;;;; history-pkg.lsp: (my first common lisp program. written in and for GCL)
;;;;
;;;; performs 'history', just like csh does.
;;;;
;;
;;	for following examples, ">" means top-level prompt.
;;
;;	>!n			re-do n'th history.
;;				same as (do-history n).
;;	>!-n			re-do current minus n'th history.
;;				same as (do-history -n).
;;	>!!			re-do last input. same as !-1.
;;				same as (do-history) or (do-history -1).
;;	>!do m n		do history from m to n.
;;				same as (do-history m n).
;;	>!history		print all history.
;;				same as (history).
;;	>!hist n		print last n history.
;;				same as (history n).
;;	>!h			print (last-histories) history.
;;				same as !hist (last-histories).
;;	>!r or !read		read history from file.
;;				same as (read-history), not (restore-history).
;;	>!w or !write		write history to file.
;;				same as (write-history).
;;	>!c or !clear		clear current history in memory, not a file.
;;				same as (clear-history).
;;	>!edit			edit and re-read history file.
;;				same as (edit-history).
;;	>!ed n			edit and eval n'th history.
;;				same as (edit-history n).
;;	>!ed -n			edit and eval current minus n'th history.
;;				same as (edit-history -n).
;;	>!e			edit and eval last history.
;;				same as (edit-history -1).
;;	>!help			print help message.
;;
;;	un-keybinded-functions:
;;		(restore-history)	read history's history and history file.
;;		(reset-history)		reset history # to 0, then read history.
;;		(history-size)		returns current history buffer size.
;;		(history-size n)	set history buffer size to n.
;;		(history-with-number t) print history with number (default).
;;		(history-with-number nil) without number.
;;		(history-with-number)   returns current settings.
;;		(last-histories)	returns current number of prints.
;;		(last-histories n)	set default number of prints. see !h.
;;
;;	*NOTE*: "!edit family" uses function (ed ..) to invoke your editor.
;;
;; Copyright (C) 2000,2001 Isao Daigo.
;;
;;		Version 0.1.0 : 2000/12/08 Implement almost all.
;;			0.1.3 : 2000/12/20 Implement !edit family
;;			0.2.0 : 2000/12/22 Implement !load
;;			0.2.1 : 2000/12/23 Bug fix
;;			0.3.0 : 2001/01/02 Can resize history buffer
;;			0.4.3 : 2001/01/04 Delete !load & add (do-history m n).
;;                      0.5.0 : 2001/02/06 Add (history-with-number)
;;			0.6.0 : 2001/05/08 Add !do
;;			0.7.1 : 2001/05/19 Add (restore-history) & 1 bug fixed.
;;			0.7.2 : 2001/06/23 1 bug fixed in (write-history).
;;			0.8.0 : 2001/06/23 Add (last-live-history).
;;
;; Licensed under GNU Library General Public License.
;;
;; ҥȥꡦѥåѤ뤿ɬפѹˤĤơ
;;
;; ȥåץ٥롦롼פǤ'read''top-level-read'Ȥؿ̾ѹ
;; ɬפޤȥåץ٥롦롼פǤ'top-level-read'ϼΤ褦
;; ˡñʤwrapperȤƤؿ̾exportƤޤ
;;
;; (export 'top-level-read)
;; (defun top-level-read () (read *standard-input* nil *eof*))
;;
;; Υҥȥꡦѥåǽ뤿ˡtop-level-read򼡤Τ褦
;; եǺ롣
;; 
;; (defun system:top-level-read (&aux form) ; redef 'top-level-read'
;;   (setf form (read *standard-input* nil *eof*))
;;   (add-to-str-hist (format nil "~s" form)) ; 'freeze' and store it.
;;   (return-from system:top-level-read form))
;;
;; ʾǥҥȥ굡ǽѤǤ褦ˤʤ뤬ץץȤ˥ҥȥֹ椬
;; ɽ褦ˡȥåץ٥롦롼פΥץץȤɽʬؿ
;; Ȥδؿ̾exportƤΤ˾ޤ
;;
;; (export 'prompt-string)
;; (defun prompt-string () "> ")
;;
;; ץץȤΰȤƥҥȥֹɽǤ褦ˤ뤿ˡδ
;; 򡢤Υѥå(history-pkg.lsp)ǼΤ褦˺롣
;;
;; (defun system:prompt-string () (format nil "[~a]> " *hist-num*))
;;
;; ʾ塣

(make-package 'history-pkg :use '(lisp system))

(in-package 'history-pkg)

(export
 '(add-to-str-hist
   clear-history
   current-history
   current-prompt-length
   do-history
   edit-history
   get-hist
   help-history
   hist-buf
   hist-range-p
   history
   history-buffer
   history-count
   history-file
   history-search-backward
   history-search-forward
   history-size
   history-version
   history-with-number
   last-histories
   last-live-history
   read-history
   reset-history
   restore-history
   write-history ))

;;
;; editable INITIAL values
;;
(defparameter *hist-size* 1024	"history's ring buffer size")
(defparameter *number-of-prints* (min 25 *hist-size*))

(defconstant  *eos* (cons nil nil))

(defparameter *history-version* "2001/09/01 Version 0.8.1")
(defparameter *history-file*   ".gcl.history")
(defparameter *history-buffer* ".gcl.history.buffer")
(defparameter *echo-history* t "normaly t, then echo history")
(defparameter *no-echo* '(history ~edit-history help-history))
(defparameter *history-with-number* t)

(defvar *hist-num* 0 "number of forms from Lisp system invoked")
(defvar *hist-pointer* 0 "buffer pointer: 0..(1- *hist-size*)")
(defvar *hist-buf* (make-array *hist-size* :initial-element *eos*))
(defvar *read-history* nil "already read history buffer ?")
(defvar *count* 0)

;;;
;;; Хåե󥰤ꡢϥ⡼ɤˤ롣
;;; ϰ¸ؿʤΤǡܿκݤˤϽɬפˤʤ롣
;;; line-edit-pkg ʻѤƤƥ顼򵯤֤
;;; 롣ʻѤƤʤ硢Τñ nilǤ褤
;;;
(defun cooked-mode ()
  (system "stty cooked echo"))

(defun history-version () *history-version*)

(defun system:prompt-string () (format nil "[~a]> " *hist-num*))

(defun current-prompt-length ()
  (length (system:prompt-string)))

(defun add-to-str-hist (str)
  (setf (aref *hist-buf* (set-hist-pointer)) str))

(defmacro strcat (&rest str)
  (declare (string str))
  (cons 'concatenate (cons ''string str)))

(defun history (&optional m n) "print history"
  (cond
   ((null m)				;all history
    (prt-hist (- *hist-num* *hist-size*) (1- *hist-num*)))
   ((null n)				;Last m history
    (if (not (integerp m)) (history-error "Argument must be integer."))
    (prt-hist (- *hist-num* m) (1- *hist-num*)))
   ((not (and (integerp m) (integerp n)))
    (history-error "Argument must be integer."))
   (t (prt-hist m n))))			;history from m to n

(defun prt-hist (m n)
  (declare (fixnum m n))
  (cond
   ((minusp m) (prt-hist2 0 n))
   (t (prt-hist2 m n))))

(defun prt-hist2 (m n &aux fmt tmp) "print history m to n"
  (declare (fixnum m n) (string fmt))
  (setf fmt (strcat "\~" (format nil "~d" (columns (1- *hist-size*))) ",\'0d:"))
  (dolist (i (buf-index m n) t)
    (setf tmp (aref *hist-buf* (cdr i)))
    (when (not (eq tmp *eos*))
      (if (history-with-number) (format t fmt (car i)))
      (princ tmp)
      (terpri))))

(defun hist (stream char)
  (echo-hist (read-hist-exp stream char)))

(set-macro-character #\! #'hist)

(defun read-hist-exp (stream char &aux tmp i)
  (declare (ignore char))
  (if (eql (peek-char nil stream) #\!)
      (setf tmp (read-char stream))
    (setf tmp (read stream)))
  (when (symbolp tmp)
    (setf tmp (string-downcase (symbol-name tmp))))
  (cond
   ((hist-range-p tmp)			;!n
    (~do-history tmp))
   ((and				;!-n
     (integerp tmp)
     (minusp tmp))
    (~do-history tmp))
   ((equal tmp #\!)			;!!
    (~do-history -1))
   ((not (stringp tmp)) nil)
   ((string-equal tmp "do")		;!do m n
    `(do-history ,(read stream) ,(read stream)))
   ((string-equal tmp "history")	;!history
    '(history))
   ((string-equal tmp "hist")		;!hist n
    `(history ,(read stream)))
   ((string-equal tmp "h")		;!h
    `(history ,*number-of-prints*))
   ((or					;!read or !r
     (string-equal tmp "read")
     (string-equal tmp "r")) '(read-history))
   ((or					;!write or !w
     (string-equal tmp "write")
     (string-equal tmp "w")) '(write-history))
   ((or					;!clear or !c
     (string-equal tmp "clear")
     (string-equal tmp "c")) '(clear-history))
   ((string-equal tmp "edit")		;!edit
    '(edit-history))
   ((string-equal tmp "ed")		;!ed n or !ed -n
    (setf i (read stream))
    (if (minusp i)
	(~edit-history (- *hist-num* (- i)))
      (~edit-history i)))
   ((string-equal tmp "e")		;!e
    (~edit-history (1- *hist-num*)))
   ((string-equal tmp "help")
    '(help-history))
   (t (history-error (format nil "Don't support \'\!~a\'." tmp)))))

(defun do-history (&optional m n &aux val)
  (cond
   ((null m)
    (eval (echo (~do-history -2))))
   ((not (integerp m))
    (history-error "Argument must be integer."))
   ((and (null n) (minusp m))
    (eval (echo (~do-history (1- m)))))
   ((and (null n) (>= m 0))
    (eval (echo (~do-history m))))
   (t (do ((i m (incf i))) ((> i n) t)
	(setf val (multiple-value-list (eval (echo (~do-history i)))))
	(dolist (i val)
	  (princ #\space)
	  (prin1 i))
	(terpri)))))

(defun ~do-history (n)
  (declare (fixnum n))
  (cond
   ((>= n 0)
    (string-to-form (aref *hist-buf* (get-hist n))))
   ((minusp n)
    (string-to-form (aref *hist-buf* (get-hist (- *hist-num* (- n))))))))

(defun edit-history (&optional arg)
  (eval (echo (~edit-history arg))))

(defun ~edit-history (&optional arg)
  (cond
   ((null arg)
    (edit-history-file))
   ((integerp arg)
    (edit-nth-history arg))))

(defun edit-history-file ()
  (write-history)
  (ed *history-file*)
  (clear-history)
  (restore-history))

(defun edit-nth-history (n) "edit n'th history"
  (declare (fixnum n))
  (if (minusp n) (setf n (- *hist-num* (- n) 1)))
  (cond
   ((hist-range-p n)
    (write-to-buffer n)
    (ed *history-buffer*)
    (read-from-buffer))))

(defun string-to-form (str)
  (unless (stringp str) (return-from string-to-form nil))
  (with-input-from-string (stream str) (read stream)))

(defun hist-range-p (n)
  (declare (fixnum n))
  (cond
   ((not (integerp n)) nil)
   ((minusp n) nil)
   ((and
     (< n *hist-num*)
     (>= n (- *hist-num* *hist-size*))) t)))

(defun set-hist-pointer ()
  (prog1
   (setf *hist-pointer* (mod *hist-num* *hist-size*))
   (history-count t)
   (incf *hist-num*)))

(defun history-count (&optional inc-p)
  (cond
   ((null inc-p) *count*)
   (t (setf *count* (min *hist-size* (1+ *count*))))))

(defun get-hist (num &aux tmp)
  (declare (fixnum num))
  (setf tmp (cdr (assoc num (current-buf))))
  (if (null tmp) (history-error "Bad history number.")
    (return-from get-hist tmp)))

(defun last-hist ()
  (the string (aref *hist-buf* (get-hist (1- *hist-num*)))))

#|
;;
;; here is sample data
;;
(add-to-history '(car '(a b)))
(add-to-history '(princ 'a))
(add-to-history '*hist-pointer*)
(add-to-history '*hist-buf*)
(add-to-history '(add-to-history '(cdr '(a b))))
(add-to-history '(floor 9 4))
(add-to-history '(history))
|#

#|
for example: (assume *hist-size* == 30)

 > (buf-index 25 35)
 ((25 . 25) (26 . 26) (27 . 27) (28 . 28) (29 . 29)
  (30 . 0) (31 . 1) (32 . 2) (33 . 3) (34 . 4) (35 . 5))
 > (buf-index2 25 35)
 (25 26 27 28 29 0 1 2 3 4 5)
 > (index-list 25 7)
 (25 26 27 28 29 0 1)
 >

|#

;;
;; Pentium III-700MHz/GCL 2.3.8-Beta
;;
;; > (time (dotimes (i 1000) (current-buf)))
;; real time : 4.580 secs
;; run time  : 4.560 secs
;;
;; (current-buf) == 4.580 msecs
;;
(defun current-buf ()
  (buf-index (- *hist-num* *hist-size*) (1- *hist-num*)))

(defun buf-index (m n)
  (declare (fixnum m n))
  (if (minusp m) (setf m 0))
  (mapcar 'cons (make-num-list m n) (buf-index2 m n)))

(defun make-num-list (m n &aux tmp)
  (declare (fixnum m n))
  ;(if (minusp m) (make-num-list 0 n))
  (do ((i n (decf i))) ((< i m) tmp)
      (push i tmp)))

(defun buf-index2 (m n)
  (declare (fixnum m n))
  (cond ;((minusp m) (buf-index2 0 n))
	((out-of-index m n) nil)
	(t (index-list
	    (- *hist-pointer* (- *hist-num* m 1))
	    (1+ (- n m))))))

(defun index-list (from n &aux tmp)
  (declare (fixnum from n))
  (dotimes (i n (reverse tmp))
	   (push (mod (+ from i) *hist-size*) tmp)))

#|
where { m , n } should satisfy
	{ 0 <= m <= n < *hist-num* } and { (n - m) < *hist-size* }
|#

(defun out-of-index (m n)
  (declare (fixnum m n))
  (cond ((or
	  (minusp m)
	  (> m n)
	  (>= n *hist-num*)
	  (>= (- n m) *hist-size*)) t)))

(defun hist-buf (n)
  (aref *hist-buf* n))

;;
;; Read history from (history-file)
;;
(defun read-history ()
  (when (probe-file (history-file))
    (with-open-file (stream (history-file) :direction :input)
      (let (tmp count)
	(setf *read-history* t)
	(setf count 0)
	(loop
	  (setf tmp (read stream nil *eos*))
	  (when (eq tmp *eos*) (return count))
	  (incf count)
	  (add-to-str-hist tmp))))))

(defun read-from-buffer ()
  (when (probe-file *history-buffer*)
    (with-open-file (stream (history-buffer) :direction :input)
      (let (tmp)
	(setf tmp (read stream nil *eos*))
	(if (eq tmp *eos*) nil tmp)))))

;;
;; Write current on memory history to (history-file)
;;
(defun write-history ()
  (record-history)
  (with-open-file (stream (history-file) :direction :output)
    (let (tmp num)
      (setf num 0)
      (dolist (i (current-buf) num)
	(setf tmp (aref *hist-buf* (cdr i)))
	(when (not (eq tmp *eos*))
	  (incf num)
	  (prinx tmp stream)
	  (terpri stream))))))

(defun write-to-buffer (n) "write n'th history to buffer"
  (declare (fixnum n))
  (with-open-file (stream (history-buffer) :direction :output)
    (let (tmp i)
      (setf i (assoc n (current-buf)))
      (setf tmp (aref *hist-buf* (cdr i)))
      (princ tmp stream)
      (terpri stream))))

;;
;; Clear current history in memory, not a file.
;;
(defun clear-history ()
  (prog1
   (dotimes (i *hist-size* t)
	    (setf (aref *hist-buf* i) *eos*))
   (setf *count* 0)
   (setf *hist-num* 0)
   (setf *hist-pointer* 0)))

(defun history-file (&optional fname)
  (cond
   ((null fname)
    *history-file*)
   ((stringp fname)
    (setf *history-file* fname))
   ((pathnamep fname)
    (setf *history-file* (namestring fname)))))

(defun history-history ()	;History's history
  (strcat (history-file) ".history"))

(defun history-buffer (&optional fname)
  (cond
   ((null fname)
    *history-buffer*)
   ((stringp fname)
    (setf *history-buffer* fname))
   ((pathnamep fname)
    (setf *history-buffer* (namestring fname)))))

(defun history-size (&optional size)
  "returns current history-size, or set to it"
  (cond
   ((null size) *hist-size*)
   ((or
     (not (integerp size))
     (zerop size)) nil)
   ((discardable *history-file*)
    (write-history)
    (setf *hist-size* size)
    (setf *hist-buf* (make-array *hist-size*))
    (clear-history)
    (read-history))))

(defun restore-history ()
  (when (probe-file (history-history))
    (with-open-file (stream (history-history) :direction :input)
      (let (tmp)
	(setf tmp (read stream nil *eos*))
	(setf *hist-num* (- (car tmp) (cdr tmp)))
	(setf *hist-pointer* 0))))
  (read-history))

;;
;; Record history's history
;;	put *hist-num* and (history-count) with commented date & time
;;	such as "(2506 . 976) ;2001/05/19 16:59:27".
;;
(defun record-history ()
  (when (probe-file (history-history))
    (rename-file (history-history) (history-buffer)))
  (with-open-file (out (history-history) :direction :output)
    (prin1 (cons *hist-num* (history-count)) out)
    (princ " ; " out)
    (princ (simple-date) out)
    (terpri out)
    (with-open-file (in (history-buffer))
      (let (tmp)
	(loop
	  (setf tmp (read-line in nil *eos*))
	  (when (eq tmp *eos*) (return t))
	  (write-line tmp out))))))

;;
;; Reset history and read history file again.
;;
(defun reset-history ()
  (clear-history)
  (read-history))

;;
;; Returns date & time in "YYYY/MM/DD HH:MM:SS"
;;
(defun simple-date ()
  (let (tmp)
    (setf tmp (multiple-value-list (get-decoded-time)))
    (format nil "~4d/~2,'0d/~2,'0d ~2,'0d:~2,'0d:~2,'0d"
	    (sixth tmp) (fifth tmp) (fourth tmp)
	    (third tmp) (second tmp) (truncate (first tmp)))))

(defun discardable (fname)
  (cond
   ((not (probe-file fname)) t)	;Is it there?
   (*read-history* t)		;exist, and already read
   ((yes-or-no-p "Do not read history file yet, discard it?") t)))

(defun last-histories (&optional num)
  "set or return current default number of prints"
  (if (and (integerp num) (plusp num))
      (setf *number-of-prints* num)
      *number-of-prints*))

(defun columns (num &aux col)
  (declare (fixnum num))
  (setf col 1)
  (loop
   (setf num (truncate num 10))
   (if (zerop num) (return col))
   (incf col)))

(defun prinx (form &optional (stream *standard-output*))
  (write form :stream stream :escape t :pretty t))

(defun echo (form &optional (is-echo *echo-history*))
  (when is-echo
    (write form :escape t :pretty t)
    (terpri))
  (return-from echo form))

(defun echo-hist (arg)
  (cond
   ((atom arg) (echo arg))
   ((member (car arg) *no-echo*) arg)
   (t (echo arg))))

(defun echo-history (&optional (arg nil s))
  "echo history if (echo-history t). (echo-history) returns current setting"
  (cond
   ((and (null arg) (null s)) *echo-history*)
   (t (setf *echo-history* arg))))

(defun history-error (msg)
  (cooked-mode)
  (princ (format nil "~%~a~%Type \'\!help\' for help.~%" msg))
  (break))

(defun history-with-number (&optional (n t sw))
  (cond
   ((null sw) *history-with-number*)
   (t (setf *history-with-number* n))))

;;
;; ˤ õʸȥҥȥֹͿ롣Ϳ줿ҥȥֹ
;; 㤤ֹΥҥȥõʸ󤬸Ĥʤ nil 򡢸
;; ((ҥȥֹ . ֹ) ݥȰ) ֤
;;
(defun history-search-backward (str &optional hist-num)
  (declare (string str))
  (prog (nlist idx rev-lst pos first-search)
    (when (null hist-num)
      (setf first-search t)
      (setf hist-num (1- *hist-num*)))
    (setf rev-lst (reverse (current-buf)))
    (setf pos (position hist-num rev-lst :test #'(lambda (x y) (= x (car y)))))
    (when (null pos) (return nil))
    (if first-search (setf nlist (nthcdr pos rev-lst))
      (setf nlist (nthcdr (1+ pos) rev-lst)))
    (if (null nlist) (return nil))
    (dolist (i nlist nil)
      (setf idx (search str (aref *hist-buf* (cdr i))))
      (if (and (numberp idx) (zerop idx))
	  (return-from history-search-backward (list i idx))))))

;;
;; ˤ õʸȥҥȥֹͿ롣Ϳ줿ҥȥֹ
;; ŤֹΥҥȥõʸ󤬸Ĥʤ nil 򡢸
;; ((ҥȥֹ . ֹ) ݥȰ) ֤
;;
(defun history-search-forward (str &optional hist-num)
  (declare (string str))
  (prog (nlist idx lst pos)
    (if (null hist-num) (setf hist-num (1- *hist-num*)))
    (setf lst (current-buf))
    (setf pos (position hist-num lst :test #'(lambda (x y) (= x (car y)))))
    (when (null pos) (return nil))
    (setf nlist (nthcdr (1+ pos) lst))
    (if (null nlist) (return nil))
    (dolist (i nlist nil)
      (setf idx (search str (aref *hist-buf* (cdr i))))
      (if (and (numberp idx) (zerop idx))
	  (return-from history-search-forward (list i idx))))))

;;
;; ߤΥҥȥֹ򤬳ǼƤֹ
;; (ҥȥֹ . ֹ) η֤
;;
(defun current-history () (cons (1- *hist-num*) *hist-pointer*))

;;
;; ҥȥ˵ϿƤǤŤͭʥҥȥֹ򤬳Ǽ
;; ֹ (ҥȥֹ . ֹ) η֤
;;
(defun last-live-history ()
  (prog (n pos lst)
    (setf n (- (car (current-history)) (1- (history-count))))
    (setf lst (current-buf))
    (setf pos (position n lst :test #'(lambda (x y) (= x (car y)))))
    (if (null pos) (error "cann't happen."))
    (return (car (nthcdr pos lst)))))

;;
;; إסåɽ롣
;;
(defun help-history () (princ
"!n			re-do n'th history.
			same as (do-history n).
!-n			re-do current minus n'th history.
			same as (do-history -n).
!!			re-do last input. same as !-1.
			same as (do-history).
!do m n			do history from m to n.
			same as (do-history m n).
!history		print all history.
			same as (history).
!hist n			print last n history.
			same as (history n).
!h			print last histories.
			same as (last-histories).
!r or !read		read history from file.
			same as (read-history).
!w or !write		write history to file.
			same as (write-history).
!c or !clear		clear current history, not a file.
			same as (clear-history).
!edit			edit and re-read history file.
			same as (edit-history).
!ed n			edit and eval n'th history.
			same as (edit-history n).
!ed -n			edit and eval current minus n'th history.
			same as (edit-history -n).
!e			edit and eval last history.
			same as (edit-history -1).

un-keybinded-functions:
	(restore-history)	read history's history and history file.
	(reset-history)		reset history # to 0, then read history.
	(history-size)		returns current history buffer size.
	(history-size n)	set history buffer size to n.
	(last-histories)	returns current default number of prints.
	(last-histories n)	set default number of prints. see !h.
	(history-with-number t) print history with number (default).
	(history-with-number nil) without number.
	(history-file fname)	set history file to 'fname'.
	(history-file)		returns current history-file's name.
	(history-buffer fname)	set history's buffer to 'fname'.
	(history-buffer)	returns current history buffer's name.

	(help-edit)		show current editor mode commands.
")
 (return-from help-history t))
