;;;; line-edit-pkg : Read line with customizable editor command, like Emacs.
;;;
;;;	Licenced under GNU Library General Public Licence.
;;;	Copyright (C) 2001-2004 Isao Daigo (daigo@tkf.att.ne.jp).
;;;
;;;	(line-edit)			read and edit characters interactively.
;;;	(line-edit TEXT)		Edit string TEXT interactively.
;;;	(line-edit TEXT BATCH MODE)	Edit TEXT for BATCH command within MODE.
;;;
;;; Example:
;;;	> (line-edit)
;;;	"sample string"		-- input characters interactively.
;;;
;;;	> (line-edit "This is a sample text.")
;;;	"This is a text."	-- Initially cursor is at beginning of line.
;;;
;;;	> (line-edit "This is a pen." "\\C-u 3 \\M-f _long" "emacs-mode")
;;;	"This is a long pen."
;;;
;;; *NOTE*
;;;	compile or load with (push 'disable-hook *features*)
;;;	when DO NOT USE WITH history-pkg (C-n & C-p set to do nothing).
;;;
(make-package 'line-edit-pkg :use '(lisp))

(in-package 'line-edit-pkg)

(export
 '(add-allow-yank-pop-commands
   add-kill-commands
   add-prefix-commands
   advertised-redo
   advertised-undo
   append-break-char
   append-next-kill
   audio-bell
   auto-scroll-offset
   backward-char
   backward-down-list
   backward-kill-sexp
   backward-kill-white-space
   backward-kill-word
   backward-list
   backward-sexp
   backward-skip-to-normal-char
   backward-skip-white-space
   backward-up-list
   backward-word
   batch-str
   beginning-of-history
   beginning-of-line
   beginning-of-line-p
   beginning-of-text
   beginning-of-text-p
   blink-paren
   blink-paren-p
   blink-paren-deley
   blink-paren-just-inserted
   break-char-p
   call-by-keyboard-p
   call-last-kbd-macro
   call-last-kbd-macro-forever
   capitalize-word
   char-to-digit
   clear-global-mark
   clear-global-set-key
   clear-line
   command-trace
   command-trace-list
   command-undo-info
   control-to-char
   cooked-mode
   copy-to-register
   corres-paren
   current-char
   editor-mode
   debug-level
   debug-print
   define-key
   defined-key-list
   delete-backward-char
   delete-char
   delete-horizontal-space
   digit-to-char
   display-line
   distill-line
   do-nothing
   do-write-file
   down-list
   downcase-word
   edit-kbd-macro
   empty-text-p
   enable-audio-bell
   end-command-trace
   end-input-p
   end-input
   end-kbd-macro
   end-kbd-macro-p
   end-of-history
   end-of-line
   end-of-line-p
   end-of-text
   end-of-text-p
   exactly-matched
   exchange-point-and-mark
   expand-history
   find-file
   finish-global-set-key
   forward-char
   forward-list
   forward-sexp
   forward-word
   get-command
   getch
   getchar
   get-global-mark
   global-keymap
   global-mark-file
   global-set-key
   goto-global-mark
   goto-history
   help-edit
   highlight-mode
   highlight-mode-putc
   highlight-paren
   highlight-paren-deley
   highlight-paren-just-inserted
   hscroll-unit
   in-string-or-literal-p
   in-word-p
   init-line-edit
   insert
   insert-file
   insert-parenthesis
   insert-register
   inspect-keycode
   inverse-off
   inverse-on
   isearch-forward
   jump-and-return-balanced-paren
   just-one-space
   kbd-macro-query
   kbd-macro-query-time
   kill-commands-p
   kill-line
   kill-region
   kill-ring-max
   kill-ring-save
   kill-sexp
   kill-text
   kill-white-space
   kill-word
   last-kbd-macro
   last-char
   last-command
   last-repeat-count
   line-edit
   line-edit-break
   line-edit-command-loop
   line-edit-put-newline
   line-edit-version
   line-length
   line-size
   lisp-extension
   load-global-mark
   load-macro
   macro-file
   macro-mode-p
   mark
   mark-sexp
   mark-word
   max-column
   move-cursor-left
   move-cursor-right
   move-cursor-to
   move-point-left
   move-point-right
   move-point-to
   move-to-matching-paren
   multiply-repeat-count-for
   narrow-break-char
   need-refresh-line
   need-refresh-line-p
   newline-symbol
   next-line
   next-newline-position
   next-word
   normal-char-p
   oldest-history-p
   pack
   packed-line
   packed-text
   parenthesis-list
   point
   pop-kill-ring
   postlude-input
   prefix-argument
   prefix-commands-p
   previous-line
   previous-newline-position
   push-command-trace-list
   push-keyboard-macro
   push-kill-ring
   push-redo-info
   push-undo-info
   putc
   putch
   putnum
   quoted-insert
   raw-mode
   read-number
   recenter
   recenter-zero
   redo-info
   register
   registers-file
   remove-break-char
   repeat-count
   reset-parenthesis-list
   reset-repeat-count
   reset-scroll
   reset-undo-info
   restore-line
   restore-registers
   reverse-do
   save-buffer
   save-global-mark
   save-line
   save-macro
   save-registers
   scroll-down
   scroll-left
   scroll-right
   scroll-up
   search-from
   select-repeat-count
   self-insert
   self-insert-newline
   self-insert-space
   self-insert-string
   set-balanced-paren
   set-break-code
   set-command-trace
   set-command-trace-list
   set-command-undo-info
   set-editor-mode
   set-global-mark
   set-last-command-inhibit-list
   set-mark-command
   set-redo-info
   set-register
   set-repeat-count-for
   set-undo-control
   set-undo-info
   skip-bar-string
   skip-dquoted-string
   skip-string-or-literal
   skip-to-normal-char
   skip-white-space
   start-command-trace
   start-kbd-macro
   start-kbd-macro-p
   string-to-number
   sync-line
   sync-max
   sync-point
   sync-point-and-cursor
   sync-simple
   text-length
   to-circumflex-accent-char
   trace-list
   transpose-chars
   transpose-sexps
   transpose-words
   undo-control
   undo-info
   undo-limit
   universal-argument
   universal-argument-command-p
   unpack
   unset-balanced-paren
   up-list
   upcase-word
   view-all-registers
   view-global-mark
   view-register
   visual
   visual-end
   wide-break-char
   what-cursor-position
   white-space-p
   write-file
   write-file-to
   writeln
   xsleep
   yank
   yank-pop
   zap-to-char))

;;;
;;; This package's version string
;;;
(defconstant *line-edit-version* "2004/07/22 line-edit-pkg: Version 9.1.6")

(defconstant *small-a*			"small-a")
(defconstant *small-b*			"small-b")
(defconstant *small-c*			"small-c")
(defconstant *small-d*			"small-d")
(defconstant *small-e*			"small-e")
(defconstant *small-f*			"small-f")
(defconstant *small-g*			"small-g")
(defconstant *small-h*			"small-h")
(defconstant *small-i*			"small-i")
(defconstant *small-j*			"small-j")
(defconstant *small-k*			"small-k")
(defconstant *small-l*			"small-l")
(defconstant *small-m*			"small-m")
(defconstant *small-n*			"small-n")
(defconstant *small-o*			"small-o")
(defconstant *small-p*			"small-p")
(defconstant *small-q*			"small-q")
(defconstant *small-r*			"small-r")
(defconstant *small-s*			"small-s")
(defconstant *small-t*			"small-t")
(defconstant *small-u*			"small-u")
(defconstant *small-v*			"small-v")
(defconstant *small-w*			"small-w")
(defconstant *small-x*			"small-x")
(defconstant *small-y*			"small-y")
(defconstant *small-z*			"small-z")
(defconstant *capital-A*		"capital-A")
(defconstant *capital-B*		"capital-B")
(defconstant *capital-C*		"capital-C")
(defconstant *capital-D*		"capital-D")
(defconstant *capital-E*		"capital-E")
(defconstant *capital-F*		"capital-F")
(defconstant *capital-G*		"capital-G")
(defconstant *capital-H*		"capital-H")
(defconstant *capital-I*		"capital-I")
(defconstant *capital-J*		"capital-J")
(defconstant *capital-K*		"capital-K")
(defconstant *capital-L*		"capital-L")
(defconstant *capital-M*		"capital-M")
(defconstant *capital-N*		"capital-N")
(defconstant *capital-O*		"capital-O")
(defconstant *capital-P*		"capital-P")
(defconstant *capital-Q*		"capital-Q")
(defconstant *capital-R*		"capital-R")
(defconstant *capital-S*		"capital-S")
(defconstant *capital-T*		"capital-T")
(defconstant *capital-U*		"capital-U")
(defconstant *capital-V*		"capital-V")
(defconstant *capital-W*		"capital-W")
(defconstant *capital-X*		"capital-X")
(defconstant *capital-Y*		"capital-Y")
(defconstant *capital-Z*		"capital-Z")
(defconstant *digit-1*			"digit-1")
(defconstant *digit-2*			"digit-2")
(defconstant *digit-3*			"digit-3")
(defconstant *digit-4*			"digit-4")
(defconstant *digit-5*			"digit-5")
(defconstant *digit-6*			"digit-6")
(defconstant *digit-7*			"digit-7")
(defconstant *digit-8*			"digit-8")
(defconstant *digit-9*			"digit-9")
(defconstant *digit-0*			"digit-0")
(defconstant *plus-sign*		"plus-sign")
(defconstant *less-than-sign*		"less-than-sign")
(defconstant *equals-sign*		"equals-sign")
(defconstant *greater-than-sign*	"greater-than-sign")
(defconstant *dollar-sign*		"dollar-sign")
(defconstant *backquote*		"backquote")
(defconstant *circumflex-accent*	"circumflex-accent")
(defconstant *tilde*			"tilde")
(defconstant *sharp-sign*		"sharp-sign")
(defconstant *percent-sign*		"percent-sign")
(defconstant *ampersand*		"ampersand")
(defconstant *asterisk*			"asterisk")
(defconstant *at-sign*			"at-sign")
(defconstant *left-square-bracket*	"left-square-bracket")
(defconstant *backslash*		"backslash")
(defconstant *right-square-bracket*	"right-square-bracket")
(defconstant *left-brace*		"left-brace")
(defconstant *vertical-bar*		"vertical-bar")
(defconstant *right-brace*		"right-brace")
(defconstant *exclamation-mark*		"exclamation-mark")
(defconstant *double-quote*		"double-quote")
(defconstant *single-quote*		"single-quote")
(defconstant *left-parenthesis*		"left-parenthesis")
(defconstant *right-parenthesis*	"right-parenthesis")
(defconstant *comma*			"comma")
(defconstant *underscore*		"underscore")
(defconstant *minus-sign*		"minus-sign")
(defconstant *period*			"period")
(defconstant *slash*			"slash")
(defconstant *colon*			"colon")
(defconstant *semicolon*		"semicolon")
(defconstant *question-mark*		"question-mark")

(defconstant *letter-a*			"letter-a")
(defconstant *letter-b*			"letter-b")
(defconstant *letter-c*			"letter-c")
(defconstant *letter-d*			"letter-d")
(defconstant *letter-e*			"letter-e")
(defconstant *letter-f*			"letter-f")
(defconstant *letter-g*			"letter-g")
(defconstant *letter-h*			"letter-h")
(defconstant *letter-i*			"letter-i")
(defconstant *letter-j*			"letter-j")
(defconstant *letter-k*			"letter-k")
(defconstant *letter-l*			"letter-l")
(defconstant *letter-m*			"letter-m")
(defconstant *letter-n*			"letter-n")
(defconstant *letter-o*			"letter-o")
(defconstant *letter-p*			"letter-p")
(defconstant *letter-q*			"letter-q")
(defconstant *letter-r*			"letter-r")
(defconstant *letter-s*			"letter-s")
(defconstant *letter-t*			"letter-t")
(defconstant *letter-u*			"letter-u")
(defconstant *letter-v*			"letter-v")
(defconstant *letter-w*			"letter-w")
(defconstant *letter-x*			"letter-x")
(defconstant *letter-y*			"letter-y")
(defconstant *letter-z*			"letter-z")

(defconstant *ctrl-@*			#\^@)
(defconstant *ctrl-a*			#\^A)
(defconstant *ctrl-b*			#\^B)
(defconstant *ctrl-c*			#\^C)
(defconstant *ctrl-d*			#\^D)
(defconstant *ctrl-e*			#\^E)
(defconstant *ctrl-f*			#\^F)
(defconstant *ctrl-g*			#\^G)
(defconstant *ctrl-h*			#\^H)
(defconstant *ctrl-i*			#\^I)
(defconstant *ctrl-j*			#\^J)
(defconstant *ctrl-k*			#\^K)
(defconstant *ctrl-l*			#\^L)
(defconstant *ctrl-m*			#\^M)
(defconstant *ctrl-n*			#\^N)
(defconstant *ctrl-o*			#\^O)
(defconstant *ctrl-p*			#\^P)
(defconstant *ctrl-q*			#\^Q)
(defconstant *ctrl-r*			#\^R)
(defconstant *ctrl-s*			#\^S)
(defconstant *ctrl-t*			#\^T)
(defconstant *ctrl-u*			#\^U)
(defconstant *ctrl-v*			#\^V)
(defconstant *ctrl-w*			#\^W)
(defconstant *ctrl-x*			#\^X)
(defconstant *ctrl-y*			#\^Y)
(defconstant *ctrl-z*			#\^Z)
(defconstant *ctrl-[*			#\^[)
(defconstant *ctrl-\*			#\^\\)
(defconstant *ctrl-]*			#\^])
(defconstant *ctrl-^*			#\^^)
(defconstant *ctrl-_*			#\^_)
(defconstant *meta-code*		#\^[)	;key-code for Meta key.
(defconstant *alt-code*			#\^[)	;key-code for Alt key.

(defconstant *rubout*			"[Rubout]")
(defconstant *space*			"[Space]")
(defconstant *backspace*		"[Backspace]")
(defconstant *delete*			"[delete]")
(defconstant *return*			"[Return]")
(defconstant *linefeed*			"[Linefeed]")

(defconstant *control-prefix*		'(#\C #\-))
(defconstant *meta-prefix*		'(#\M #\-))
(defconstant *control-string*		"control")
(defconstant *meta-string*		"meta")
(defconstant *backslash-ch*		#\\)
(defconstant *left-square-bracket-ch*	#\[)
(defconstant *right-square-bracket-ch*	#\])
(defconstant *vertical-bar-ch*		#\|)
(defconstant *underscore-ch*		#\_)

(defconstant *eos* (cons nil nil))
(defconstant ESC *ctrl-[*)
(defconstant *number-char* '(#\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9))
(defconstant *white-space-char*
  '(#\Space #\Tab #\Newline #\Page #\Backspace #\Return #\Rubout))
(defconstant *narrow-break-char*
  '(#\" #\# #\' #\= #\| #\\ #\` #\; #\, #\. #\< #\> #\( #\) #\[ #\] #\{ #\} #\? #\+ #\~))
(defconstant *wide-break-char* (append *narrow-break-char* '(#\- #\* #\: #\/ #\$ #\@)))

(defparameter *left-parenthesis-list*  '(#\( #\< #\{) )
(defparameter *right-parenthesis-list* '(#\) #\> #\}) )
(defparameter *parenthesis-list*
  (append *left-parenthesis-list* *right-parenthesis-list*))
(defparameter *otherwise-function* 'self-insert)
(defparameter *kill-commands*
  '(kill-line kill-region kill-word backward-kill-word kill-sexp zap-to-char
	      kill-ring-save kill-text append-next-kill))
(defparameter *prefix-commands* '(universal-argument))
(defparameter *allow-yank-pop-commands*
  (append '(yank yank-pop) *prefix-commands*))
(defparameter *last-command-inhibit-list* nil)

;;;
;;; ƥȤƤγƼ͡
;;;
(defvar ctrl-g *ctrl-g*)		;break code for C-g.
(defvar *kill-ring-max* 30)		;kill-ring stack size.
(defvar *blink-paren-deley* 3/10)	;how many sec. for BLINKING paren ?
(defvar *blink-paren-just-inserted* t)	;Is blink paren only inserted ?
(defvar *undo-limit* 300)		;undo limit. nil means unlimited.
(defvar *line-size* 72)			;physical line size.
(defvar *auto-scroll-offset* t)		;auto scroll offset. t means auto.
(defvar *newline-symbol* #\$)		;display character for #\Newline.
(defvar *kbd-macro-query-time* 1)	;How many sec. waiting for C-x q ?
(defvar *audio-bell* t)			;enable audio bell ?
(defvar *break-char* *wide-break-char*)	;default definition.
(defvar *editor-mode* "emacs-mode")	;default editor mode.
(defvar *lisp-extension* ".lsp")	;default source file extension.
(defvar *registers-file* ".gcl.registers")
(defvar *global-mark-file* ".gcl.global-mark")
(defvar *macro-file* ".gcl.last-kbd-macro")

;;;
;;; ΥѥåФѿ
;;;
(defvar *text* nil)			;current text (=list of characters).
(defvar *point* 0)			;point position (just before cursor).
(defvar *ncl* 0)			;Number of Characters in Line.
(defvar *mark* nil)			;Mark of beginning or end of region.
(defvar *kill-ring* nil)		;kill-ring buffer.
(defvar *hscroll-unit* nil)		;horizontal scrolling unit.
(defvar *hscroll-size* 0)
(defvar *suspend-switch* t)
(defvar *line-stack* nil)
(defvar *echo* t)
(defvar *last-yank* 0)
(defvar *search-from* nil)
(defvar *registers* nil)
(defvar *repeat-for* nil)
(defvar *last-repeat-count* nil)
(defvar *last-prefix-argument* nil)
(defvar *line-edit-put-newline* t)
(defvar *allow-yank-pop* nil)
(defvar *backward-kill-commands* nil)
(defvar *save-mark-info* nil)
(defvar *command-undo-info* nil)
(defvar *undo-control* nil)
(defvar *undo-info* nil)
(defvar *last-undo-info* nil)
(defvar *exec-undo* nil)
(defvar *redo-info* nil)
(defvar *exec-redo* nil)
(defvar *max-undo-length* 0)		;for statistics information.
(defvar *cursor-position* 0)
(defvar *need-refresh-line* nil)
(defvar *newline-position* nil)
(defvar *last-newline-position* nil)
(defvar *keyboard-macro* nil)
(defvar *record-macro* nil)
(defvar *macro-mode* nil)
(defvar *last-command* nil)
(defvar *command-trace* nil)
(defvar *command-trace-list* nil)
(defvar *debug-level* 0)
(defvar *last-char* nil)
(defvar *global-keymap* nil)
(defvar *defined-key-list* nil)
(defvar *inspect-keycode* nil)
(defvar *call-by-keyboard* nil)
(defvar *global-mark* nil)
(defvar *implicit-load-path* nil)

;;
;; MACRO DEFINITION PART
;;

;;;
;;; ƬץץȤɽ;ɾ
;;; ץץȤǾä줿ԤƤ롣
;;;
(defmacro with-prompt (prompt &rest body)
  `(progn
     (clear-line)
     (dolist (ch (unpack ,prompt)) (putch ch))
     (finish-output)
     ,@body
     (display-line)))

;;;
;;; Хåե󥰤ʤʤϥ⡼ɤǰɾ롣
;;;
(defmacro with-raw-input-mode (&rest body)
  `(progn
     (raw-mode)
     (unwind-protect
	 ,@body
       (cooked-mode)))) 

;;;
;;; ¸ؿ
;;;
;;; »֤ǤιԽ¸뤿ɬ׺Ǿ¤ε¸ؿƤ롣
;;; Υѥåܿݤˡǽ 6Ĥε¸ؿ򵡼˹碌
;;; ɬפ롣
;;;
;;; pure-delete-lineؿpure-delete-charؿˤäǤ뤬ǽ
;;; иΨΤ˲浡ǽˤäľ뤳Ȥ˾ޤ
;;;
;;; Ĵɽ⡼ɤʸζĴɽ˻Ѥ롣դɽɽ
;;; ɡ̾ɽȰۤʤΤǤвǤ褤ڤΥס
;;; 󥹤ݡȤƤʤ硢ؿΤ nilǤ褤
;;; ʸζĴɽϹԤʤ
;;;

;;;
;;; Хåե󥰤ʤʤϥ⡼ɤˤ롣
;;;
(defun raw-mode ()
  (system "stty raw -echo"))

;;;
;;; Хåե󥰤ꡢϥ⡼ɤ᤹
;;;
(defun cooked-mode ()
  (system "stty cooked echo"))

;;;
;;; 򱦤 nʸư롣
;;;
(defun pure-move-right (n)
  (declare (fixnum n))          ;GCL 2.4.0ǤϥХΤ (declare ...)
  (when (> n 0)                 ;ѥǤʤ2.3.8beta, 2.4.4, 2.5.2
    (putc ESC)			;ʤȤǧƤ롣2.4.0ꡣ
    (putc #\[)
    (putnum n)
    (putc #\C)))

;;;
;;; 򺸤 nʸư롣
;;;
(defun pure-move-left (n)
  (declare (fixnum n))
  (when (> n 0)
    (putc ESC)
    (putc #\[)
    (putnum n)
    (putc #\D)))

;;;
;;; ֤ 1ʸõơʹߤʸ1ʬŤˤĤ롣
;;; ɽΤߤä
;;;
(defun pure-delete-char ()
  (putc ESC)
  (putc #\[)
  (putc #\P))

;;;
;;; raw-modeǤ(terpri)ԤιԤƬ˥ư롣
;;;
(defun writeln ()
  (putc #\Return)
  (putc #\Linefeed))

;;;
;;; ֤ޤǤʸõ롣
;;; б륨ס󥹤ݡȤƤʤ
;;; ȡȤƤؤ롣
#|
;Pentium III 700MHz/Linux 2.4.0-test8 GNU Common Lisp 2.3.8 Beta - Interpreted.
;
;> (time (dotimes (i 1000) (line-edit-pkg::pure-delete-line)))
;real time : 1.250 secs
;run time  : 1.150 secs
;
(defun pure-delete-line ()
  (dotimes (i (- (line-size) (cursor-position))) (pure-delete-char)))
|#
;;;
;> (time (dotimes (i 1000) (line-edit-pkg::pure-delete-line)))
;
;real time : 0.020 secs
;run time  : 0.020 secs
;
(defun pure-delete-line ()
  (putc ESC)
  (putc #\[)
  (putc #\K))

;;;
;;; ֶĴɽפΥ󡿥դ椹롣
;;;
(defun highlight-mode (mode)
  (cond
   ((null mode) (inverse-off))
   (t (inverse-on))))

;;;
;;; ȿžɽ⡼ɤˤ롣
;;;
(defun inverse-on ()
  (putc ESC)
  (putc #\[)
  (putnum 7)
  (putc #\m))

;;;
;;; ȿžɽ⡼ɤλ롣
;;;
(defun inverse-off ()
  (putc ESC)
  (putc #\[)
  (putc #\m))

;;;
;;; Helpեɽ뤿 lessƤӽФ
;;; ޥɤƤӽФ systemؿȤʤ
;;; ƥǤ read-line Ȥä롣
;;;
(defun less (fname)
  (declare (string fname))
  (system (concatenate 'string "less " fname))
  (return-from less t))

;;;
;;; ʸηϤ˰¸ؿASCIIʸݤ֤
;;; GCL 2.4.4 graphic-char-p˥ߥꡢ(char-code ch) >= 32
;;; Ǥ褦ʸ chФ (graphic-char-p ch)ϡ٤ t ֤
;;; Ƥޤϥե character.d655
;;;
;;;        if (' ' <= i && ' ' < '\177')
;;; 
;;;        if (' ' <= i && i < '\177')
;;;
;;; Ƚƺƥѥ뤹Ф褤ХΤ륷ƥǤư
;;; 褦ǰΤ˰ʲδؿƻѤƤ롣
;;;
;;; ΥХ GCL 2.5.1ǽ줿
;;;
(defun ascii-char-p (ch)
  (declare (character ch))
  (< (char-code ch) 128))


;;; MACHINE INDEPENDENT FUNCTIONS.

;;;
;;; ΥѥåΥСȸߤΥǥ⡼֤̾
;;;
(defun line-edit-version ()
  (format nil "~a (~a)" *line-edit-version* (editor-mode)))

;;;
;;; ǥ⡼ɤꤹ롣ʤǼ¹Ԥȸߤ
;;; ⡼֤̾⡼̾ʸǤꡢ
;;; 	"emacs-mode"
;;;	"vi-mode"
;;; 	"WordStar"
;;; ʤǤդǤ뤬
;;;
;;; ե̾
;;;	emacs-mode.{o,lsp}
;;;	vi-mode.{o,lsp}
;;;	WordStar.{o,lsp}
;;; إסե̾
;;; 	emacs-mode.help
;;; 	vi-mode.help
;;; 	WordStar.help
;;;
;;; ȴϢդ롣⡼̾硦ʸ϶̤롣
;;; 裲Ȥƥѥ̾ꤹȻꤷѥ
;;; եɤ߹ࡣꤵ줿ѥ̾ϵ
;;; ʹߡ꤬ʤΥǥեȤȤ
;;; Ȥ롣
;;;
(defun set-editor-mode (&optional mode-name load-path)
  (declare (string mode-name load-path))
  (when (null mode-name)
    (return-from set-editor-mode (values *editor-mode* *implicit-load-path*)))
  (setf *editor-mode* mode-name)
  (cond
   ((null load-path)
    (load (merge-pathnames *editor-mode* *implicit-load-path*)))
   (t (setf *implicit-load-path* load-path)
    (load (merge-pathnames *editor-mode* load-path)))))

;;;
;;; ߤΥǥ⡼֤̾(set-editor-mode)Ʊ
;;;
(defun editor-mode ()
  (values *editor-mode* *implicit-load-path*))

;;;
;;; ߤΥǥ⡼ѤΥإסեɽ롣
;;; 裲Ȥƥإסեɤѥ̾
;;; Ǥ롣ꤵʤä set-editor-modeǻꤵ
;;; ѥ̾Ȥ롣
;;;
;;;  help-editǻꤷ⡼ɤȥѥ̾ϡֵʤס
;;;
(defun help-edit (&optional (mode (editor-mode)) load-path)
  (let (fname)
    (cond
     ((null load-path)
      (setf fname (concatenate 'string *implicit-load-path* mode ".help")))
     (t (setf fname (concatenate 'string load-path mode ".help"))))
    (less fname)))

;;;
;;; ܡɤƤӽФ줿ݤˤäơ֤
;;; Ȥ򤹤٤֤ͤüʴؿδؿ
;;; Ѥ뤳ȤǷ֤Ǥ롣
;;;
;;;  1ˤϸƤӽФ¦ؿǷ֤
;;; Ϥ
;;;
(defun select-repeat-count (n)
  (let ((i nil))
    (if (call-by-keyboard-p) (setf i (repeat-count)))
    (setf *call-by-keyboard* nil)
    (if (null i) (setf i n))
    (return-from select-repeat-count i)))

;;;
;;; ⤷ʤؿ
;;;
(defun do-nothing () nil)

;;;
;;; Ͻλ֤ؿñʤޡʤΤΤס
;;;
(defun end-input () nil)

;;;
;;; ΥޥɤǤ뤿ʸɤɽꤹ롣
;;;
(defun set-break-code (&optional (code nil sw))
  (cond
   ((null sw) ctrl-g)
   ((characterp code)
    (setf ctrl-g code))
   (t (format nil "set-break-code: single character expected~%"))))

;;;
;;; debugɽ椹ؿ
;;;
(defun debug-level (&optional (level nil exist-p))
  (cond
   ((null exist-p) *debug-level*)
   (t (setf *debug-level* level))))

;;; M-i
;;; ܡɤϤ޷ʸʸɤɽ롣
;;;
;;; 	C-u M-i		޷ʸʸɤɽ롣
;;; 	C-u 0 M-i	޷ʸʸɤɽʤ
;;;
(defun inspect-keycode (&optional (count nil))
  (setf count (select-repeat-count count))
  (cond
   ((null count) *inspect-keycode*)
   ((zerop count) (setf *inspect-keycode* nil))
   (t (setf *inspect-keycode* t))))

;;;
;;; ݥȰ֤֤
;;; ޥȤ exportǤʤΤǴؿȤƤ롣
;;; ƤӽФСإåɤȤʤսǤ inlineŸꡣ
;;;
(defun point () *point*)

;;;
;;; ֤֤Ʊ塣ʲƱ¿
;;;
(defun cursor-position () *cursor-position*)

;;;
;;; ޡ֤֤
;;;
(defun mark () *mark*)

;;;
;;; ݥȤľʸ֤
;;;
(defun current-char ()
  (elt *text* *point*))
; (nth *point* *text*))

;;;
;;; ݥȤƥȤƬˤ뤫ݤ֤
;;;
(defun beginning-of-text-p ()
  (= *point* 0))

;;; M-a
;;; ݥȤƥȤƬ˰ư롣
;;;
(defun beginning-of-text ()
  (setf *point* 0))

;;;
;;; ݥȤƥȤνüˤ뤫ݤ֤
;;;
(defun end-of-text-p ()
  (= *point* *ncl*))

;;; M-e
;;; ݥȤƥȤνü˰ư롣
;;;
(defun end-of-text ()
  (setf *point* *ncl*))

;;;
;;; Ԥݤ֤
;;;
(defun empty-text-p ()
  (= *ncl* 0))

;;;
;;; ǽηֹ֤
;;;
(defun max-column ()
  (1- *line-size*))

;;;
;;; ٥Ĥ餹
;;;
(defun audio-bell ()
  (when (enable-audio-bell) (putc ctrl-g))
  (return-from audio-bell nil))

;;;
;;; ݥȤƬˤ뤫ݤ֤
;;;
(defun beginning-of-line-p ()
  (cond
   ((beginning-of-text-p) t)
   ((char= (elt *text* (1- *point*)) #\Newline) t)))
;  ((char= (nth (1- *point*) *text*) #\Newline) t)))

;;; C-a
;;; ݥȤꤵ줿ԿʬιԤƬ˰ư롣
;;;
(defun beginning-of-line (&optional (n 1))
  (declare (fixnum n))
  (setf n (select-repeat-count n))
  (cond
   ((zerop n) nil)
   ((> n 0)
    (dotimes (i n)
      (true-beginning-of-line)
      (if (< i (1- n)) (backward-char 1))))
   ((< n 0)
    (dotimes (i (- n))
      (true-end-of-line)
      (if (< i (1- (- n))) (forward-char 1))))))

(defun true-beginning-of-line ()
  (loop
    (if (beginning-of-line-p) (return))
    (backward-char 1)))

;;;
;;; ݥȤˤ뤫ݤ֤
;;;
(defun end-of-line-p ()
  (cond
   ((end-of-text-p) t)
   ((char= (current-char) #\Newline) t)))

;;; C-e
;;; ݥȤꤵ줿Կʬι˰ư롣
;;;
(defun end-of-line (&optional (n 1))
  (declare (fixnum n))
  (setf n (select-repeat-count n))
  (cond
   ((zerop n) nil)
   ((> n 0)
    (dotimes (i n)
      (true-end-of-line)
      (if (< i (1- n)) (forward-char 1))))
   ((< n 0)
    (dotimes (i (- n))
      (true-beginning-of-line)
      (if (< i (1- (- n))) (backward-char 1))))))

(defun true-end-of-line ()
  (loop
    (if (end-of-line-p) (return))
    (forward-char 1)))

;;;
;;; ƥΤʸ֤
;;;
(defun text-length () (length *text*))

;;;
;;; ʸ˥ѥåƥΤ֤
;;;
(defun packed-text () (pack *text*))

;;;
;;; ݥȤ¸ߤԤʸ֤
;;;
(defun line-length ()
  (let (p n)
    (setf p (previous-newline-position))
    (setf n (next-newline-position))
    (cond
     ((and (null p) (null n))
      (text-length))
     ((null p)
      (1+ n))
     ((null n)
      (1- (- (text-length) p)))
     (t (- n p)))))

;;;
;;; ݥȤ¸ߤΤʸ˥ѥå֤
;;;
(defun packed-line () (pack (current-line)))

;;;
;;; ݥȤ¸ߤΤ֤
;;;
(defun current-line ()
  (let (p n)
    (setf p (previous-newline-position))
    (setf n (next-newline-position))
    (cond
     ((and (null p) (null n)) *text*)
     ((null p)
      (subseq *text* 0 (1+ n)))
     ((null n)
      (subseq *text* (1+ p)))
     (t (subseq *text* (1+ p) (1+ n))))))

;;;
;;; ݥȤޤԤγϰ֤֤
;;;
(defun current-line-base ()
  (let (pos base)
    (setf pos (previous-newline-position))
    (if (null pos) (setf base 0) (setf base (1+ pos)))
    (return-from current-line-base base)))

;;;
;;; 򺸤 nư롣ݥȤѲʤ
;;; 뤬 0ۤƺưȤϤʤ
;;;
(defun move-cursor-left (n)
  (declare (fixnum n))
  (if (minusp n) (return-from move-cursor-left nil))
  (setf n (min n (cursor-position)))
  (pure-move-left n)
  (decf *cursor-position* n))

;;;
;;; 򱦤 nư롣ݥȤѲʤ
;;; 뤬(1- (line-size))ۤƱư
;;; Ϥʤ
;;;
(defun move-cursor-right (n)
  (declare (fixnum n))
  (if (minusp n) (return-from move-cursor-right nil))
  (setf n (min n (- (max-column) (cursor-position))))
  (pure-move-right n)
  (incf *cursor-position* n))

;;;
;;;  nܤ˰ư롣ݥȤѲʤ
;;; ưΥ֤֤
;;;
(defun move-cursor-to (n)
  (declare (fixnum n))
  (if (minusp n) (return-from move-cursor-to nil))
  (let (col)
    (setf col (- n (cursor-position)))
    (cond
     ((zerop col) (cursor-position))
     ((> col 0) (move-cursor-right col))
     ((< col 0) (move-cursor-left (- col))))))

;;;
;;; ݥȤ򺸤 nʸʬư롣
;;; ݥȤǤ˥ƥȤƬˤϲ⤻nil֤
;;; ưǤϡưΥݥȤ֤
;;;
(defun move-point-left (n)
  (declare (fixnum n))
  (if (minusp n) (return-from move-point-left nil))
  (if (beginning-of-text-p) (return-from move-point-left nil))
  (setf n (min n *point*))
  (decf *point* n))

;;;
;;; ݥȤ򱦤 nʸʬư롣
;;; ݥȤǤ˥ƥȤνüˤϲ⤻nil֤
;;; ưǤϡưΥݥȤ֤
;;;
(defun move-point-right (n)
  (declare (fixnum n))
  (if (minusp n) (return-from move-point-right nil))
  (if (end-of-text-p) (return-from move-point-right nil))
  (setf n (min n (- *ncl* *point*)))
  (incf *point* n))

;;;
;;; ݥȤǻꤵ줿֤˰ư롣
;;; ưΥݥȤ֤
;;;
(defun move-point-to (n)
  (declare (fixnum n))
  (if (minusp n) (return-from move-point-to nil))
  (let (col)
    (setf col (- n *point*))
    (cond
     ((zerop col) *point*)
     ((> col 0) (move-point-right col))
     ((< col 0) (move-point-left (- col))))))

;;; C-f
;;; ꤵ줿ʬݥȤʤ롣
;;; ưΥݥȤ֤
;;;
(defun forward-char (&optional (n 1))
  (declare (fixnum n))
  (setf n (select-repeat-count n))
  (cond
   ((zerop n) *point*)
   ((> n 0)
    (move-point-right n))
   ((< n 0)
    (move-point-left (- n)))))

;;; C-b
;;; ꤵ줿ʬݥȤ˿ʤ롣
;;; ưΥݥȤ֤
;;;
(defun backward-char (&optional (n 1))
  (declare (fixnum n))
  (setf n (select-repeat-count n))
  (cond
   ((zerop n) *point*)
   ((> n 0)
    (move-point-left n))
   ((< n 0)
    (move-point-right (- n)))))

;;; C-d
;;; ݥȰ֤ꤵ줿ʸ뤹롣
;;;
;;; ʸꤵ줿(1ʸǤ)ǤϤʤ뤹롣
;;; ꤵ줿ϥݥȰ֤ʸоݤȤʤ롣
;;;
;;; ʸκ뤬ԤʤäȤ nil֤Ǥ
;;; ȤϻĤʸ֤
;;;
(defun delete-char (&optional (n nil))
  (let (result)
    (setf n (select-repeat-count n))
    (cond
     ((empty-text-p) nil)
     ((end-of-text-p) nil)
     ((null n)
      (save-mark-info)
      (save-delete-undo-info (current-char))
      (setf result (true-delete-char))
      (align-mark)
      (return-from delete-char result))
     ((zerop n) nil)
     ((< n 0)
      (delete-backward-char (- n)))
     ((> n 0)
      (save-mark-info)
      (set-mark-command)
      (forward-char n)
      (exchange-point-and-mark)
      (kill-region)
      (align-mark)
      (return-from delete-char *ncl*)))))

;;; C-h (BS)
;;; ݥȰ֤ꤵ줿ʸո˺뤹롣
;;;
;;; ʸꤵ줿(1ʸǤ)ǤϤʤ뤹롣
;;; ꤵ줿ϥݥȰ֤ʸоݤȤʤ롣
;;;
;;; ʸκ뤬ԤʤäȤ nil֤Ǥ
;;; ȤϻĤʸ֤
;;;
(defun delete-backward-char (&optional (n nil))
  (let (result)
    (setf n (select-repeat-count n))
    (cond
     ((beginning-of-text-p) nil)
     ((null n)
      (save-mark-info)
      (backward-char 1)
      (save-backward-delete-undo-info (current-char))
      (setf result (true-delete-char))
      (align-mark)
      (return-from delete-backward-char result))
     ((zerop n) nil)
     ((< n 0)
      (delete-char (- n)))
     ((> n 0)
      (save-mark-info)
      (set-mark-command)
      (backward-char n)
      (exchange-point-and-mark)
      (setf result (kill-region))
      (pop-kill-ring)
      (align-mark)
      (return-from delete-backward-char result)))))

;;;
;;; ݥȰ֤1ʸ롣
;;;
;;; In GCL 2.3.8-Beta on Pentium-III 700MHz/Linux 2.4.0-test8 (Interpreted).
;;;
;;; > (setq x (unpack "(compile-file \"line-edit-pkg.lsp\")"))
;;;
;;; > (time (dotimes (i 10000) (remove-if #'identity x :start 1 :count 1)))
;;; real time : 0.060 secs
;;; run time  : 0.050 secs
;;;
;;; > (time (dotimes (i 10000) (append (subseq x 0 1) (cdr (subseq x 1)))))
;;; real time : 0.170 secs
;;; run time  : 0.170 secs
;;;
(defun true-delete-char ()
  (if (or (empty-text-p) (end-of-text-p)) (return-from true-delete-char nil))
  (setf *text* (remove-if #'identity *text* :start *point* :count 1))
  (need-refresh-line t)
  (decf *ncl*)
  (return-from true-delete-char *ncl*))

;;; M-\
;;; ݥȤμζʸ򤹤٤ƺ롣
;;;
(defun delete-horizontal-space ()
  (save-mark-info)
  (backward-kill-white-space)
  (kill-white-space)
  (align-mark))

;;; M-s
;;; ݥȤμζʸο˴طʤ
;;; (ȤĤǤ)ݥȤľ˶1ĤĤ
;;;
(defun just-one-space ()
  (save-mark-info)
  (delete-horizontal-space)
  (self-insert #\Space 1)
  (align-mark))

;;; C-k
;;; ݥȰ֤ޤǤʸ򥭥뤹롣
;;; ݥȰ֤֤
;;;
;;; 0ޤꤵ줿ϥݥȰ֤ΥƥȤ򥭥뤹롣
;;;
(defun kill-line (&optional (n nil))
  (setf n (select-repeat-count n))
  (save-mark-info)
  (set-mark-command)
  (cond
   ((null n)
    (if (end-of-line-p) (forward-char 1)
      (end-of-line 1)))
   ((zerop n)
    (beginning-of-line 1))
   ((< n 0)
    (beginning-of-line (- n)))
   ((> n 0)
    (end-of-line n)))
  (exchange-point-and-mark)
  (kill-region)
  (align-mark)
  (return-from kill-line *ncl*))

;;; M-f
;;; ñľ˥ݥȤư롣ñƬؤΰưǤϤʤ
;;; ݥȤGNU EmacsפȤƥȤƬ=GľˤˤȤ
;;; forward-wordȥݥȤ'U'ľ˰ư롣
;;;
;;; ưʤäȤ nil֤ǤʤȤϰưΥݥȤ
;;; ֤
;;;
(defun forward-word (&optional (n 1) &aux result)
  (declare (fixnum n))
  (setf n (select-repeat-count n))
  (cond
   ((zerop n) nil)
   ((end-of-text-p) (audio-bell))
   ((> n 0) (dotimes (i n result) (setf result (true-forward-word))))
   (t (backward-word (- n))))
  (return-from forward-word n))

(defun true-forward-word ()
  (loop
    (if (end-of-text-p) (return-from true-forward-word *point*))
    (if (normal-char-p (current-char)) (return nil))
    (forward-char 1))
  (loop
    (if (end-of-text-p) (return-from true-forward-word *point*))
    (if (not (normal-char-p (current-char)))
	(return-from true-forward-word *point*))
    (forward-char 1)))

;;; M-n
;;; ñƬ˥ݥȤư롣ݼŪʰưޥɤ˴Ƥ
;;; 桼ѤƤ롣Ϳ줿ȤľñƬ
;;; ΰưȤʤ롣
;;;
(defun next-word (&optional (n 1))
  (declare (fixnum n))
  (let (result)
    (setf n (select-repeat-count n))
    (cond
     ((zerop n) nil)
     ((end-of-text-p) (audio-bell))
     ((> n 0) (dotimes (i n result) (setf result (true-next-word))))
     (t (backward-word (- n))))))

(defun true-next-word ()
  (loop
    (if (end-of-text-p) (return-from true-next-word nil))
    (if (not (normal-char-p (current-char))) (return))
    (forward-char 1))
  (loop
    (if (end-of-text-p) (return-from true-next-word nil))
    (if (normal-char-p (current-char)) (return-from true-next-word *point*))
    (forward-char 1)))

;;;
;;; ʸɤФʸ *white-space-char* ǤȤ
;;; Ƥ롣
;;;
;;; ưʤäȤ nil֤ǤʤȤϰưΥݥȤ
;;; ֤
;;;
(defun skip-white-space ()
  (let (moved)
    (setf moved nil)
    (loop
      (if (end-of-text-p) (return-from skip-white-space moved))
      (if (not (white-space-p (current-char)))
	  (return-from skip-white-space moved))
      (forward-char 1)
      (setf moved *point*))))

;;;
;;; *break-char*Ǥ⡢*white-space-char*ǤʤʸޤǥݥȤư롣
;;; ˥ݥȰ֤֤
;;;
(defun skip-to-normal-char ()
  (loop
    (if (end-of-text-p) (return-from skip-to-normal-char *point*))
    (if (normal-char-p (current-char)) (return-from skip-to-normal-char *point*))
    (forward-char 1)))

;;; M-b
;;; ˣñʬư롣ݥȤñľ˰ư롣
;;;
;;; ưʤäȤ nil֤ǤʤȤϰưΥݥȤ֤
;;;
(defun backward-word (&optional (n 1) &aux result)
  (declare (fixnum n))
  (setf n (select-repeat-count n))
  (cond
   ((zerop n) nil)
   ((beginning-of-text-p) (audio-bell))
   ((> n 0)
    (dotimes (i n result) (setf result (true-backward-word))))
   (t (forward-word (- n)))))

(defun true-backward-word ()
  (loop
    (when (beginning-of-text-p) (return-from true-backward-word *point*))
    (backward-char 1)
    (when (normal-char-p (current-char)) (return)))
  (loop
    (if (beginning-of-text-p) (return-from true-backward-word *point*))
    (backward-char 1)
    (when (not (normal-char-p (current-char)))
      (forward-char 1)
      (return-from true-backward-word *point*))))

;;;
;;; ո˶ʸɤФʸ *white-space-char*
;;; ǤȤƤ롣
;;;
;;; ưʤäȤ nil֤ǤʤȤϰưΥݥȤ֤
;;;
(defun backward-skip-white-space ()
  (let (pos)
    (setf pos *point*)
    (loop
      (if (beginning-of-text-p) (return))
      (backward-char 1)
      (when (not (white-space-p (current-char)))
	(forward-char 1)
	(return)))
    (if (/= pos *point*) *point* nil)))

;;;
;;; *break-char*Ǥ⡢*white-space-char*Ǥʤʸޤǵո
;;; ݥȤư롣˥ݥȰ֤֤
;;;
(defun backward-skip-to-normal-char ()
  (loop
    (if (beginning-of-text-p) (return *point*))
    (if (normal-char-p (current-char)) (return *point*))
    (backward-char 1)))

;;; M-d
;;; ꤵ줿Ŀʬñ򥭥뤹롣
;;; ꤵ줿Ȥϸ˥뤹롣
;;;
;;; 뤬ԤʤäȤ nil֤ǤʤȤϥݥȰ֤֤
;;;
(defun kill-word (&optional (n 1) &aux result)
  (declare (fixnum n))
  (setf n (select-repeat-count n))
  (cond
   ((zerop n) nil)
   ((> n 0)
    (dotimes (i n result) (setf result (true-kill-word))))
   (t (backward-kill-word (- n)))))

;;;
;;; ñޤǤ˥뤹롣
;;;
;;; 뤬ԤʤäȤ nil֤ǤʤȤϥݥȰ֤֤
;;;
(defun true-kill-word ()
  (let (result)
    (save-mark-info)
    (set-mark-command)
    (forward-word 1)
    (exchange-point-and-mark)		;force forward kill.
    (setf result (kill-region))
    (align-mark)
    (return-from true-kill-word result)))

;;;
;;; ʸ򥭥뤹롣
;;; 뤬ԤʤäȤ nil֤ǤʤȤϥݥȰ֤֤
;;;
(defun kill-white-space ()
  (let (result)
    (save-mark-info)
    (set-mark-command)
    (skip-white-space)
    (exchange-point-and-mark)		;force forward kill.
    (setf result (kill-region))
    (align-mark)
    (return-from kill-white-space result)))

;;; M-C-h
;;; ꤵ줿Ŀʬñ˥뤹롣
;;; ꤵ줿Ȥ˥뤹롣
;;;
;;; 뤬ԤʤäȤ nil֤ǤʤȤϥݥȰ֤֤
;;;
(defun backward-kill-word (&optional (n 1) &aux result)
  (declare (fixnum n))
  (setf n (select-repeat-count n))
  (cond
   ((zerop n) nil)
   ((> n 0)
    (dotimes (i n result) (setf result (true-backward-kill-word))))
   (t (kill-word (- n)))))

;;;
;;; ñƬޤǤ˥뤹롣
;;;
;;; 뤬ԤʤäȤ nil֤ǤʤȤϥݥȰ֤֤
;;;
(defun true-backward-kill-word ()
  (let (result)
    (save-mark-info)
    (set-mark-command)
    (backward-word 1)
    (exchange-point-and-mark)		;force backward kill.
    (setf result (kill-region))
    (align-mark)
    (return-from true-backward-kill-word result)))

;;;
;;; ո˶ʸ򥭥뤹롣
;;;
;;; 뤬ԤʤäȤ nil֤ǤʤȤϥݥȰ֤֤
;;;
(defun backward-kill-white-space ()
  (let (result)
    (save-mark-info)
    (set-mark-command)
    (backward-skip-white-space)
    (exchange-point-and-mark)		;force backward kill.
    (setf result (kill-region))
    (align-mark)
    (return-from backward-kill-white-space result)))

;;; M-@
;;; ݥȤưˡꤵ줿ʬ M-fˤäưư֤
;;; ޡꤹ롣ꤵ줿Ȥ M-bˤäưư
;;; ˥ޡꤹ롣ޡ֤֤
;;;
(defun mark-word (&optional (n 1))
  (declare (fixnum n))
  (setf n (select-repeat-count n))
  (set-mark-command)
  (cond
   ((zerop n) nil)
   ((> n 0) (forward-word n))
   (t (backward-word (- n))))
  (exchange-point-and-mark))

;;;
;;; kill-ringؤ¸κͤꤹ롣ꤷʤȸߤ
;;; ¸֤¸Ͻͤ 30ȤƤ뤬ץ
;;; ξ¤Ϥʤ
;;;
;;; kill-ringƥȤߤ륹åȹͤȡδؿ
;;; kill-ringåʿ꡿ؿǤ롣
;;;
(defun kill-ring-max (&optional n)
  (cond
   ((null n) *kill-ring-max*)
   (t (setf *kill-ring-max* n))))

;;;
;;; kill-ringƬإƥȤɲä롣
;;;
(defun push-kill-ring (lst)
  (declare (list lst))
  (push lst *kill-ring*))

;;;
;;; kill-ringƬƥȤФkill-ringƬΤƤ롣
;;;
(defun pop-kill-ring ()
  (pop *kill-ring*))

;;; C-y
;;; kill-ringƥȤ󥯤롣 (nȤ)ꤷ
;;; ǽ󥯡ݥ󥿤 nܤΥƥȤ󥯤롣
;;;
;;; 󥯤ƥȤƬ˥ޡꤵ졢ݥȤϥ󥯤ƥ
;;; Ȥľ˰ư롣
;;;
;;; 󥯤줿ȤϥݥȰ֤֤롣ǤʤȤ nil֤롣
;;;
(defun yank (&optional (n 1) &aux result)
  (declare (fixnum n))
  (setf n (select-repeat-count n))
  (cond
   ((zerop n) nil)
   ((= n 1)
    (true-yank))
   ((> n 1)
    (true-yank)
    (dotimes (i (1- n) result) (if (true-yank-pop) (setf result *point*))))))

;;;
;;; Ǹ˥뤷ƥȤ󥯡kill-ringƬΥƥȤ򡢥ݥȰ
;;; ˤ롣
;;;
(defun true-yank ()
  (let (result)
    (setf *last-yank* 0)
    (setf *allow-yank-pop* t)
    (set-mark-command)
    (setf result (insert (car *kill-ring*)))
    ; check and cut *kill-ring*, if its length is over the (kill-ring-max)
    (if (> (length *kill-ring*) (kill-ring-max))
	(setf *kill-ring*
	      (butlast *kill-ring* (- (length *kill-ring*) (kill-ring-max)))))
    (return-from true-yank result)))

;;; M-y
;;; ľ˥󥯤ƥȤ򡢤˥뤷ƥȤ֤롣
;;;
;;; 󥯤ƥȤƬ˥ޡꤵ졢ݥȤϥ󥯤ƥ
;;; Ȥľ˰ư롣
;;;
;;; 󥯤줿ȤϥݥȰ֤֤롣ǤʤȤ nil֤롣
;;;
(defun yank-pop (&optional (n 1) &aux result)
  (declare (fixnum n))
  (setf n (select-repeat-count n))
  (cond
   ((zerop n) nil)
   ((> n 0)
    (dotimes (i n result) (if (true-yank-pop) (setf result *point*))))))

;;;
;;; ľ˥󥯤ƥȤ򡢤˥뤷ƥȤ֤롣
;;; kill-ring򥹥åȹͤȡ󥯤ƥȤ¹ԤΤӤ
;;; kill-ringåΣܡܡƤ֤
;;; 롣kill-ringκǸãȡƤkill-ringκǽ롣
;;;
(defun true-yank-pop ()
  (let (result)
    (when (not *allow-yank-pop*) (return-from true-yank-pop nil))
    (when (> *point* *mark*) (exchange-point-and-mark))
    (when (equal
	   (car (nthcdr *last-yank* *kill-ring*))
	   (subseq *text* *point* *mark*))
      (incf *last-yank*)
      (setf *last-yank* (mod *last-yank* (length *kill-ring*)))
      (kill-ring-suspend t)
      (kill-region)
      (pop-kill-ring)
      (set-mark-command)
      (setf result (insert (car (nthcdr *last-yank* *kill-ring*)))))
    (return-from true-yank-pop result)))

;;; C-@
;;; ߤΥݥȰ֤꡼ƬޤȤƵ롣
;;;
(defun set-mark-command ()
  (setf *mark* *point*))

;;; C-x C-x
;;; ߤΥݥȰ֤ȥޡ֤ؤ롣
;;; ޡ¸ߤƤʤäϡ⤻ nil֤
;;; ǤʤȤؤäޡ֤֤
;;;
(defun exchange-point-and-mark ()
  (let (pos)
    (when (null *mark*)
      (audio-bell)
      (return-from exchange-point-and-mark nil))
    (setf pos *point*)
    (move-point-to *mark*)
    (setf *mark* pos)))

;;; M-w
;;; ݥȤȥޡǰϤޤ줿꡼򥭥롦󥰤˥ԡ
;;; 롣ݥȤȥޡѲʤ
;;;
;;; ºݤ˥꡼򥳥ԡnilǤʤ֤ͤǤʤ
;;; nil֤
;;;
;;; ˥뤹(backward-kill-commands nil)
;;; ˥뤹(backward-kill-commands t)
;;; ¹ԤƵϿĤ
;;;
(defun kill-ring-save ()
  (cond
   ((empty-text-p) nil)
   ((null *mark*)
    (audio-bell))
   ((= *mark* *point*) nil)
   ((< *point* *mark*)
    (backward-kill-commands nil)
    (append-to-kill-ring (subseq *text* *point* *mark*)))
   (t (backward-kill-commands t)
      (append-to-kill-ring (subseq *text* *mark* *point*)))))

;;; C-w
;;; ݥȰ֤ޡ֤ޤǤΥƥȤ򥭥뤹ʥåȡˡ
;;; ꡼󤬰Ǥ⥭뤵줿ȤϡΥݥȰ֤֤
;;; ǤʤȤ nil֤
;;;
(defun kill-region ()
  (let (str)
    (when (kill-ring-save)
      (if (> *point* *mark*) (exchange-point-and-mark))
      (setf str (copy-seq (subseq *text* *point* *mark*)))
      (save-mark-info)
      (dotimes (i (- *mark* *point*) *point*) (true-delete-char))
      (align-mark)
      (if (backward-kill-commands)
	  (save-backward-kill-undo-info str)
	(save-kill-undo-info str)))))

;;;
;;; kill-ringإƥȤѡޤϿܤȤ¸롣
;;;
(defun append-to-kill-ring (lst)
  (if (characterp lst) (setf lst (list lst)))
  (cond
   ((kill-ring-suspend)
    (setf *suspend-switch* nil)
    (setf *kill-ring* (cons lst *kill-ring*)))
   ((backward-kill-commands)
    (setf *kill-ring*
	  (cons (append lst (car *kill-ring*)) (cdr *kill-ring*))))
   (t (setf *kill-ring*
	    (cons (append (car *kill-ring*) lst) (cdr *kill-ring*))))))

;;;
;;; ˥뤹륳ޥɤǤ nil򡢸˥뤹륳ޥ
;;; Ǥ tϿ롣ꤷʤȡλǤ֤
;;;
(defun backward-kill-commands (&optional (flag nil sw))
  (cond
   ((null sw) *backward-kill-commands*)
   (t (setf *backward-kill-commands* flag))))

;;;
;;; kill-ringƬؤΥƥѤβݤ椹롣
;;;
(defun kill-ring-suspend (&optional (set-to nil sw))
  (cond
   ((null sw) *suspend-switch*)
   ((null set-to) nil)
   (t (setf *suspend-switch* t))))

;;; M-C-w
;;; ľΥ򡢺Ǹ˥뤷ƥȤ(ɬ)դä롣
;;;
(defun append-next-kill ()
  (setf *suspend-switch* nil))

;;; M-c
;;; ꤵ줿Ŀñ1ʸܤʸѴĤ
;;; ʸѴ롣
;;;
(defun capitalize-word (&optional (n 1))
  (declare (fixnum n))
  (let (pos)
    (setf n (select-repeat-count n))
    (cond
     ((zerop n) nil)
     ((> n 0)
      (dotimes (i n *point*) (true-capitalize-word)))
     ((< n 0)
      (setf pos *point*)
      (self-insert #\Space 1)		;this is a trick.
      (backward-word 1)
      (true-capitalize-word)
      (delete-char nil)			;cancel trick here.
      (backward-word 1)
      (dotimes (i (- (- n) 1))
	(backward-word 1)
	(true-capitalize-word)
	(backward-word 1))
      (move-point-to pos)))))

;;;
;;; ³ñ1ʸܤʸѴĤʸѴ롣
;;; ݥȤñˤϡݥȤθˤʬ
;;; Ѵ롣
;;;
(defun true-capitalize-word ()
  (let (word)
    (if (empty-text-p) (return-from true-capitalize-word nil))
    (loop
      (when (end-of-text-p) (return-from true-capitalize-word nil))
      (when (normal-char-p (current-char)) (return))
      (forward-char 1))
    (kill-ring-suspend t)
    (kill-word 1)
    (setf word (pop-kill-ring))
    (setf word (unpack (string-capitalize (pack word))))
    (insert word)))

;;; M-u
;;; ꤵ줿ĿñʸѴ롣
;;; ꤵ줿Ȥϸ˻Ĥñʸ롣
;;;
(defun upcase-word (&optional (n 1))
  (declare (fixnum n))
  (let (pos)
    (setf n (select-repeat-count n))
    (cond
     ((zerop n) nil)
     ((> n 0)
      (dotimes (i n *point*) (true-upcase-word)))
     ((< n 0)
      (setf pos *point*)
      (self-insert #\Space 1)		;this is a trick.
      (backward-word 1)
      (true-upcase-word)
      (delete-char nil)			;cancel trick here.
      (backward-word 1)
      (dotimes (i (- (- n) 1))
	(backward-word 1)
	(true-upcase-word)
	(backward-word 1))
      (move-point-to pos)))))

;;;
;;; ³ñʸѴ롣
;;; ݥȤñˤϡݥȤθˤʬ
;;; Ѵ롣
;;;
(defun true-upcase-word ()
  (let (word)
    (if (empty-text-p) (return-from true-upcase-word nil))
    (save-mark-info)
    (set-mark-command)
    (forward-word 1)
    (exchange-point-and-mark)
    (kill-ring-suspend t)
    (kill-word 1)
    (setf word (pop-kill-ring))
    (setf word (unpack (string-upcase (pack word))))
    (insert word)
    (align-mark)
    (return-from true-upcase-word *point*)))

;;; M-l
;;; ꤵ줿ĿñʸѴ롣
;;; ꤵ줿Ȥϸ˻Ĥñʸ롣
;;;
(defun downcase-word (&optional (n 1))
  (declare (fixnum n))
  (let (pos)
    (setf n (select-repeat-count n))
    (cond
     ((zerop n) nil)
     ((> n 0)
      (dotimes (i n *point*) (true-downcase-word)))
     ((< n 0)
      (setf pos *point*)
      (self-insert #\Space 1)		;this is a trick.
      (backward-word 1)
      (true-downcase-word)
      (delete-char nil)		;cancel trick here.
      (backward-word 1)
      (dotimes (i (- (- n) 1))
	(backward-word 1)
	(true-downcase-word)
	(backward-word 1))
      (move-point-to pos)))))
;;;
;;; ³ñʸѴ롣
;;; ݥȤñˤϡݥȤθˤʬ
;;; Ѵ롣
;;;
(defun true-downcase-word ()
  (let (word)
    (if (empty-text-p) (return-from true-downcase-word nil))
    (save-mark-info)
    (set-mark-command)
    (forward-word 1)
    (exchange-point-and-mark)
    (kill-ring-suspend t)
    (kill-word 1)
    (setf word (pop-kill-ring))
    (setf word (unpack (string-downcase (pack word))))
    (insert word)
    (align-mark)
    (return-from true-downcase-word *point*)))

;;; M-z
;;; ݥȰ֤Ϥ줿ʸܤ˸֤ޤǤ
;;; ʸޤƥ뤹롣Ǥʤ nil֤Ǥ
;;; ȤϥݥȰ֤֤
;;;
(defun zap-to-char (&optional (n 1) &aux result)
  (declare (fixnum n))
  (let (ch)
    (setf n (select-repeat-count n))
    (setf ch (getchar))
    (cond
     ((zerop n) nil)
     ((> n 0)
      (dotimes (i n result) (setf result (true-zap-to-char ch))))
     (t
      (dotimes (i (- n) result) (setf result (true-backward-zap-to-char ch)))))))

(defun true-zap-to-char (ch)
  (declare (character ch))
  (let (rest-line pos result)
    (declare (list rest-line))
    (setf rest-line (subseq *text* *point*))
    (setf pos (position ch rest-line))
    (when (null pos)
      (audio-bell)
      (return-from true-zap-to-char nil))
    (save-mark-info)
    (set-mark-command)
    (forward-char (1+ pos))
    (exchange-point-and-mark)	;force forward kill.
    (setf result (kill-region))
    (align-mark)
    (return-from true-zap-to-char result)))

(defun true-backward-zap-to-char (ch)
  (declare (character ch))
  (let (prev-line pos result)
    (declare (list prev-line))
    (setf prev-line (subseq *text* 0 *point*))
    (setf pos (position ch (reverse prev-line)))
    (when (null pos)
      (audio-bell)
      (return-from true-backward-zap-to-char nil))
    (save-mark-info)
    (set-mark-command)
    (backward-char (1+ pos))
    (exchange-point-and-mark)	;force backward kill.
    (setf result (kill-region))
    (align-mark)
    (return-from true-backward-zap-to-char result)))

;;; C-t
;;; ݥȤʸؤ롣
;;;
;;; 012354 ȤƥȤǥݥȤ4ľʥ뤬4ΰ֡ˤ
;;; Ȥ transpose-charԤ 012345 ֤ݥȤ
;;; 5ľ˰ư롣ưΥݥȰ֤֤
;;;
;;;  0ΤȤϥݥȤľʸȥޡľʸ
;;; ؤ롣ݥȰ֤ȥޡ֤ؤ롣ޡ
;;; ƤʤäȤϲ⤻ nil֤
;;;
(defun transpose-chars (&optional (n 1))
  (declare (fixnum n))
  (setf n (select-repeat-count n))
  (cond
   ((> n 0)
    (dotimes (i n *point*) (true-transpose-chars)))
   ((< n 0)
    (dotimes (i (- n) *point*)
      (backward-char 1)
      (true-transpose-chars)
      (backward-char 1)))
   ((zerop n)
    (cond
     ((null *mark*) (audio-bell))
     (t
      (kill-ring-suspend t)
      (delete-char 1)			;KILL one character.
      (exchange-point-and-mark)
      (yank 1) (pop-kill-ring)
      (kill-ring-suspend t)
      (delete-char 1)			;KILL one character.
      (exchange-point-and-mark)
      (yank 1) (pop-kill-ring)
      (backward-char 1)
      (exchange-point-and-mark)
      (backward-char 1)
      (return-from transpose-chars *point*))))))

(defun true-transpose-chars ()
  (let (ch)
    (when (beginning-of-text-p)
      (audio-bell)
      (return-from true-transpose-chars nil))
    (if (end-of-text-p) (backward-char 1))
    (setf ch (current-char))
    (delete-char nil)
    (backward-char 1)
    (self-insert ch 1)
    (forward-char 1)))

;;; M-t
;;; ݥȤñؤ롣
;;;
;;;  0ΤȤϥݥȤľñȥޡľñ
;;; ؤ롣ݥȰ֤ȥޡ֤ؤ롣ޡ
;;; ƤʤäȤϲ⤻ nil֤
;;;
(defun transpose-words (&optional (n 1))
  (declare (fixnum n))
  (let (pos pos2)
    (setf n (select-repeat-count n))
    (cond
     ((> n 0)
      (dotimes (i n *point*) (true-transpose-words)))
     ((< n 0)
      (dotimes (i (- n) *point*)
	(backward-word 1)
	(true-transpose-words)
	(backward-word 1)))
     ((zerop n)
      (cond
       ((null *mark*) (audio-bell))
       (*mark*
	(kill-ring-suspend t)
	(setf pos *point*)
	(kill-word 1)
	(exchange-point-and-mark)
	(yank 1) (pop-kill-ring)
	(kill-ring-suspend t)
	(setf pos2 *point*)
	(kill-word 1)
	(move-point-to pos)
	(yank 1) (pop-kill-ring)
	(move-point-to pos)
	(set-mark-command)
	(move-point-to pos2)
	(return-from transpose-words *point*)))))))


(defun true-transpose-words ()
  (let (pos)
    (if (or (empty-text-p) (beginning-of-text-p) (end-of-text-p))
	(return-from true-transpose-words nil))
    (save-mark-info)
    (set-mark-command)
    ;;Already in a word ?
    (when (in-word-p)
      (forward-word 1)
      (set-mark-command))
    ;; Is there word backward ?
    (backward-word 1)
    (when (not (normal-char-p (current-char)))
      (exchange-point-and-mark)
      (align-mark)
      (audio-bell)
      (return-from true-transpose-words nil))
    ;; Is there word forward ?
    (exchange-point-and-mark)
    (set-mark-command)
    (loop
      (if (end-of-text-p) (return))
      (if (normal-char-p (current-char)) (return))
      (forward-char 1))
    (when (end-of-text-p) 
      (exchange-point-and-mark)
      (align-mark)
      (audio-bell)
      (return-from true-transpose-words nil))
    ;; Main block
    (exchange-point-and-mark)
    (backward-word 1)
    (kill-ring-suspend t)
    (kill-word 1)
    (setf pos *point*)
    (skip-to-normal-char)
    (yank 1) (pop-kill-ring)
    (kill-ring-suspend t)
    (kill-word 1)
    (move-point-to pos)
    (yank 1) (pop-kill-ring)
    (forward-word 1)
    (align-mark)))

;;;
;;; ݥȤñˤ뤫ݤĴ٤롣
;;;
(defun in-word-p ()
  (let (ch)
    (cond
     ((empty-text-p) nil)
     ((beginning-of-text-p)
      (normal-char-p (current-char)))
     ((end-of-text-p) nil)
     ((not (normal-char-p (current-char))) nil)
     (t (backward-char 1)
	(setf ch (current-char))
	(forward-char 1)
	(normal-char-p ch)))))

;;; M-C-f
;;; Ӽñ̤Ȥƻʬ˰ư롣
;;; ξϻʬ˰ư롣
;;;
;;; ʬưǤȤϥݥȰ֤֤
;;; ǤʤȤ nil֤
;;;
(defun forward-sexp (&optional (n 1) &aux result)
  (declare (fixnum n))
  (setf n (select-repeat-count n))
  (cond
   ((zerop n) nil)
   ((> n 0)
    (dotimes (i n result) (setf result (true-forward-sexp))))
   (t (backward-sexp (- n)))))

;;;
;;; ݥȤ LispΣӼҤȤʬ˰ư롣
;;; ưоݤΣӼʤХݥȤϰưnil֤
;;; ưǤХݥȰ֤֤
;;;
(defun true-forward-sexp ()
  (let (depth in-string pos)
    (setf depth 0)
    (setf pos *point*)
    (setf in-string (in-string-or-literal-p))
    (when in-string
      (move-point-to in-string)
      (skip-string-or-literal)
      (return-from true-forward-sexp *point*))
    (loop
      (skip-white-space)
      (cond
       ((end-of-text-p)
	(move-point-to pos)
	(return-from true-forward-sexp nil))
       ((member (current-char) *left-parenthesis-list* :test #'char=)
	(decf depth)
	(forward-char 1))
       ((member (current-char) *right-parenthesis-list* :test #'char=)
	(incf depth)
	(forward-char 1)
	(if (= depth 0) (return-from true-forward-sexp *point*)))
       ((char= (current-char) #\#)
	(forward-char 1)
	(cond
	 ((member (current-char) '(#\\ #\'))
	  (forward-char 1)
	  (forward-word 1))
	 (t (forward-word 1)))
	(if (= depth 0) (return-from true-forward-sexp *point*)))
       ((member (current-char) '(#\' #\` #\, #\@))
	(forward-char 1))
       ((member (current-char) '(#\" #\|))
	(skip-string-or-literal)
	(if (= depth 0) (return-from true-forward-sexp *point*)))
       ((char= (current-char) #\\)
	(forward-char 2)
	(if (= depth 0) (return-from true-forward-sexp *point*)))
       ((normal-char-p (current-char))
	(forward-word 1)
	(if (= depth 0) (return-from true-forward-sexp *point*)))
       (t (forward-char 1))))))

;;; M-C-b
;;; Ӽñ̤Ȥƻʬ˰ư롣
;;; ξϡʬ˰ư롣
;;;
;;; ʬưǤȤϥݥȰ֤֤
;;; ǤʤȤ nil֤
;;;
(defun backward-sexp (&optional (n 1) &aux result)
  (declare (fixnum n))
  (setf n (select-repeat-count n))
  (cond
   ((zerop n) nil)
   ((> n 0)
    (dotimes (i n result) (setf result (true-backward-sexp))))
   (t (forward-sexp (- n)))))

;;;
;;; ݥȤLispΣӼҤȤʬ˰ư롣
;;; ưоݤΣӼʤХݥȤϰưnil֤
;;; ưǤХݥȰ֤֤
;;;
(defun true-backward-sexp ()
  (let (line pos len result)
    (setf pos *point*)
    (setf line (copy-seq *text*))
    (setf len (length *text*))
    (setf *text* (invert-list line))
    (move-point-to (- len pos))
    (setf result (forward-sexp 1))
    (setf pos (- len *point*))
    (setf *text* line)
    (move-point-to pos)
    (return-from true-backward-sexp result)))

;;; M-C-d
;;; ˻ʬΤä٥˰ư롣
;;;
;;; ʬưǤȤϥݥȰ֤֤
;;; ǤʤȤ nil֤
;;;
(defun down-list (&optional (n 1) &aux result)
  (declare (fixnum n))
  (setf n (select-repeat-count n))
  (cond
   ((zerop n) nil)
   ((> n 0)
    (dotimes (i n result) (setf result (true-down-list))))
   (t (backward-down-list (- n)))))

;;;
;;; ˤҤȤĲΤä٥˰ư롣
;;; ưоݤΣӼʤХݥȤϰưnil֤
;;; ưǤХݥȰ֤֤
;;;
(defun true-down-list ()
  (let (in-string pos)
    (setf pos *point*)
    (setf in-string (in-string-or-literal-p))
    (when in-string
      (move-point-to in-string)
      (skip-string-or-literal))
    (loop
      (skip-white-space)
      (cond
       ((end-of-text-p)
	(move-point-to pos)
	(return-from true-down-list nil))
       ((member (current-char) *left-parenthesis-list* :test #'char=)
	(forward-char 1)
	(return-from true-down-list *point*))
       ((member (current-char) *right-parenthesis-list* :test #'char=)
	(move-point-to pos)
	(return-from true-down-list nil))
       ((member (current-char) '(#\" #\|))
	(skip-string-or-literal))
       ((char= (current-char) #\\)
	(forward-char 2))
       ((normal-char-p (current-char))
	(forward-word 1))
       (t (forward-char 1))))))

;;;
;;; ˻ʬΤä٥˰ư롣
;;;
;;; ʬưǤȤϥݥȰ֤֤
;;; ǤʤȤ nil֤
;;;
(defun backward-down-list (&optional n &aux result)
  (if (null n) (setf n (repeat-count)))
  (if (null n) (setf n 1))
  (cond
   ((zerop n) nil)
   ((> n 0)
    (dotimes (i n result) (setf result (true-backward-down-list))))
   (t (down-list (- n)))))

;;;
;;; 1ĲΤä٥˰ư롣
;;;
;;; ʬưǤȤϥݥȰ֤֤
;;; ǤʤȤ nil֤
;;;
(defun true-backward-down-list ()
  (let (line pos len result)
    (setf pos *point*)
    (setf line (copy-seq *text*))
    (setf len (length *text*))
    (setf *text* (invert-list line))
    (move-point-to (- len pos))
    (setf result (down-list 1))
    (setf pos (- len *point*))
    (setf *text* line)
    (move-point-to pos)
    (return-from true-backward-down-list result)))

;;; C-x u (for debug binding)
;;; ˻ʬΤä٥˰ư롣
;;;
;;; ʬưǤȤϥݥȰ֤֤
;;; ǤʤȤ nil֤
;;;
(defun up-list (&optional (n 1) &aux result)
  (declare (fixnum n))
  (setf n (select-repeat-count n))
  (cond
   ((zerop n) nil)
   ((> n 0)
    (dotimes (i n result) (setf result (true-up-list))))
   (t (backward-up-list (- n)))))

;;;
;;; ˤҤȤľΤä٥˰ư롣
;;; ưоݤΣӼʤХݥȤϰưnil֤
;;; ưǤХݥȰ֤֤
;;;
(defun true-up-list ()
  (let (in-string pos level)
    (setf level 0)
    (setf pos *point*)
    (setf in-string (in-string-or-literal-p))
    (when in-string
      (move-point-to in-string)
      (skip-string-or-literal))
    (loop
      (skip-white-space)
      (cond
       ((end-of-text-p)
	(move-point-to pos)
	(return-from true-up-list nil))
       ((member (current-char) *left-parenthesis-list* :test #'char=)
	(decf level)
	(forward-char 1))
       ((member (current-char) *right-parenthesis-list* :test #'char=)
	(incf level)
	(forward-char 1)
	(if (= level 1) (return-from true-up-list *point*)))
       ((member (current-char) '(#\" #\|))
	(skip-string-or-literal))
       ((char= (current-char) #\\)
	(forward-char 2))
       ((normal-char-p (current-char))
	(forward-word 1))
       (t (forward-char 1))))))

;;; M-C-u
;;; ˻ʬΤä٥˰ư롣
;;;
;;; ʬưǤȤϥݥȰ֤֤
;;; ǤʤȤ nil֤
;;;
(defun backward-up-list (&optional (n 1) &aux result)
  (declare (fixnum n))
  (setf n (select-repeat-count n))
  (cond
   ((zerop n) nil)
   ((> n 0)
    (dotimes (i n result) (setf result (true-backward-up-list))))
   (t (up-list (- n)))))

;;;
;;; ˤҤȤľΤä٥˰ư롣
;;; ưоݤΣӼʤХݥȤϰưnil֤
;;; ưǤХݥȰ֤֤
;;;
(defun true-backward-up-list ()
  (let (pos line len level in-string result)
    (setf result t)
    (setf pos *point*)
    (setf line (copy-seq *text*))
    (setf len (length line))
    (setf *text* (invert-list line))
    (move-point-to (- len pos))
    (block main
      (setf level 0)
      (setf in-string (in-string-or-literal-p))
      (when in-string
	(move-point-to in-string)
	(skip-string-or-literal))
      (loop
	(skip-white-space)
	(cond
	 ((end-of-text-p)
	  (if (= level 0) (setf result nil))
	  (return result))
	 ((member (current-char) *left-parenthesis-list* :test #'char=)
	  (decf level)
	  (forward-char 1))
	 ((member (current-char) *right-parenthesis-list* :test #'char=)
	  (incf level)
	  (forward-char 1)
	  (if (= level 1) (setf result *point*))
	  (return result))
	 ((member (current-char) '(#\" #\|))
	  (skip-string-or-literal))
	 ((char= (current-char) #\\)
	  (forward-char 2))
	 ((normal-char-p (current-char))
	  (forward-word 1))
	 (t (forward-char 1)))))
    (when result (setf pos (- len *point*)))
    (setf *text* line)
    (move-point-to pos)
    (return-from true-backward-up-list result)))

;;; M-C-n (move to Next list)
;;; ʬꥹñ̤ǿʤࡣ
;;; ΰꤵ줿Ȥϡո˿ʤࡣ
;;;
;;; ʬưǤȤϥݥȰ֤֤
;;; ǤʤȤ nil֤
;;;
(defun forward-list (&optional (n 1) &aux result)
  (declare (fixnum n))
  (setf n (select-repeat-count n))
  (cond
   ((zerop n) nil)
   ((> n 0)
    (dotimes (i n result) (setf result (true-forward-list))))
   (t (backward-list (- n)))))

;;;
;;; ꥹȤҤȤʬʤ
;;; ưоݤΣӼʤХݥȤϰưnil֤
;;; ưǤХݥȰ֤֤
;;;
(defun true-forward-list ()
  (when (down-list 1) (up-list 1)))

;;; M-C-p (move to Previous list)
;;; ʬꥹñ̤ǵո˿ʤࡣ
;;; ΰꤵ줿Ȥϡ˿ʤࡣ
;;;
;;; ʬưǤȤϥݥȰ֤֤
;;; ǤʤȤ nil֤
;;;
(defun backward-list (&optional (n 1) &aux result)
  (declare (fixnum n))
  (setf n (select-repeat-count n))
  (cond
   ((zerop n) nil)
   ((> n 0)
    (dotimes (i n result) (setf result (true-backward-list))))
   (t (forward-list (- n)))))

;;;
;;; ꥹȤҤȤʬ
;;; ưоݤΣӼʤХݥȤϰưnil֤
;;; ưǤХݥȰ֤֤
;;;
(defun true-backward-list ()
  (let (pos line len result)
    (setf pos *point*)
    (setf line (copy-seq *text*))
    (setf len (length line))
    (setf *text* (invert-list line))
    (move-point-to (- len pos))
    (setf result (forward-list 1))
    (when result (setf pos (- len *point*)))
    (setf *text* line)
    (move-point-to pos)
    (return-from true-backward-list result)))

;;;
;;; ʸꥹȤսˤ롣ոФƽ
;;; ؿŬѤǤ褦ˤ뤿ˡʲѴԤ
;;;
;;;	'('		--> ')'
;;;	')'		--> '('
;;;	'\x'		--> '\x'
;;;	'#\\word'	--> '#\\word'
;;;
;;; ȤС
;;;	"(load \"line-edit-pkg.lsp\")"	==> "(\"psl.gkp-tide-enil\" daol)"
;;;	"(list #\\a #\\Space #\\b)"	==> "(#\\b #\\Space #\\a tsil)"
;;; Ȥʤ롣
;;; 	
(defun invert-list (lst)
  (declare (list lst))
  (let (inv-lst rest buf ch)
    (if (not (listp lst)) (return-from invert-list nil))
    (setf inv-lst nil)
    (setf rest (length lst))
    (loop
      (if (null lst) (return inv-lst))
      (setf ch (pop lst))
      (decf rest)
      (cond
       ((< rest 0) inv-lst)
       ((member ch *left-parenthesis-list* :test #'char=)
 	(push (corres-paren ch) inv-lst))
       ((member ch *right-parenthesis-list* :test #'char=)
 	(push (corres-paren ch) inv-lst))
       ((char= ch #\\)
	(when lst
	  (setf ch (pop lst))
	  (decf rest)
	  (push ch inv-lst))
	(push #\\ inv-lst))
       ((char= ch #\#)
	(setf buf nil)
	(push #\# buf)
	(setf ch (pop lst))
	(decf rest)
	(cond
	 ((null lst)
	  (push ch buf)
	  (setf inv-lst (append (reverse buf) inv-lst)))
	 ((member ch '(#\\ #\') :test #'char=)
	  (push ch buf)
	  (push (pop lst) buf)
	  (decf rest)
	  (cond
	   ((null lst)
	    (setf inv-lst (append (reverse buf) inv-lst)))
	   (t
	    (loop
	      (setf ch (pop lst))
	      (decf rest)
	      (cond
	       ((null lst)
		(cond
		 ((normal-char-p ch)
		  (push ch buf)
		  (setf inv-lst (append (reverse buf) inv-lst))
		  (return))
		 (t (push ch lst)
		    (incf rest)
		    (setf inv-lst (append (reverse buf) inv-lst))
		    (return))))
	       ((not (normal-char-p ch))
		(push ch lst)
		(incf rest)
		(setf inv-lst (append (reverse buf) inv-lst))
		(return))
	       (t (push ch buf)))))))
	 (t (setf inv-lst (append (reverse buf) inv-lst))
	    (push ch lst)
	    (incf rest))))
       (t (push ch inv-lst))))))

;;;
;;; Ʊб뤫ä֤
;;;
(defun corres-paren (paren)
  (let (n)
    (cond
     ((setf n (search (list paren) *left-parenthesis-list* :test #'char=))
      (nth n *right-parenthesis-list*))
     ((setf n (search (list paren) *right-parenthesis-list* :test #'char=))
      (nth n *left-parenthesis-list*)))))

;;;
;;; äȻꤵ줿Τΰ֤
;;;
(defun parenthesis-list ()
  (pairlis *left-parenthesis-list* *right-parenthesis-list*))

;;;
;;; äꥻåȤ롣
;;;
(defun reset-parenthesis-list ()
  (setf *left-parenthesis-list* nil)
  (setf *right-parenthesis-list* nil))

;;;
;;; ФΤäȡꥹȤƬɲä롣
;;;	(set-balanced-paren #\( #\))
;;; ǿΤä֤
;;;
(defun set-balanced-paren (left right)
  (when (and (characterp left) (characterp right))
    (push left *left-parenthesis-list*)
    (push right *right-parenthesis-list*)
    (parenthesis-list)))

;;;
;;; ꤵ줿ڥˤʤ뤫ä򤫤ä롣
;;;
;;; ⤷ҤȤĤꤵƤʤХꥹȤ򺸤
;;; 鸡ǽ˸ĤääȡФˤʤ뱦
;;; 롣
;;;
;;; ⤷ĻꤵƤСפФΤä
;;; 롣
(defun unset-balanced-paren (left &optional right)
  (cond
   ((not (characterp left)) nil)
   ((or
     (null right)
     (characterp right))
    (make-parenthesis-list (true-unset-balanced-paren left right (parenthesis-list))))
   (t nil)))

(defun true-unset-balanced-paren (left right alst)
  (cond
   ((null alst) nil)
   ((char= left (caar alst))
    (cond
     ((null right) (cdr alst))
     ((char= right (cdar alst)) (cdr alst))
     (t (cons (car alst) (true-unset-balanced-paren left right (cdr alst))))))
   (t (cons (car alst) (true-unset-balanced-paren left right (cdr alst))))))

;;;
;;; dotted listɽ蘆줿Фˤʤ뤫äΥꥹȤ餫ä
;;; ꥹȤ롣
;;;
(defun make-parenthesis-list (alst)
  (reset-parenthesis-list)
  (loop
    (when (null alst) (return))
    (when (not (and (characterp (caar alst)) (characterp (cdar alst)))) (return))
    (setf *left-parenthesis-list* (append *left-parenthesis-list* (list (caar alst))))
    (setf *right-parenthesis-list* (append *right-parenthesis-list* (list (cdar alst))))
    (pop alst))
  (parenthesis-list))

;;; M-C-@
;;; ꤵ줿ĿΣӼޤ꡼ꤹ롣
;;;
(defun mark-sexp (&optional (n 1))
  (declare (fixnum n))
  (setf n (select-repeat-count n))
  (forward-sexp n)
  (set-mark-command)
  (backward-sexp n))

;;; M-(
;;; ꤵ줿ĿʬΣӼ򤫤äǰϤߡݥȤ򳫤äľ
;;; ư롣Ӽ¸ߤʤУФΤä롣
;;;
(defun insert-parenthesis (&optional (n 1))
  (declare (fixnum n))
  (let (deley)
    (setf n (select-repeat-count n))
    (setf deley (blink-paren-deley))
    (blink-paren-deley 0)
    (save-mark-info)
    (cond
     ((zerop n) nil)
     ((> n 0)
      (self-insert (first *left-parenthesis-list*) 1)
      (set-mark-command)
      (forward-sexp n)
      (self-insert (first *right-parenthesis-list*) 1)
      (exchange-point-and-mark))
     ((< n 0)
      (self-insert (first *left-parenthesis-list*) 1)
      (backward-sexp (- n))
      (self-insert (first *right-parenthesis-list*) 1)))
    (align-mark)
    (blink-paren-deley deley)))

;;; M-C-k
;;; ߤΥݥȤʬΣӼνޤǤ򥭥뤹롣
;;;
(defun kill-sexp (&optional (n 1))
  (declare (fixnum n))
  (let (result)
    (setf n (select-repeat-count n))
    (cond
     ((zerop n) nil)
     ((< n 0)
      (backward-kill-sexp (- n)))
     ((> n 0)
      (save-mark-info)
      (set-mark-command)
      (forward-sexp n)
      (exchange-point-and-mark)		;force forward kill.
      (setf result (kill-region))
      (align-mark)
      (return-from kill-sexp result)))))

;;;
;;; ߤΥݥȤʬΣӼγϰ֤ޤǤ򥭥뤹롣
;;;
(defun backward-kill-sexp (&optional n)
  (let (result)
    (setf n (select-repeat-count n))
    (cond
     ((zerop n) nil)
     ((< n 0)
      (kill-sexp (- n)))
     ((> n 0)
      (save-mark-info)
      (set-mark-command)
      (backward-sexp n)
      (exchange-point-and-mark)		;force backward kill.
      (setf result (kill-region))
      (align-mark)
      (return-from backward-kill-sexp result)))))

;;; M-C-t
;;; ݥȤΣӼؤ롣ӼȣӼδ֤ζʸϡ
;;; Τޤ¸ߤ롣ݥȤؤӼľ˰ư롣
;;;
;;;  0ΤȤϥݥȤľΣӼȥޡľΣӼ
;;; ؤ롣ݥȰ֤ȥޡ֤ؤ롣
;;;
(defun transpose-sexps (&optional (n 1)) ;transpose Symbolic EXPressionS
  (declare (fixnum n))
  (let (pos pos2 tm)
    (setf n (select-repeat-count n))
    (setf tm (blink-paren-deley))
    (blink-paren-deley 0)
    (cond
     ((> n 0)
      (dotimes (i n) (true-transpose-sexps)))
     ((< n 0)
      (dotimes (i (- n))
	(true-backward-sexp)
	(true-transpose-sexps)
	(true-backward-sexp)))
     ((zerop n)
      (cond
       ((null *mark*) nil)
       (*mark*
	(kill-ring-suspend t)
	(setf pos *point*)
	(kill-sexp 1)
	(exchange-point-and-mark)
	(yank 1) (pop-kill-ring)
	(kill-ring-suspend t)
	(setf pos2 *point*)
	(kill-sexp 1)
	(move-point-to pos)
	(yank 1) (pop-kill-ring)
	(move-point-to pos)
	(set-mark-command)
	(move-point-to pos2)))))
    (blink-paren-deley tm)))

(defun true-transpose-sexps ()
  (if (empty-text-p) (return-from true-transpose-sexps nil))
  (if (beginning-of-text-p) (return-from true-transpose-sexps nil))
  (if (end-of-text-p) (return-from true-transpose-sexps nil))
  (save-mark-info)
  (set-mark-command)
  ;; Is previous sexp exists ?
  (backward-skip-white-space)
  ;Ignore quote character, if exists. 2001/08/12
  (when (not (beginning-of-text-p))
    (backward-char 1)
    (if (char= (current-char) #\')
	(backward-skip-white-space)
      (forward-char 1)))
  (when (beginning-of-text-p)
    (exchange-point-and-mark)
    (align-mark)
    (audio-bell)
    (return-from true-transpose-sexps nil))
  ;; Is next sexp exists ?
  (exchange-point-and-mark)
  (set-mark-command)
  (skip-white-space)
  (when (in-word-p) (forward-word 1))
  (when (not (forward-sexp 1))
    (exchange-point-and-mark)
    (align-mark)
    (audio-bell)
    (return-from true-transpose-sexps nil))
  ;; Main block
  (exchange-point-and-mark)
  (backward-sexp 1)
  (kill-ring-suspend t)
  (kill-sexp 1)
  (kill-ring-suspend t)
  (kill-white-space)
  (forward-sexp 1)
  (yank 1)
  (yank 1) (yank-pop 1)
  (pop-kill-ring)
  (pop-kill-ring)
  (align-mark))

;;; M-p
;;; ֤ʸäǤСб뤫ä˰ư롣
;;; б뤫ä¸ߤʤХݥȤϰưʤ
;;; ֤ʸäʳǤСΤä˰ư롣
;;;
(defun move-to-matching-paren (&optional (n 1))
  (declare (fixnum n))
  (setf n (select-repeat-count n))
  (cond
   ((zerop n) nil)
   ((< n 0) nil)
   ((= (mod n 2) 1) (true-move-to-matching-paren))))

(defun true-move-to-matching-paren ()
  (cond
   ((end-of-text-p) nil)
   (t (pure-move-to-matching-paren))))

(defun pure-move-to-matching-paren ()
  (let (pos)
    (cond
     ((member (current-char) *left-parenthesis-list* :test #'char=)
      (when (forward-list 1) (backward-char 1)))
     ((member (current-char) *right-parenthesis-list* :test #'char=)
      (forward-char 1)
      (when (backward-list 1) (return-from pure-move-to-matching-paren t))
      (backward-char 1)
      (return-from pure-move-to-matching-paren nil))
     (t
      (setf pos *point*)
      (loop
	(forward-char 1)
	(when (end-of-text-p)
	  (setf *point* pos)
	  (return))
	(if (member (current-char) *parenthesis-list* :test #'char=)
	    (return)))))))

;;;
;;; б뤫äÿư롣
;;;
(defun blink-paren ()
  (if (blink-paren-p) (jump-and-return-balanced-paren)))

;;;
;;; б뤫ä˰ưƤ뤫ݤ֤
;;;
(defun blink-paren-p ()
  (and (numberp (blink-paren-deley)) (plusp (blink-paren-deley))))

;;; ǤȤθߴΤ
(defun highlight-paren ()
  (if (blink-paren-p) (jump-and-return-balanced-paren)))

;;;
;;; б뤫ä˰ưÿꤹ롣
;;; ʤꤵƤÿ֤
;;; б뤫äؤΰư򥪥դˤˤ 0ޤϿͰʳͤꤹ롣
;;;
(defun blink-paren-deley (&optional (sec 0 sw))
  (cond
   ((null sw) *blink-paren-deley*)
   (t (setf *blink-paren-deley* sec))))

;;; ǤȤθߴΤ
(defun highlight-paren-deley (&optional (sec 0 sw))
  (cond
   ((null sw) (blink-paren-deley))
   (t (blink-paren-deley sec))))

;;;
;;; б뤫ä (blink-paren-deley)äư롣
;;; ݥȤǤʤưԤ
;;;
(defun jump-and-return-balanced-paren ()
  (let (pos)
    (setf pos *point*)
    (when (move-to-matching-paren 1)
      (sync-point-and-cursor)
      (xsleep (blink-paren-deley))
      (move-point-to pos)
      (sync-point-and-cursor)
      (return-from jump-and-return-balanced-paren t))
    (return-from jump-and-return-balanced-paren nil)))

;;;
;;; (sleep sec)Ʊ "sleep is permitted to use approximate timing."
;;; ʤΤǡɤΤϽϤ˰¸롣GCL 2.3.8-BetaǤ
;;;
;;;	> (time (sleep 0.5))
;;;	real time : 0.000 secs
;;;	run time  : 0.000 secs
;;;	
;;;	> (time (sleep 0.6))
;;;	real time : 1.000 secs
;;;	run time  : 0.000 secs
;;;
;;; Ǥäʲxsleep٤Ϥȼ¹ԥޥ®ٰ¸Ǥ뤬
;;; 0.1ñ̤٤ԤƤ롣
;;; Pentium III 700MHz + GCL 2.3.8-Beta/Linux 2.4.0-test8ǤϡԤɤ
;;; ٤Ǥä
;;;
(defun xsleep (sec)
  (let (ctime second)
    (setf second (* sec internal-time-units-per-second))
    (setf ctime (get-internal-real-time))
    (loop
      (when (>= (- (get-internal-real-time) ctime) second)
	(return-from xsleep sec)))))

;;;
;;; ʸϻΤߡб뤫ä餻뤫ɤꤹ롣
;;; ꤷʤȸߤ֤
;;;
(defun blink-paren-just-inserted (&optional (blink-p nil sw))
  (cond
   ((null sw) *blink-paren-just-inserted*)
   (t (setf *blink-paren-just-inserted* blink-p))))

;;; ǤȤθߴΤ
(defun highlight-paren-just-inserted (&optional (highlight-p nil sw))
  (cond
   ((null sw) (blink-paren-just-inserted))
   (t (blink-paren-just-inserted highlight-p))))

;;;
;;; ݥȤ Common Lisp ˤʸˤ뤫ɤ
;;; Ĵ٤롣⤷ǤСʸƬ֤֤
;;;
;;; Is pointer in string (defined by Common Lisp) ?
;;; returns start position of its string, or nil.
;;;
(defun in-string-or-literal-p ()
  (prog (pos start in-string-or-literal)
    (save-mark-info)
    (set-mark-command)
    (setf pos *point*)
    (setf in-string-or-literal nil)
    (setf start nil)
    (beginning-of-line 1)
    (loop
      (if (end-of-text-p) (return))
      (loop
	(if (end-of-text-p) (return))
	(when (member (current-char) '(#\" #\|))
	  (setf start *point*)
	  (return))
	(if (char= (current-char) #\\) (forward-char 2))
	(forward-char 1))
      (when (or (null start) (>= *point* pos))
	(return))
      (skip-string-or-literal)
      (when (<= start pos *point*)
	(setf in-string-or-literal t)
	(return))
      (forward-char 1))
    (move-point-to pos)
    (align-mark)
    (if in-string-or-literal (return start) (return nil))))

;;;
;;; ʸޤϥƥɤФ
;;;
(defun skip-string-or-literal ()
  (cond
   ((end-of-text-p) nil)
   ((char= (current-char) #\")
    (skip-dquoted-string))
   ((char= (current-char) #\|)
    (skip-bar-string))))

;;;
;;; ʸɤФ
;;;
(defun skip-dquoted-string ()
  (block nil
    (if (or (end-of-text-p) (char/= (current-char) #\")) (return nil))
    (forward-char 1)
    (loop
      (if (end-of-text-p) (return *point*))
      (cond
       ((char= (current-char) #\")
	(forward-char 1)
	(return-from skip-dquoted-string *point*))
       ((char= (current-char) #\\)
	(forward-char 2))
       (t (forward-char 1))))))

;;;
;;; ƥɤФ
;;;
(defun skip-bar-string ()
  (block nil
    (if (or (end-of-text-p) (char/= (current-char) #\|)) (return nil))
    (forward-char 1)
    (loop
      (if (end-of-text-p) (return *point*))
      (cond
       ((char= (current-char) #\|)
	(forward-char 1)
	(return-from skip-bar-string *point*))
       ((char= (current-char) #\\)
	(forward-char 2))
       (t (forward-char 1))))))

;;;
;;; ݥȰ֤˰ǻꤵʸ롣
;;; ʸХݥȰ֤֤Ǥʤ nil֤
;;;
;;;	In GCL-2.3.8 on Linux 2.4.0/Pentium III-700MHz
;;;
;;;	> (setf x (concatenate 'list "This is a sample code."))
;;;	> (setf y (concatenate 'list " and this is trailing code."))
;;;	> (time (dotimes (i 100000) (append x (list #\&) y)))
;;;	real time : 0.940 secs
;;;	run time  : 0.930 secs
;;;
;;;	> (time (dotimes (i 100000) (concatenate 'list x (list #\&) y)))
;;;	real time : 4.060 secs
;;;	run time  : 4.030 secs
;;;
;;;	So, I use append in this case.
;;;
(defun self-insert (ch &optional n)
  (declare (character ch))
  (let (char-lst)
    (if (null n) (setf n (repeat-count)))
    (if (null n) (setf n 1))
    (if (not (characterp ch)) (return-from self-insert nil))
    (cond
     ((zerop n) nil)
     ((= n 1)
      (save-self-insert-undo-info ch)
      (true-self-insert ch))
     ((> n 1)
      (setf char-lst nil)
      (dotimes (i n) (push ch char-lst)) 
      (insert char-lst))
     ((< n 0)
      (setf char-lst nil)
      (dotimes (i (- n)) (push ch char-lst))
      (insert char-lst)
      (backward-char (length char-lst))))))

(defun true-self-insert (ch)
  (declare (character ch))
  (if (not (characterp ch)) (return-from true-self-insert nil))
  (if (end-of-text-p) (setf *text* (nconc *text* (list ch)))
    (setf *text* (pure-self-insert *text* *point* ch)))
  (if (and (numberp *mark*) (> *mark* *point*)) (incf *mark*))
  (incf *point*)
  (incf *ncl*)
  (when
      (or
       (member ch *right-parenthesis-list* :test #'char=)
       (and
	(not (end-of-text-p))
	(member ch *left-parenthesis-list* :test #'char=)))
    (when (blink-paren-p)
      (backward-char 1)
      (if (jump-and-return-balanced-paren)
	  (forward-char 1))))
  (need-refresh-line t)
  (return-from true-self-insert *point*))

(defun pure-self-insert (lst n ch)
  (append (subseq lst 0 n) (list ch) (subseq lst n)))

;;;
;;; ʸݥȰ֤롣
;;;
(defun self-insert-string (str)
  (insert (unpack str)))

;;;
;;; ɽΡʸפݥȰ֤롣
;;; ˥ݥȰ֤֤
;;;
(defun insert (char-lst)
  (save-insert-undo-info char-lst)
  (dolist (ch char-lst *point*) (true-self-insert ch)))

;;; C-q
;;; ǻꤵ줿ĿʬȤʸ롣
;;;
(defun quoted-insert (&optional (n 1))
  (declare (fixnum n))
  (setf n (select-repeat-count n))
  (self-insert (getch) n))

;;; C-j
;;; #\Newline롣
;;;
(defun self-insert-newline (&optional (n 1))
  (declare (fixnum n))
  (setf n (select-repeat-count n))
  (self-insert #\Newline n))

;;;
;;; #\Space롣
;;;
(defun self-insert-space (&optional n)
  (setf n (select-repeat-count n))
  (self-insert #\Space n))

;;; C-l
;;; ߤιΤľ
;;; #\Newline ʳΤ٤ƤʸݥȤ
;;; ǻꤵ줿֤ɽ褦˹Ԥɽľ
;;;
;;;  nilξ硢(auto-scroll-offset)ؿ
;;;
;;;     ͤꤵƤиߤɽϰϤѹɽ
;;;     ϰΤ褹롣
;;;
;;;     nil ǤͤǤʤͤꤵƤС
;;;     Ԥɽ̤Ȥʤ褦˥ݥ
;;;     ֤Ĵƺ褹롣
;;;
(defun recenter (&optional (col nil))
  (let (pos)
    (setf col (select-repeat-count col))
    (kill-ring-suspend t)
    (distill-line nil)
    (cond
     ((numberp col)
      (sync-point-and-cursor col))
     ((numberp (auto-scroll-offset))
      (display-line))
     ((auto-scroll-offset)
      (setf pos *point*)
      (move-point-to (car (display-range)))
      (sync-point-and-cursor 0)
      (move-point-to pos)
      (need-refresh-line t)))))

;;; M-r
;;; ߤΥ֤Ȥʤ褦˹Ԥľ
;;;
(defun recenter-zero ()
  (recenter 0))

;;;
;;; ľɬפȤʤ볫ϰ֤Ͽ롣
;;;
;;; ߤϽľɬפݤξȤƤΤ߻ѤƤꡢ
;;; ϰ֤ˤϴΤƤʤ
;;;
(defun need-refresh-line (is-set)
  (cond
   (is-set
    (cond
     ((null *need-refresh-line*)
      (setf *need-refresh-line* *point*))
     (t (setf *need-refresh-line* (min *point* *need-refresh-line*)))))
   (t (setf *need-refresh-line* nil))))

;;;
;;; ľɬפݤ֤
;;;
(defun need-refresh-line-p () *need-refresh-line*)

;;;
;;;  nilʤ#\NewlineʳƤ޷ʸ롣
;;;  nilʳʤ#\NewlineޤƤ޷ʸ롣
;;;
(defun distill-line (&optional del-all-p)
  (set-mark-command)
  (beginning-of-text)
  (loop
    (if (end-of-text-p) (return))
    (cond
     ((and (not del-all-p) (char= (current-char) #\Newline))
      (forward-char 1))
     ((not (graphic-char-p (current-char)))
      (delete-char nil))
     (t (forward-char 1))))
  (exchange-point-and-mark))

;;; M-k
;;; ߤƥȤ򥭥뤷򸡺ϰ֤ꥻåȤ롣
;;; undoΤƤ롣
;;;
(defun kill-text ()
  (beginning-of-text)
  (set-mark-command)
  (end-of-text)
  (kill-ring-suspend t)
  (kill-region)
  (pop-kill-ring)
  (reset-scroll)
  (reset-undo-info)
  (search-from nil))

;;; C-x r s x
;;; 쥸 'x'˥꡼򥳥ԡ롣
;;; ְꤵƤϡְ
;;; ꡼ʣ̤ƥԡ롣
;;;
(defun copy-to-register (&optional (n 1))
  (let (r lst tmp)
    (setf n (select-repeat-count n))
    (kill-ring-suspend t)
    (kill-ring-save)
    (setf r (getchar))
    (setf lst (pop-kill-ring))
    (setf tmp nil)
    (dotimes (i n)
      (setf tmp (append tmp lst)))
    (set-register r tmp)))

;;; C-x r i x
;;; 쥸 'x'ݥȰ֤˥ƥȤ롣
;;; ְꤵƤϡְ쥸
;;;  'x'Ƥ򷫤֤롣
;;;
(defun insert-register (&optional (n 1))
  (let (lst)
    (setf n (select-repeat-count n))
    (setf lst (cdr (register (getchar))))
    (dotimes (i n) (insert lst))))

;;;
;;; 쥸 'r'˥ƥȤ¸롣
;;; 裲nilʤ顢쥸 'r'ΤΤ롣
;;;
(defun set-register (r lst)
  (declare (character r) (list lst))
  (cond
   ((null r) nil)
   ((null (assoc r *registers*))
      (push (cons r lst) *registers*))
   (t (setf (cdr (assoc r *registers*)) lst)))
  (setf *registers* (remove-if #'(lambda (x) (null (cdr x))) *registers*)))

;;;
;;; 쥸 'r'Ƥ֤
;;;
(defun register (r)
  (declare (character r))
  (assoc r *registers*))

;;;
;;; 쥸Ƥե¸롣
;;;
(defun save-registers ()
  (with-open-file (stream (registers-file) :direction :output)
    (dolist (i *registers* t)
      (prin1 (cons (car i) (pack (cdr i))) stream)
      (terpri stream))))

;;;
;;; ե¸Ƥ쥸Ƥɤ߹ࡣ
;;;
(defun restore-registers ()
  (let (form)
    (when (probe-file (registers-file))
      (with-open-file (stream (registers-file) :direction :input)
	(loop
	  (setf form (read stream nil *eos*))
	  (when (eq form *eos*) (return-from restore-registers t))
	  (set-register (car form) (unpack (cdr form))))))))

;;;
;;; 쥸Ƥ¸ե̾ꤹ롣ʤξ
;;; ꤵƤե֤̾
;;;
(defun registers-file (&optional fname)
  (cond
   ((null fname) *registers-file*)
   (t (setf *registers-file* fname))))

;;;
;;; ǻꤵ줿쥸̾ȡƤɽ롣
;;; ꤵƤʤϻѤƤ뤹٤
;;; Υ쥸̾ȡƤɽ롣
;;;
(defun view-register (&optional ch)
  (cond
   ((null ch)
    (view-all-registers))
   (t (pure-view-register ch))))

(defun pure-view-register (ch)
  (let (reg)
;   (declare (list reg))
    (setf reg (register ch))
    (when (null reg) (return-from pure-view-register nil))
    (prin1 (car reg))
    (princ #\Tab)
    (princ (pack (cdr reg)))
    (terpri)
    (return-from pure-view-register t)))

(defun view-all-registers ()
  (if (null *registers*) (return-from view-all-registers nil))
  (setf *registers* (sort *registers* #'(lambda (x y) (char< (car x) (car y)))))
  (dolist (i *registers* t) (pure-view-register (car i))))

;;; C-s
;;; ݥȰְʹߤιоݤ˥󥯥󥿥롦Ԥ
;;; õʸʸꤹʸȾʸ̤õ롣
;;; õʸְ㤨C-hǼä롣
;;; õʸ󤫤(٤Ƥ)ʸäȡõФʸ
;;;   ʸζ̤Ԥʤ
;;; C-hˤäõϰ֤ޤǥݥȤä硢󥯥󥿥
;;;   ⽪λ롣
;;; C-gõΤäݥȤõϻΰ֤᤹
;;; õʸC-s򥿥פõʸμνиĽõ롣
;;; Enter, Linefeed, ޤ Esc򥿥פȸλ롣
;;;
(defun isearch-forward ()
  (let (str ch line idx pos first-pos case-sensitive)
    (setf case-sensitive nil)
    (setf first-pos *point*)
    (setf pos first-pos)
    (setf str "")
    (setf line (pack (nthcdr *point* *text*)))
    (loop
      (setf ch (getch))		; MUST be 'getch', NOT 'getchar'
      (cond
       ((member ch '(#\Return #\Linefeed #\^[))
	(return t))
       ((char= ch #\^H)
	(when (= (length str) 0)
	  (move-point-to first-pos)
	  (sync-point-and-cursor)
	  (return nil))
	(setf str (subseq str 0 (1- (length str)))))
       ((char= ch ctrl-g)
	(move-point-to first-pos)
	(sync-point-and-cursor)
	(return nil))
       ((char= ch #\^S)
	(setf pos *point*)
	(setf line (pack (nthcdr *point* *text*))))
       ((not (graphic-char-p ch))
	(unread-char ch)
	(return nil))
       (t (setf str (concatenate 'string str (list ch)))))
      (if (string= str (string-downcase str))
	  (setf case-sensitive nil)
	(setf case-sensitive t))
      (if case-sensitive
	  (setf idx (search str line :test #'(lambda (x y) (char= x y))))
	(setf idx (search str line :test #'(lambda (x y) (char-equal x y)))))
      (when idx
	(move-point-to (+ pos idx (length str)))
	(sync-point-and-cursor)))))

;;; C-x i
;;; ݥȰ֤˥ե롣
;;; ҥȥɽƤޥɹԡ(insert-file)μ¹ԵϿ
;;; ξĤ˺Ǹ(throw...)¹Ԥ롣
;;;
(defun insert-file ()
  (let (from line result)
    (setf result "")
    (save-line)
    (writeln)
    (format t "Insert file from: ")
    (setf from (line-edit))
    (when (string/= from "")
      (setf from (merge-pathnames from))
      (with-open-file (s from :direction :input :if-does-not-exist nil)
	(loop
	  (setf line (read-line s nil *eos*))
	  (when (eq line *eos*) (return result))
	  (setf result (concatenate 'string result line (list #\Newline))))))
    (restore-line)
    (save-mark-info)
    (insert (unpack (string-right-trim '(#\Newline) result)))
    (align-mark)
    (sync-point-and-cursor)
    (throw 'exit-line-edit (line-edit-command-loop))))

;;; C-x C-f
;;; ߤιԤꤷեƤ֤롣
;;;
(defun find-file ()
  (beginning-of-line 1)
  (kill-ring-suspend t)
  (kill-line 1)
  (pop-kill-ring)
  (insert-file))

;;; C-x C-w
;;; ߤιԤꤷե¸롣
;;;
(defun write-file ()
  (let (save-to path line message)
    (save-line)
    (writeln)
    (format t "Save line to: ")
    (setf save-to (line-edit))
    (when (string= save-to "")
      (return-from write-file nil))
    (setf path (merge-pathnames save-to))
    (restore-line)
    (setf line (pack *text*))
    (do-write-file line path)
    (setf message (concatenate 'string "(write-file-to \"" save-to "\")" ))
    (throw 'exit-line-edit message)))

(defun do-write-file (line path)
  (with-open-file (s path :direction :output)
    (write line :stream s :escape nil :pretty t)
    (terpri s)))

(defun write-file-to (fname) fname)

;;; C-x C-s
;;; ߤιԤե¸롣
;;; ¸ե̾򸵤˼ưŪ˺롣
;;;	yyyymmdd-hhmmss.lsp
;;;
(defun save-buffer ()
  (let (save-to path message)
    (setf save-to
	  (concatenate 'string (make-date-and-time-string) *lisp-extension*))
    (setf path (merge-pathnames save-to))
    (do-write-file (pack *text*) path)
    (setf message (concatenate 'string "(write-file-to \"" save-to "\")" ))
    (writeln)
    (throw 'exit-line-edit message)))

;;;
;;; C-x C-sǽ񤭽ФեγĥҤꤹ롣
;;; ʤǸƤӽФȸߤ֤ͤ
;;; ǥեȤ ".lsp"
;;;
(defun lisp-extension (&optional ext)
  (cond
   ((or
     (null ext)
     (not (stringp ext)))
    *lisp-extension*)
   (t (setf *lisp-extension* ext))))

;;;
;;; ߤλyyyymmdd-hhmmssȤʸȤ֤
;;;
;;; (get-decoded-time)ϥƥΥॾ˽ä
;;; ֤ॾδͤˡLispƥ˰¸
;;; 롣(decode-universal-time (get-universal-time) timezone)
;;; Ū˥ॾꤹ뤳ȤǤ뤬ξ
;;; ϲƻ־̵뤵ΤǸߴ꤬롣
;;;
;;; GCLǥॾδͤꤹˤ
;;;	(setf si::*default-time-zone* -9)
;;; Ȥ(ܤξ)Common Lispˤ륿ॾͤ
;;; GMTλֿȤɽפȵꤵƤ
;;; GMTȤ椬դˤʤ롣GMT+0900Ǥ -9
;;;
(defun make-date-and-time-string ()
  (let (tm)
    (setf tm (multiple-value-list (get-decoded-time)))
    (format nil "~4,'0d~2,'0d~2,'0d-~2,'0d~2,'0d~2,'0d"
	    (nth 5 tm) (nth 4 tm) (nth 3 tm)
	    (nth 2 tm) (nth 1 tm) (truncate (nth 0 tm)))))

;;;
;;; ȥ꡼फ飱ʸɤ߹ࡣ
;;; ʸ ctrl-g ʸȤ֤
;;;
(defun getch ()
  (let (ch)
    (setf ch (read-char *standard-input* nil *eos*))
    (if (eq ch *eos*) (throw 'control-g nil) ch)))

;;;
;;; peek-ahead
;;;
(defun peek-ahead ()
  (let (ch)
    (setf ch (peek-char nil *standard-input* nil *eos*))
    (if (eq ch *eos*) (throw 'control-g nil) ch)))

;;;
;;; 1ʸɤ߹ࡣctrl-gϤ줿饿control-gפܤ
;;; Ĥޤꡢʸ ctrl-g ̤Ȥ֤뤳ȤϤʤ
;;;
(defun getchar ()
  (let (ch)
    (setf ch (getch))
    (if (char/= ch ctrl-g) ch (throw 'control-g nil))))

;;;
;;; ֤ʸ chɽ֤1ʬʤ롣
;;;
;;; ʸ ch#\^A#\^Zξ硢
;;; 	*print-case*  :downcaseʤ	ʸAZ
;;; 			:upcase  ʤ	ʸaz
;;; 			:capitalizeʤ	ʸAZ
;;; ȿžɽ(¸ؿ򻲾)롣
;;;
;;; ʸ chʳ޷ʸξ(to-circumflex-accent-char)
;;; ֤ʸ2ʸܤȿžɽ롣
;;;
;;; ȿžɽݡȤƤʤǤʸΰ֤ȼ
;;; Τ뤳ȤǤ롣
;;;
;;; C-l¹ԤȹΤ٤Ƥ޷ʸ롣ä
;;; C-l¹Ԥƺ줿ʸ޷ʸǤ롣򸵤᤹
;;; ˤ C-x u¹ԤФ褤
;;;
;;; 嵭ˡΤäʸΰ֤˥ưC-x =
;;; ɤȤСʸμबɽ롣
;;;
(defun putch (ch)
  (declare (character ch))
  (let (char)
    (when (< -1 (cursor-position) (line-size))
      (setf char (control-to-char ch))
      (if (graphic-char-p ch) (putc char)
	(highlight-mode-putc char))
      (incf *cursor-position*))))

;;;
;;; ʸбʸ֤
;;;
(defun control-to-char (ch)
  (declare (character ch))
  (if (graphic-char-p ch) (return-from control-to-char ch))
  (when (and (characterp (newline-symbol)) (char= ch #\Newline))
    (return-from control-to-char (newline-symbol)))
  (setf ch (pure-control-to-char ch))
  (if (not (graphic-char-p ch)) (return-from control-to-char #\?))
  (cond
   ((eq *print-case* ':downcase)
    (char-upcase ch))
   ((eq *print-case* ':upcase)
    (char-downcase ch))
   ((eq *print-case* ':capitalize)
    (char-upcase ch))))

(defun pure-control-to-char (ch)
  (declare (character ch))
  (let (str-or-ch)
    (setf str-or-ch (to-circumflex-accent-char ch))
    (cond
     ((stringp str-or-ch)
      (elt str-or-ch 1))
;     (nth 1 str-or-ch))
     ((characterp str-or-ch)
      str-or-ch))))

;;;
;;; ʸǤдˤä '^'֤롣
;;; ǤʤСʸΤΤ֤
;;;
(defun to-circumflex-accent-char (ch)
  (declare (character ch))
  (case ch
    (#\^@	"^@")
    (#\^A	"^A")
    (#\^B	"^B")
    (#\^C	"^C")
    (#\^D	"^D")
    (#\^E	"^E")
    (#\^F	"^F")
    (#\^G	"^G")
    (#\^H	"^H")
    (#\^I	"^I")
    (#\^J	"^J")
    (#\^K	"^K")
    (#\^L	"^L")
    (#\^M	"^M")
    (#\^N	"^N")
    (#\^O	"^O")
    (#\^P	"^P")
    (#\^Q	"^Q")
    (#\^R	"^R")
    (#\^S	"^S")
    (#\^T	"^T")
    (#\^U	"^U")
    (#\^V	"^V")
    (#\^W	"^W")
    (#\^X	"^X")
    (#\^Y	"^Y")
    (#\^Z	"^Z")
    (#\^[	"^[")
    (#\^\\	"^\\")
    (#\^]	"^]")
    (#\^^	"^^")
    (#\^_	"^_")
    (otherwise	ch)))

;;;
;;; ƥ#\Newlineɽʸ꡿롣
;;; ꤷʸϡֶĴɽפ롣
;;;
(defun newline-symbol (&optional (ch #\$ sw))
  (cond
   ((null sw) *newline-symbol*)
   (t (setf *newline-symbol* ch))))

;;; C-x =
;;; ݥȤľˤʸʸɤȥݥȰ֡
;;; ֤ɽ롣ǤդʸϤȸɽ롣
;;;
;;; Char: c (0143, 99, 0x63) point=32 of 127(25%) column 12
;;;
;;; ݥȤƥȽüˤʸ˴ؤɽʤ
;;;
;;; End-of-text: point=64 column 12
;;;
;;; Ƭη֤ 0ȿ롣
;;;
(defun what-cursor-position ()
  (let (n ch pos col str fmt)
    (setf pos *point*)
    (setf col (cursor-position))
    (setf fmt "Char: ~a (0~o, ~d, 0x~2,'0x) point=~d of ~d(~d\%) column ~d")
    (cond
     ((not (end-of-text-p))
      (setf ch (current-char))
      (setf n (char-code ch))
      (setf ch (to-circumflex-accent-char ch))
      (setf str
	(format nil fmt
		ch n n n pos *ncl* (floor (* 100 (/ (1+ pos) *ncl*))) col)))
     (t (setf str (format nil "End-of-text: point ~d column ~d" pos col))))
    (with-prompt str
	(getch))))

;;;
;;; ʸϴؿ
;;;
(defun putc (c)
  (declare (character c))
  (if *echo* (write-char c)))

(defun putnum (num)
  (declare (fixnum num))
  (if *echo* (princ num)))

;;;
;;; Ĵɽ⡼ɤʸϤ롣
;;;
(defun highlight-mode-putc (c)
  (declare (character c))
  (highlight-mode t)
  (putc c)
  (highlight-mode nil))

;;;
;;; ƥȤɽ̾ΡʸѴ롣
;;;
(defun pack (char-list)
  (declare (list char-list))
  (concatenate 'string char-list))

;;;
;;; ʸɽΡ˥ƥȤѴ롣
;;;
(defun unpack (str)
  (declare (string str))
  (concatenate 'list str))

;;;
;;; ʸηϤ˰¸ʤֿ͡פѴԤ
;;;
(defun char-to-digit (ch)
  (case ch
    (#\0	0)
    (#\1	1)
    (#\2	2)
    (#\3	3)
    (#\4	4)
    (#\5	5)
    (#\6	6)
    (#\7	7)
    (#\8	8)
    (#\9	9)
    (otherwise	nil)))

;;;
;;; ֿ͢פѴԤ
;;;
(defun digit-to-char (n)
  (case n
    (0		#\0)
    (1		#\1)
    (2		#\2)
    (3		#\3)
    (4		#\4)
    (5		#\5)
    (6		#\6)
    (7		#\7)
    (8		#\8)
    (9		#\9)
    (otherwise nil)))

;;;
;;; ٥벻ͭˤ뤫ݤβݤꤹ롣
;;; nil ʤ̵nil ʳʤͭꤵ롣
;;; ʤϡߤ֤ͤ
;;;
(defun enable-audio-bell (&optional (s t sw))
  (cond
   ((null sw) *audio-bell*)
   (t (setf *audio-bell* s))))

;;;
;;; 1ʸundo¸롣
;;;
(defun save-delete-undo-info (ch)
  (declare (character ch))
  (save-undo-info '- (list ch)))

;;;
;;; ؤ1ʸundo¸롣
;;;
(defun save-backward-delete-undo-info (ch)
  (declare (character ch))
  (save-undo-info '-- (list ch)))

;;;
;;; ꡼󡦥undo¸롣
;;;
(defun save-kill-undo-info (char-lst)
  (save-undo-info '- char-lst))

;;;
;;; ؤΥ꡼󡦥undo¸롣
;;;
(defun save-backward-kill-undo-info (char-lst)
  (save-undo-info '-- char-lst))

;;;
;;; 1ʸundo¸롣
;;;
(defun save-self-insert-undo-info (ch)
  (declare (character ch))
  (save-undo-info '+ (list ch)))

;;;
;;; ʸundo¸롣
;;;
(defun save-insert-undo-info (char-lst)
  (save-undo-info '+ char-lst))

;;;
;;; undo¸롣¸ƥå׿ۤƤ
;;; ǿʬĤƸŤΤƤ롣
;;;
;;; ߤǤǤϥޡͤ¸ƤǻѤƤʤ
;;;
;;; undoΥǡ¤ϼΤȤꡣ
;;; 	(ݥ ޡ ( (ʸǡ)))
;;; ǡ㤨
;;; 	(0 nil (- (#\( #\l #\o #\a #\d)))
;;;
(defun save-undo-info (insert-or-kill char-lst)
  (push (list *point* *mark* (cons insert-or-kill (list char-lst)))
	*command-undo-info*)
  (setf *max-undo-length* (max *max-undo-length* (length *undo-info*)))
  (if (and (undo-limit) (> (length *undo-info*) (undo-limit)))
    (setf *undo-info* (head (undo-limit) *undo-info*))))

;;;
;;; 2ͿꥹȤƬ numܤޤǤǤĥꥹȤ֤
;;; ꥹȤĹ numİʲΤȤ lstΤ֤롣
;;;
(defun head (num lst)
  (let (len)
    (setf len (length lst))
    (cond
     ((>= num len) lst)
     (t (reverse (nthcdr (- len num) (reverse lst)))))))

;;;
;;; undoꥻåȤ롣
;;;
(defun reset-undo-info ()
  (set-command-undo-info nil)
  (set-undo-control nil)
  (set-undo-info nil)
  (set-redo-info nil))

(defun set-command-undo-info (lst)
  (setf *command-undo-info* lst))

(defun command-undo-info () *command-undo-info*)

(defun set-undo-control (s) (setf *undo-control* s))

(defun undo-control () *undo-control*)

(defun undo-info () *undo-info*)

(defun set-undo-info (n) (setf *undo-info* n))

(defun push-undo-info (n) (push n *undo-info*))

(defun redo-info () *redo-info*)

(defun set-redo-info (n) (setf *redo-info* n))

(defun push-redo-info (n) (push n *redo-info*))

;;;
;;; undoФͤꤹ롣
;;; ξ¤Ϥʤnilꤹ̵¤Ȥʤ롣
;;; ꤷʤꤵƤ֤
;;; ꤵ줿²ͭˤʤΤϼιԤϳϤ顣
;;;
(defun undo-limit (&optional (step 300 sw))
  (cond
   ((null sw) *undo-limit*)
   (t (setf *undo-limit* step))))

;;; C-x u [M-,]
;;; ѹ()undo롣
;;; Ϣ³Ƽ¹Ԥ뤳ȤǼundoǤ롣
;;;
(defun advertised-undo ()
  (let (tm)
    (setf tm (blink-paren-deley))
    (blink-paren-deley 0)
    (undo)
    (blink-paren-deley tm)))

(defun undo ()
  (let (undo-list)
    (setf *exec-undo* nil)
    (when (not
	   (or (advertised-undo-command-p (last-command))
	       (advertised-redo-command-p (last-command))))
      (setf *redo-info* nil))
    (when (null *undo-info*)
      (audio-bell)
      (return-from undo nil))
    (setf *undo-control* 'undo)
    (setf undo-list (pop *undo-info*))
    (dolist (cmds undo-list) (true-undo cmds))
    (setf *exec-undo* t)))

;;; C-x U [M-.]
;;; undoundo롣
;;; Ǹ(ưǤʤºݤ)¹Ԥޥɤ
;;; undoredoΤȤΤ߼¹ԤǤ롣Ϣ³Ƽ¹Ԥ뤳
;;; ȤǼredoǤ롣
;;;
(defun advertised-redo ()
  (let (tm)
    (setf tm (blink-paren-deley))
    (blink-paren-deley 0)
    (redo)
    (blink-paren-deley tm)))

(defun redo ()
  (let (redo-list)
    (when (null *redo-info*)
      (audio-bell)
      (return-from redo nil))
    (cond
     ((and (advertised-undo-command-p (last-command)) *exec-undo*)
      (setf *undo-control* 'redo))
     ((and (advertised-redo-command-p (last-command)) *exec-redo*)
      (setf *undo-control* 'redo))
     ((null *undo-info*)
      (setf *undo-control* 'redo))
     (t (setf *exec-redo* nil)
	(audio-bell)
	(return-from redo nil)))
    (setf redo-list (pop *redo-info*))
    (dolist (cmds redo-list) (true-undo cmds))
    (setf *exec-redo* t)))

(defun true-undo (cmds)
  (let (point mark insert-or-kill char-lst match-p)
    (save-mark-info)
    (setf point (nth 0 cmds))
    (setf mark (nth 1 cmds))
    (setf insert-or-kill (nth 0 (nth 2 cmds)))
    (setf char-lst (nth 1 (nth 2 cmds)))
    (cond
     ((equal insert-or-kill '-)		;delete/kill, undo is insert.
      (move-point-to point)
      (insert char-lst)
      (backward-char (length char-lst)))
     ((equal insert-or-kill '--)	;backward delete/kill, undo is insert.
      (move-point-to point)
      (insert char-lst))
     ((equal insert-or-kill '+)		;insert, undo is kill (not delete).
      (setf match-p (search char-lst (subseq *text* point) :test #'equal))
      (cond
       ((exactly-matched match-p)	;fail-safe guard.
	(move-point-to point)
	(kill-ring-suspend t)
	(forward-char (length char-lst))
	(delete-backward-char (length char-lst)))
       (t (audio-bell)))))
    (align-mark)))

;;;
;;; *redo-info*/*undo-info*˵ϿƤundoǤä
;;; ꥹȤ֤
;;;
(defun reverse-do (info-list)
  (let (cmds result)
    (setf result nil)
    (loop
      (when (null info-list) (return result))
      (setf cmds (pop info-list))
      (push (mapcar #'pure-reverse-do cmds) result))))

(defun pure-reverse-do (cmd)
  (let (point mark insert-or-kill char-list)
    (when (null cmd) (return-from pure-reverse-do nil))
    (setf point (nth 0 cmd))
    (setf mark (nth 1 cmd))
    (setf insert-or-kill (nth 0 (nth 2 cmd)))
    (setf char-list (nth 1 (nth 2 cmd)))
    (cond
     ((equal insert-or-kill '-)
      (list point mark (cons '+ (list char-list))))
     ((equal insert-or-kill '--)
      (list point mark (cons '+ (list char-list))))
     ((equal insert-or-kill '+)
      (list point mark (cons '-- (list char-list))))
     (t (line-edit-break (format nil "break at (pure-reverse-do ~a)" cmd))))))

;;;
;;; ؿ search֤̤ƬǰפƤ뤫ɤĴ٤롣
;;;
(defun exactly-matched (p)
  (and (numberp p) (zerop p)))

;;;
;;; (line-size)बʪԤΥɥȤʤꡢݻƤ
;;; ƥȤ (visual)(visual-end)ǶޤϰϤɽ롣
;;;
;;;        0----+----+----+----+----+----+----+----+----+----+-
;;;					       	        |
;;;					           (line-size)
;;; |----+----+----+----+----+----+----+----+----+----+----+---+--|
;;;        |              |                        |
;;;     (visual)       *point*                (visual-end)
;;;

;;;
;;; ʪԤηꤹ롣
;;; ǽΥ򥼥Ȥ롣
;;;
(defun line-size (&optional (n 72 sw))
  (cond
   ((null sw) *line-size*)
   ((numberp n) (setf *line-size* n))))

;;;
;;; Ļ벽Ƭݥȡ
;;;
(defun visual ()
  (let (base)
    (setf base (current-line-base))
    (min (+ base *hscroll-size*) (+ base (line-length)))))

;;;
;;; Ļ벽ϰϤνü
;;;
(defun visual-end ()
  (let (np)
    (cond
     ((setf np (next-newline-position))
      (min (+ (visual) (line-size)) (1+ np)))
     (t (min (+ (visual) (line-size)) (text-length))))))

;;;
;;; ݥȰְʹߤ#\Newlineΰ֤֤
;;; ⤷ݥȰְʹߤ#\Newline¸ߤʤ nil֤
;;;
(defun next-newline-position ()
  (position #\Newline *text* :test #'char= :start *point*))

;;;
;;; ݥȰְ#\Newlineΰ֤֤
;;; ⤷ݥȰְ#\Newline¸ߤʤ nil֤
;;;
(defun previous-newline-position ()
  (position #\Newline *text* :from-end t :test #'char= :end *point*))

;;;
;;; ʿԤݤΥǥեʸꤹ롣
;;; ꤷʤ֤ͤ
;;; ǥեͤ *line-size*20%
;;;
(defun hscroll-unit (&optional (size nil sw))
  (when (null *hscroll-unit*)
    (setf *hscroll-unit* (floor (* (line-size) 20/100))))
  (cond
   ((null sw) *hscroll-unit*)
   (t (setf *hscroll-unit* size))))

;;;
;;; ǻꤵ줿֤˥ݥȤɽ롣 nilǤ
;;; ǽʸ¤긽ߤɽϰϤѹ˥ݥȰ֤˥ɽ롣
;;; ɬפǤвԤ֤֤
;;;
(defun sync-point-and-cursor (&optional col)
  (sync-line)
  (cond
   ((not (numberp col))
    (if (numberp (auto-scroll-offset))
	(sync-simple (auto-scroll-offset))
      (sync-max)))
   (t (sync-point col)))
  (if (need-refresh-line-p) (display-line))
  (move-cursor-to (- *point* (visual)))
  (finish-output)
  (return-from sync-point-and-cursor (cursor-position)))

;;;
;;; #\NewlineޤǰưݤɽĴ᤹롣
;;;
(defun sync-line ()
  (let (pos)
    (setf *newline-position* (previous-newline-position))
    (setf pos *point*)
    (cond
     ((equal *newline-position* *last-newline-position*) nil)
     ((empty-text-p) nil)
     ((beginning-of-line-p)
      (sync-point 0))
     ((end-of-line-p)
      (sync-point (cdr (display-range))))
     ((< *point* (visual-end))
      (beginning-of-line 1)
      (sync-point 0))
     ((>= *point* (visual-end))
      (sync-max)))
    (move-point-to pos)
    (need-refresh-line t)
    (setf *last-newline-position* *newline-position*)))

;;;
;;; ǻꤵ줿֤˥ݥȤɽ롣
;;; ɬפǤвԤ
;;;
(defun sync-point (col)
  (declare (fixnum col))
  (cond
   ((< col 0)
    (setf col 0))
   ((> col (max-column))
    (setf col (max-column))))
  (cond
   ((empty-text-p) nil)
   ((end-of-text-p)
    (reset-scroll)
    (scroll-left *point*)
    (scroll-right col))
   ((< *point* (visual))
    (scroll-right (+ col (- (visual) *point*))))
   ((>= *point* (visual-end))
    (scroll-left (+ (- (line-size) col) (- *point* (visual-end)))))
   ((and (<= (visual) *point*) (< *point* (visual-end)))
    (cond
     ((> (- *point* (visual)) col)
      (reset-scroll)
      (scroll-left (- (- *point* (visual)) col)))
     ((< (- *point* (visual)) col)
      (reset-scroll)
      (scroll-right (- col (- *point* (visual)))))))))

;;;
;;; (visual)(visual-end)ϰϤ˥ݥȤɽ롣
;;; (auto-scroll-offset)ꤵ줿ʸʬ;ʬ
;;; 뤹롣
;;;
(defun sync-simple (offset)
  (declare (fixnum offset))
  (cond
   ((empty-text-p) nil)
   ((beginning-of-line-p)
    (sync-point 0))
   ((end-of-line-p)
    (sync-point (cdr (display-range))))
   ((< *point* (visual))
    (scroll-right (+ offset (- (visual) *point*))))
   ((>= *point* (visual-end))
    (scroll-left (1+ (+ offset (- *point* (visual-end))))))
   ((and (<= (visual) *point*) (< *point* (visual-end)))
    (do-nothing)))
  (need-refresh-line t))

;;;
;;; ȼԤν񤭴ǽʸ¤޻ߤ뤿ˡ
;;; ݥȤμϤʸɽ褦Ĵ᤹롣
;;;
(defun sync-max ()
  (let (pos)
    (setf pos *point*)
    (cond
     ((empty-text-p) nil)
     ((beginning-of-line-p)
      (sync-point 0))
     ((end-of-line-p)
      (sync-point (cdr (display-range))))
     ((< *point* (visual))
      (move-point-to (car (display-range)))
      (sync-point 0))
     ((>= *point* (visual-end))
      (move-point-to (car (display-range)))
      (sync-point 0))
     ((and (<= (visual) *point*) (< *point* (visual-end)))
      (do-nothing)))
    (need-refresh-line t)
    (move-point-to pos)))

;;;
;;; ݥȤμϤʸɽ뤿ɽ٤
;;; ݥϰϤβüȾü򥳥󥹡ڥȤ֤
;;;
(defun display-range ()
  (let (p n half rem range-low range-high low high low-rem high-rem)
    (setf half (floor (/ (line-size) 2)))
    (if (< (* half 2) (line-size))
	(setf rem 1)
      (setf rem 0))
    (setf p (previous-newline-position))
    (setf n (next-newline-position))
    (if (null p) (setf p 0) (incf p))
    (if (null n) (setf n (text-length)))
    (setf range-low (- *point* half))
    (setf range-high (+ *point* half rem))
    (setf low (max p range-low))
    (setf high (min n range-high))
    (if (> p range-low) (setf low-rem (- p range-low)))
    (if (< n range-high) (setf high-rem (- range-high n)))
    (cond
     (low-rem
      (setf high (min n (+ high low-rem))))
     (high-rem
      (setf low (max p (- low high-rem)))))
    (return-from display-range (cons low high))))

;;; C-x <
;;; ƥȤ򺸤 nࡢ뤹롣
;;;
(defun scroll-left (&optional (n nil))
  (setf n (select-repeat-count n))
  (cond
   ((null n)
    (pure-scroll-left (hscroll-unit)))
   ((zerop n) nil)
   ((> n 0)
    (pure-scroll-left n))
   ((< n 0)
    (pure-scroll-right (- n)))))

(defun pure-scroll-left (n)
  (when (empty-text-p)
    (return-from pure-scroll-left nil))
  (setf *hscroll-size* (min (line-length) (+ n *hscroll-size*)))
  (if (< *point* (visual)) (move-point-to (visual)))
  (need-refresh-line t))
  
;;; C-x >
;;; ƥȤ򱦤 nࡢ뤹롣
;;;
(defun scroll-right (&optional (n nil))
  (setf n (select-repeat-count n))
  (cond
   ((null n)
    (pure-scroll-right (hscroll-unit)))
   ((zerop n) nil)
   ((> n 0)
    (pure-scroll-right n))
   ((< n 0)
    (pure-scroll-left (- n)))))

(defun pure-scroll-right (n)
  (when (empty-text-p)
    (return-from pure-scroll-right nil))
  (setf *hscroll-size* (max 0 (- *hscroll-size* n)))
  (if (> *point* (visual-end)) (move-point-to (visual-end)))
  (need-refresh-line t))

;;;
;;; ̤򥼥᤹
;;;
(defun reset-scroll ()
  (when (> *hscroll-size* 0)
    (setf *hscroll-size* 0)
    (need-refresh-line t)))

;;; C-v
;;; ƥȤ˥뤷ƼԤɽ롣
;;; Ԥ¸ߤʤв⤷ʤ
;;;
(defun scroll-up (&optional (n 1))
  (setf n (select-repeat-count n))
  (dotimes (i n) (pure-scroll-up)))

(defun pure-scroll-up ()
  (let (np cp)
    (setf cp (cursor-position))
    (setf np (next-newline-position))
    (when (and np (>= np *point*))
      (move-point-to np)
      (if (not (end-of-text-p)) (forward-char 1))
      (sync-line)
      (setf cp (min cp (1- (- (visual-end) (visual)))))
      (move-point-right cp)
      (need-refresh-line t))))

;;; M-v
;;; ƥȤ˥뤷ľԤɽ롣
;;; ľԤ¸ߤʤв⤷ʤ
;;;
(defun scroll-down (&optional (n 1))
  (setf n (select-repeat-count n))
  (dotimes (i n) (pure-scroll-down)))

(defun pure-scroll-down ()
  (let (np cp)
    (if (beginning-of-text-p) (return-from pure-scroll-down nil))
    (setf cp (cursor-position))			;Cursor  Position
    (setf np (previous-newline-position))	;Newline Position
    (if (or (null np) (zerop np)) (return-from pure-scroll-down nil))
    (move-point-to (1- np))
    (beginning-of-line 1)
    (sync-line)
    (setf cp (min cp (1- (- (visual-end) (visual)))))
    (move-point-right cp)
    (need-refresh-line t)))

;;;
;;; ưΥեåͤꤹ롣
;;;
;;; ˥뤷Ƥ֤Ǻü鹹˺˰ưݤ
;;; ;ʬ˥뤹륫ꤹ롣
;;;
;;; 0 ޤ nilꤹ;ʬʥԤʤ 1
;;; ȤʤꡢͰʳ (t侩)ꤹȥݥȼϤ
;;; ǽʸ¤ʸɽǤ褦˥뤹롣
;;;
;;; ʤϸߤ֤
;;;
(defun auto-scroll-offset (&optional (num 4 sw))
  (cond
   ((null sw) *auto-scroll-offset*)
   ((null num) (setf *auto-scroll-offset* 0))
   (t (setf *auto-scroll-offset* num))))

;;;
;;; (visual) #\Newlineޤ(visual-end)ޤǤϰϤɽ롣
;;;
(defun display-line ()
  (clear-line)
  (dolist (ch (subseq *text* (visual) (visual-end))) (putch ch))
  (finish-output)
  (need-refresh-line nil))

;;;
;;; ϤɽΤߤä
;;;
(defun clear-line ()
  (move-cursor-to 0)
  (pure-delete-line))

;;; C-u
;;; Υޥ{command}{positive-number}¹Ԥ롣
;;;
;;; ˻ꤵ륳ޥɤ¹Ԥ٤֤
;;;
;;; 	C-u ['+'|'-'] [{positive-number}] {command}
;;;
;;; {positive-number}ά 4ꤵ줿ȸʤ
;;; '-'ꤵ줿 -1'+'
;;; 줿 +1ꤵ줿ȸʤ
;;;
;;; ͤǤ¹Բۤλ꤫Ū꤫̤
;;; ֤롣ͤnil ǤоάͤŬѤ줿
;;; Ȥ򼨤nil ʳǤŪ˻ꤵ줿ͤŬ
;;; 줿Ȥ򼨤
;;;
;;; {positive-number}ϤǤ褦ˤ뤿
;;; '/' {positive-number}ϽλޡȤƤ롣
;;; '/' {positive-number}Ϥ'/'2٥
;;; פ롣
;;;
(defun universal-argument ()	;'count' for dummy.
  (let (num sign exist-digit)
    (setf num 0)
    (setf sign nil)
    (setf exist-digit nil)
    (case (peek-ahead)
      (#\+	(getchar) (setf sign +1))
      (#\-	(getchar) (setf sign -1)))
    (loop
      (when (char= (peek-ahead) #\/)
	(getchar)
	(return))
      (if (not (digit-char-p (peek-ahead))) (return))
      (setf exist-digit t)
      (setf num (+ (* num 10) (char-to-digit (getchar)))))
    (cond
     ((and sign exist-digit)
      (setf num (* sign num)))
     ((and sign (not exist-digit))
      (setf num sign))
     ((and (null sign) exist-digit)
      (do-nothing))
     ((and (null sign) (not exist-digit))
      (setf num 4)))
    (multiply-repeat-count-for num)
    (values num (or sign exist-digit))))

(defun multiply-repeat-count-for (count)
  (cond
   ((not (numberp count)) nil)
   ((null (repeat-count)) (set-repeat-count-for count))
   (t (set-repeat-count-for (* *repeat-for* count)))))

(defun set-repeat-count-for (count)
  (setf *last-prefix-argument* *repeat-for*)
  (setf *last-repeat-count* *repeat-for*)
  (setf *repeat-for* count))

(defun reset-repeat-count ()
  (setf *repeat-for* nil))

(defun repeat-count () *repeat-for*)

(defun last-repeat-count () *last-repeat-count*)

(defun last-command () *last-command*)

;;;
;;; Ǹ˼¹ԤؿϿƤ
;;;
(defun set-last-command (cmd)
  (when (not (member cmd *last-command-inhibit-list*))
    (setf *last-command* cmd)))

;;;
;;; vi ϥǥ redoޥɤʤ *last-command*ؤ
;;; оݤȤƤϤʤؿϿ롣ͤnil
;;;
(defun set-last-command-inhibit-list (&optional lst)
  (cond
   ((null lst)
    *last-command-inhibit-list*)
   ((atom lst)
    (set-last-command-inhibit-list (list lst)))
   (t
    (setf *last-command-inhibit-list* (append *last-command-inhibit-list* lst)))))

(defun last-char () *last-char*)

;;;
;;; ɤߡͤȤ֤
;;;
(defun read-number ()
  (let (num)
    (setf num (char-to-digit (last-char)))
    (loop
      (if (not (digit-char-p (peek-ahead))) (return-from read-number num))
      (setf num (+ (* num 10) (char-to-digit (getchar)))))))

;;;
;;; ͤѴ¿ʹؿǥեͤϤʤ
;;;
;;; ͤѴͤ͡μ (+1, -1, nil)
;;; 軰ͤϿͤѴʤäĤΰ
;;;
;;; + ޤ -Τߤꤵ줿 nil֤ͤ
;;; μ̤֤
;;;
;;; (string-to-number "123") ==> 123 nil ""
;;; (string-to-number "  + 123") ==> 123 1 ""
;;; (string-to-number "-123") ==> -123 -1 ""
;;; (string-to-number " -  123 abc ") ==> -123 -1 " abc "
;;;
;;; (string-to-number "+1") ==> 1 1 ""
;;; (string-to-number " +") ==> nil 1 ""
;;; (string-to-number "-1") ==> -1 -1 ""
;;; (string-to-number " -") ==> nil -1 ""
;;;
;;; (string-to-number "+0") ==> 0 1 ""
;;; (string-to-number " 0") ==> 0 nil ""
;;; (string-to-number "-0") ==> 0 -1 ""
;;;
;;; (string-to-number "")   ==> nil nil ""
;;; (string-to-number " ?") ==> nil nil "?"
;;;
(defun string-to-number (str)
  (let (lst num sign exist-digit)
    (when (not (stringp str))
      (return-from string-to-number (values nil nil "")))
    (setf lst (unpack str))
    (setf num 0)
    (setf sign nil)
    (setf exist-digit nil)
    (loop
      (if (null lst) (return-from string-to-number (values nil nil "")))
      (if (not (member (car lst) *white-space-char*)) (return))
      (pop lst))
    (case (car lst)
      (#\+	(pop lst) (setf sign +1))
      (#\-	(pop lst) (setf sign -1))
      (otherwise nil))
    (loop
      (if (null lst) (return))
      (if (not (member (car lst) *white-space-char*)) (return))
      (pop lst))
    (loop
      (if (null lst) (return))
      (when (not (digit-char-p (car lst))) (return))
      (setf exist-digit t)
      (setf num (+ (* num 10) (char-to-digit (pop lst)))))
    (cond
     ((and sign exist-digit)
      (values (* sign num) sign (pack lst)))
     ((and sign (not exist-digit))
      (values nil sign (pack lst)))
     ((and (null sign) exist-digit)
      (values num nil (pack lst)))
     ((and (null sign) (not exist-digit))
      (values nil nil (pack lst))))))

;;;
;;; ͤѴ¿ʹؿǥեͤġ
;;;
;;; ͤѴͤ͡(+1,-1,nil)
;;; 軰ͤϿͤѴʤäĤΰ
;;;
;;; '+' ޤ '-'Τߤοξϡ줾 +1 -1
;;; ֤
;;;
;;; ͤ¸ߤʤäϥɰ default
;;; ǻꤵ줿 (ǥեͤ 4) ֤ͤ
;;; nil ֤
;;;
;;; ˤǥեͤŪ˻ꤵ줿Τ¸
;;; ߤʤä˥ǥե֤ͤ줿Τ̤Ǥ
;;; 롣
;;;
(defun meta-string-to-number (str &key (default 4))
  (let (num sign rest)
    (multiple-value-setq (num sign rest) (string-to-number str))
    (cond
     ((and sign num)
      (values num sign rest))
     ((and sign (null num))
      (values sign sign rest))
     ((and (null sign) num)
      (values num sign rest))
     ((and (null sign) (null num))
      (values default nil rest)))))

;;; C-x (
;;; ܡɡޥ򳫻Ϥ롣
;;; ȿꤵ줿ϸߤΥܡɡޥ
;;; ¹Ԥ˥ܡɤϤ륳ޥɤ
;;; ä롣
;;;
(defun start-kbd-macro ()
  (cond
   ((repeat-count)
    (pop *keyboard-macro*)	;removes (repeat-count)
    (call-last-kbd-macro 1))
   (t (setf *keyboard-macro* nil)))
  (setf *record-macro* t))

(defun push-keyboard-macro (s)
  (push s *keyboard-macro*))

;;; C-x )
;;; ܡɡޥλ롣
;;; ȿꤵλȤȤ˻ꤷ
;;; ܡɡޥ¹Ԥ롣ƤȤ
;;; ʼ¹ԤƤΤǡˣܤμ¹Ԥȿ롣
;;;
(defun end-kbd-macro ()
  (let (n)
    (setf n (repeat-count))
    (setf *record-macro* nil)
    (cond
     ((null n) nil)
     ((zerop n)
      (call-last-kbd-macro 0))
     ((> (1- n) 0)
      (call-last-kbd-macro (1- n))))))

;;; C-x e
;;; ǿΥޥ¹Ԥ롣
;;;
(defun call-last-kbd-macro (&optional (n nil))
  (setf n (select-repeat-count n))
  (set-repeat-count-for nil)
  (setf *macro-mode* t)
  (cond
   ((null n)
    (true-call-last-kbd-macro 1))
   ((zerop n)
    (call-last-kbd-macro-forever))
   ((> n 0)
    (true-call-last-kbd-macro n)))
  (setf *macro-mode* nil))

(defun macro-mode-p () *macro-mode*)

;;;
;;; 顼뤫 C-gϤޤǥޥ̵¤˷֤
;;;
(defun call-last-kbd-macro-forever ()
  (let (ch)
    (catch 'quit-macro
      (loop
	(dolist (cmd (last-kbd-macro))
	  (setf ch (read-char-no-hang))
	  (when (and ch (char= ch ctrl-g))
	    (throw 'quit-macro nil))
	  (complex-do-command cmd)
	  (sync-point-and-cursor))))))

;;;
;;; ޥ֤
;;;
(defun true-call-last-kbd-macro (n)
  (catch 'quit-macro
    (dotimes (i n)
      (catch 'abort-macro
	(dolist (cmd (last-kbd-macro))
	  (complex-do-command cmd)
	  (sync-point-and-cursor))))))

;;; C-x q
;;; ޥ˳ǧɲä롣
;;;
(defun kbd-macro-query ()
  (let (ch)
    (when (not *record-macro*)
      (sync-point-and-cursor)
      (xsleep (kbd-macro-query-time))
      (loop
	(with-prompt "Proceed with macro? (y/n/q, or C-l)"
	 (setf ch (getchar)))
	(cond
	 ((member ch '(#\y #\Space) :test #'char=)
	  (return-from kbd-macro-query nil))
	 ((member ch '(#\n #\Rubout) :test #'char=)
	  (throw 'abort-macro nil))
	 ((member ch '(#\q #\Return) :test #'char=)
	  (throw 'quit-macro nil))
	 ((char= ch #\^L)
	  (need-refresh-line t)
	  (sync-point-and-cursor)
	  (xsleep (kbd-macro-query-time))))))))

;;;
;;; C-x qǽ³뤫ɤ䤤碌ɽԤ
;;; ƥȤôɽƤꤹ롣
;;;
;;; ʤϸߤ֤ͤ
;;;
(defun kbd-macro-query-time (&optional (tm 2 sw))
  (cond
   ((null sw) *kbd-macro-query-time*)
   ((numberp tm) (setf *kbd-macro-query-time* tm))))

;;; C-x C-k
;;; ǿΥޥե˽񤭽ФƤԽ롣
;;;
(defun edit-kbd-macro ()
  (save-macro)
  (ed (macro-file))
  (load-macro))

;;;
;;; ǿΥޥƤ֤
;;;
(defun last-kbd-macro ()
  (reverse *keyboard-macro*))

;;;
;;; ǿΥޥե¸롣
;;;
(defun save-macro (&optional (fname (macro-file)))
  (with-open-file (s fname :direction :output)
    (prin1 (last-kbd-macro) s)
    (terpri s)
    (return-from save-macro fname)))

;;;
;;; ꤷե뤫ޥɤ롣
;;;
(defun load-macro (&optional (fname (macro-file)))
  (with-open-file (s fname :direction :input)
    (let (tmp)
      (setf tmp (read s nil *eos*))
      (when (not (eq tmp *eos*))
	(setf *keyboard-macro* (reverse tmp))))))

;;;
;;; ޥ¸ɤǥեȤΥե̾ꤹ롣
;;; ʤиߤե֤̾
;;;
(defun macro-file (&optional (fn *macro-file* sw))
  (cond
   ((null sw) *macro-file*)
   ((or (stringp fn) (pathnamep fn))
    (setf *macro-file* fn))))

;;;
;;; Ԥ򥹥å¸롣
;;;
(defun save-line ()
  (push *text* *line-stack*)
  (push *mark* *line-stack*)
  (push *ncl* *line-stack*)
  (push *point* *line-stack*))

;;;
;;; Ԥ򥹥å롣
;;;
(defun restore-line ()
  (setf *point* (pop *line-stack*))
  (setf *ncl* (pop *line-stack*))
  (setf *mark* (pop *line-stack*))
  (setf *text* (pop *line-stack*)))

;;;
;;; ޡ֤Ĵ뤿ɬפʾ¸롣
;;;
(defun save-mark-info ()
  (push *point* *save-mark-info*)
  (push *mark* *save-mark-info*)
  (push *ncl* *save-mark-info*))

;;;
;;; 롿˥ޡ֤Ĵ롣
;;;
(defun align-mark ()
  (setf *mark* (pure-align-mark)))

(defun pure-align-mark ()
  (let (point mark len killed)
    (setf len (pop *save-mark-info*))
    (setf mark (pop *save-mark-info*))
    (setf point (pop *save-mark-info*))
    (setf killed (abs (- len *ncl*)))
    (if (null mark) (return-from pure-align-mark nil))
    (if (empty-text-p) (return-from pure-align-mark nil))
    (if (< *point* point)
      (cond	;backward-kill
       ((<= mark (- point killed)) mark)
       ((<= point mark) (- mark killed))
       (t nil))
      (cond	;forward-kill
       ((<= mark point) mark)
       ((>= mark (+ point killed)) (- mark killed))
       (t nil)))))

;;;
;;; ŪʽҸȴؿ
;;;

;;;
;;; ʸफݤ֤
;;;
(defun white-space-p (ch)
  (declare (character ch))
  (member ch *white-space-char*))

;;;
;;; ڤʸफݤ֤
;;;
(defun break-char-p (ch)
  (declare (character ch))
  (member ch *break-char*))

;;;
;;; ʤζڤʸǶڤʸ롣
;;;
(defun narrow-break-char ()
  (setf *break-char* *narrow-break-char*))

;;;
;;; ¿ζڤʸǶڤʸ롣
;;;
(defun wide-break-char ()
  (setf *break-char* *wide-break-char*))

;;;
;;; ǻꤹʸڤʸɲä롣
;;;
(defun append-break-char (&optional lst)
  (cond
   ((null lst) *break-char*)
   ((atom lst)
    (if (not (member lst *break-char*)) (push lst *break-char*)))
   ((listp lst)
    (if (not (member (car lst) *break-char*)) (push (car lst) *break-char*))
    (append-break-char (cdr lst)))))

;;;
;;; ǻꤹʸڤʸफ롣
;;;
(defun remove-break-char (&optional lst)
  (cond
   ((null lst) *break-char*)
   ((atom lst)
    (remove lst *break-char*))
   ((listp lst)
    (remove (car lst) (remove-break-char (cdr lst))))))

;;;
;;; ʸफݤ֤
;;;
(defun normal-char-p (ch)
  (declare (character ch))
  (not (or (white-space-p ch) (break-char-p ch))))

;;;
;;; ƼҸ
;;;
(defun start-kbd-macro-p (cmd) (eq cmd 'start-kbd-macro))
(defun end-kbd-macro-p (cmd) (eq cmd 'end-kbd-macro))
(defun end-input-p (cmd) (eq cmd 'end-input))
(defun advertised-undo-command-p (cmd) (eq cmd 'advertised-undo))
(defun advertised-redo-command-p (cmd) (eq cmd 'advertised-redo))
(defun universal-argument-command-p (cmd) (eq cmd 'universal-argument))
(defun allow-yank-pop-commands-p (cmd) (member cmd *allow-yank-pop-commands*))
(defun kill-commands-p (cmd) (member cmd *kill-commands*))
(defun prefix-commands-p (cmd) (member cmd *prefix-commands*))
(defun call-by-keyboard-p () *call-by-keyboard*)

;;;
;;; ɲåޥɤ򥰥롼פɲä뤿Υޥɷ
;;;
(defun add-allow-yank-pop-commands (&optional lst)
  (cond
   ((null lst) *allow-yank-pop-commands*)
   ((atom lst) (add-allow-yank-pop-commands (list lst)))
   (t (setf *allow-yank-pop-commands* (append *allow-yank-pop-commands* lst)))))

(defun add-kill-commands (&optional lst)
  (cond
   ((null lst) *kill-commands*)
   ((atom lst) (add-kill-command (list lst)))
   (t (setf *kill-commands* (append *kill-commands* lst)))))

(defun add-prefix-commands (&optional lst)
  (cond
   ((null lst) *prefix-commands*)
   ((atom lst) (add-prefix-command (list lst)))
   (t (setf *prefix-commands* (append *prefix-commands* lst)))))

;;; M-C-c
;;; ΥѥåǥХåݤ˻Ѥbreakؿ
;;;
(defun line-edit-break (&optional (msg "line-edit-break") (quiet nil))
  (let (pkg)
    (cooked-mode)
    (setf pkg *package*)
    (in-package 'line-edit-pkg)
    (format t "~a~%" msg)
    (if (null quiet) (debug-print))
    (unwind-protect
	(break)
      (setf *package* pkg)
      (raw-mode)
      (init-line-edit nil)
      (finish-output))))

(defun debug-print ()
  (dprt)
  (format t "(text-length) = ~d~%" (text-length))
  (format t "*text* = ~s~%" *text*)
  (format t "inverted *text* = ~s~%" (invert-list *text*))
  (format t "packed text = ~s~%" (pack *text*))
  (format t "inverted packed text = ~s~%" (pack (invert-list *text*)))
  (format t "(line-size) = ~d~%" (line-size))
  (format t "(line-length) = ~d~%" (line-length))
  (format t "(current-line) = ~s~%" (current-line))
  (format t "(last-command) = ~s~%" (last-command))
  (format t "*undo-control* = ~s~%" *undo-control*)
  (format t "(length *undo-info*) = ~s~%" (length *undo-info*))
  (format t "*undo-info* = ~s~%" *undo-info*)
  (format t "(length *redo-info*) = ~s~%" (length *redo-info*))
  (format t "*redo-info* = ~s~%" *redo-info*)
  (format t "(hscroll-unit) = ~d~%" (hscroll-unit))
  (format t "*hscroll-size* = ~d~%" *hscroll-size*)
  (format t "*newline-position* = ~s~%" *newline-position*)
  (format t "*last-newline-position* = ~s~%" *last-newline-position*)
  (format t "*need-refresh-line* = ~d~%" *need-refresh-line*))

(defun dprt ()
  (format t "*point* = ~d~%" *point*)
  (format t "*mark* = ~d~%" *mark*)
  (format t "*cursor-position* = ~d~%" *cursor-position*)
  (format t "(visual) = ~d~%" (visual))
  (format t "(visual-end) = ~d~%" (visual-end))
  (format t "current display =~%")
  (display-line)
  (terpri))

;;;
;;; ޥɤμ¹ꥹȤ֤
;;;	ޥɼ¹εϿϤ	M-C-s (start)
;;;	ޥɼ¹εϿλ	M-C-e (end)
;;; ޥɼ¹ꥹȤˤ
;;;	(֤ޥ ƥ ݥ ޡ *undo-info* *redo-info*)
;;; ʤꥹȤս(=Τۤ)˵ϿƤ롣
;;;

;;;
;;; M-C-s
;;;
(defun start-command-trace ()
  (set-command-trace 0)
  (set-command-trace-list
   (list '("\#num" "repeat-count" "Command" "Text" "point" "mark" "undo-info" "redo-info")))
  (return-from start-command-trace t))

;;;
;;; M-C-e
;;;
(defun end-command-trace ()
  (push-command-trace-list
   (list (command-trace) (repeat-count) (packed-text)
	 (point) (mark) (undo-info) (redo-info)))
  (set-command-trace nil))

(defun trace-list (&optional help)
  (cond
   ((null help)
    (cdr (reverse *command-trace-list*)))
   (t (reverse *command-trace-list*))))

(defun command-trace () *command-trace*)

(defun set-command-trace (s)
  (setf *command-trace* s))

(defun command-trace-list () *command-trace-list*)

(defun set-command-trace-list (s)
  (setf *command-trace-list* s))

(defun push-command-trace-list (s)
  (push s *command-trace-list*))

;;;
;;; global-set-key ˡΥޥʸʸפѴ롣
;;;
;;;	\C-x	=>	*ctrl-x*
;;;	\M-x	=>	*meta-code* x
;;;	_	=>	#\space
;;;	\_	=>	_
;;;	\x	=>	x
;;;	x	=>	x
;;;	|str|	=>	str
;;;	[str]	=>	(defined-key-list) ˽äʸѴ
;;;
(defun batch-str (str)
  (let (lst word ch result tmp)
    (setf lst (unpack str))
    (setf result nil)
    (loop
      (if (null lst) (return))
      (setf lst (remove-preceding-white-spaces lst))
      (cond
       ((char= (first lst) *backslash-ch*)
	(cond
	 ((and				;\C-x or \M-x ?
	   (>= (length lst) 4)
	   (or
	    (equal (subseq lst 1 3) *meta-prefix*)
	    (equal (subseq lst 1 3) *control-prefix*)))
	  (pop lst)			;remove '\'
	  (cond
	   ((char= (first lst) (first *control-prefix*))
	    (pop lst)			;remove 'C'
	    (pop lst)			;remove '-'
	    (push (control-code (pop lst)) result))
	   ((char= (first lst) (first *meta-prefix*))
	    (pop lst)			;remove 'M'
	    (pop lst)			;remove '-'
	    (push *meta-code* result)
	    (push (pop lst) result))))
	 (t
	  (pop lst)			;remove '\'
	  (push (pop lst) result))))
       ((char= (first lst) *left-square-bracket-ch*)
	(setf word nil)
	(setf ch (pop lst))
	(push ch word)
	(loop
	  (if (null lst) (return))
	  (setf ch (pop lst))
	  (push ch word)
	  (if (char= ch *right-square-bracket-ch*) (return)))
	(setf tmp (pack (reverse word)))
	(setf result (append (reverse (unpack (get-def tmp))) result)))
       ((char= (first lst) *vertical-bar-ch*) ;for |string| form
	(pop lst)			;remove '|'
	(loop
	  (if (null lst) (return))
	  (if (char= (first lst) *vertical-bar-ch*) (return (pop lst)))
	  (push (pop lst) result)))
       ((char= (first lst) *underscore-ch*) ;'_' for visual space.
	(pop lst)
	(push #\space result))
       (t (push (pop lst) result))))
    (push #\return result)
    (pack (reverse result))))

(defun get-def (str)
  (cdr (true-get-def str (defined-key-list))))

(defun true-get-def (str lst)
  (cond
   ((null lst) nil)
   ((atom (car lst)) nil)
   ((not (stringp (caar lst))) nil)
   ((string= str (caar lst))
    (car lst))
   (t (true-get-def str (cdr lst)))))

;;;
;;; ʸ ch б륳ȥ롦ɤ֤
;;;
(defun control-code (ch)
  (case (char-upcase ch)
    (#\@ *ctrl-@*)
    (#\A *ctrl-a*)
    (#\B *ctrl-b*)
    (#\C *ctrl-c*)
    (#\D *ctrl-d*)
    (#\E *ctrl-e*)
    (#\F *ctrl-f*)
    (#\G *ctrl-g*)
    (#\H *ctrl-h*)
    (#\I *ctrl-i*)
    (#\J *ctrl-j*)
    (#\K *ctrl-k*)
    (#\L *ctrl-l*)
    (#\M *ctrl-m*)
    (#\N *ctrl-n*)
    (#\O *ctrl-o*)
    (#\P *ctrl-p*)
    (#\Q *ctrl-q*)
    (#\R *ctrl-r*)
    (#\S *ctrl-s*)
    (#\T *ctrl-t*)
    (#\U *ctrl-u*)
    (#\V *ctrl-v*)
    (#\W *ctrl-w*)
    (#\X *ctrl-x*)
    (#\Y *ctrl-y*)
    (#\Z *ctrl-z*)
    (#\\ *ctrl-\*)
    (#\] *ctrl-]*)
    (#\^ *ctrl-^*)
    (#\_ *ctrl-_*)
    (#\[ *ctrl-[*)
    (otherwise ch)))

;;;
;;; ΥѥåΥᥤؿ
;;;
;;;	init-text	ƥȡnilʤƥȡ
;;;	batch-cmd	ХåԽѥޥɡ
;;;	mode-name	ХåޥɤΥޥηϤ򼨤⡼̾
;;;	load-path	ޥեɤѥ̾
;;;
;;;	(line-edit "This is a pen." "\\C-u 3 \\M-f _long" "emacs-mode")
;;; and
;;;	(line-edit "This is a pen." "3w i long_" "vi-mode")
;;; both returns
;;;	"This is a long pen."
;;;
(defun line-edit (&optional init-text batch-cmd mode-name load-path)
  (let (tmp saved-mode-name saved-load-path)
    (when (stringp mode-name)
      (multiple-value-setq (saved-mode-name saved-load-path) (editor-mode))
      (set-editor-mode mode-name load-path))
    (setf tmp (true-line-edit init-text batch-cmd))
    (when (stringp mode-name)
      (set-editor-mode saved-mode-name saved-load-path))
    (return-from line-edit tmp)))

(defun true-line-edit (&optional init-text batch-cmd)
  (let (tmp)
    (init-line-edit init-text)
    (cond
     ((null batch-cmd)
      (line-edit-command-loop))
     ((stringp batch-cmd)
      (with-input-from-string
       (*standard-input* (batch-str batch-cmd))
       (with-output-to-string
	 (*standard-output*)
	 (setf tmp (line-edit-command-loop))))
      (return-from true-line-edit tmp))
     (t (line-edit-command-loop)))))

;;;
;;; Ԥ
;;;
(defun init-line-edit (init-text)
  (clear-input)
  (setf *point* 0)
  (setf *ncl* (length init-text))
  (setf *text* (unpack init-text))
  (setf *mark* nil)
  (setf *suspend-switch* t)
  (setf *record-macro* nil)
  (setf *macro-mode* nil)
  (need-refresh-line nil)
  (reset-scroll)
  (reset-undo-info))

;;;
;;; line-editλԤԤݤ椹롣
;;;
(defun line-edit-put-newline (&optional (use t sw))
  (cond
   ((null sw) *line-edit-put-newline*)
   (t (setf *line-edit-put-newline* use))))

;;;
;;; ѥޥɡ롼ס
;;; ޥϤꡢɽ򹹿롣
;;;
(defun line-edit-command-loop ()
  (let (cmd)
    (setf *echo* t)
    (need-refresh-line nil)
    (reset-scroll)
    (sync-point-and-cursor)
    (setf *newline-position* (previous-newline-position))
    (setf *last-newline-position* *newline-position*)
    (with-raw-input-mode
     (catch 'exit-line-edit
       (loop
	 (catch 'control-g
	   (setf *need-refresh-line* nil)
	   (set-last-command cmd)
	   (setf cmd (get-command))
	   (when (and *record-macro* (not (end-kbd-macro-p cmd)))
	     (push cmd *keyboard-macro*))
	   (when (end-input-p cmd)
	     (postlude-input)
	     (return-from line-edit-command-loop (packed-text)))
	   (complex-do-command cmd)
	   (sync-point-and-cursor)
	   (when (and
		  (blink-paren-p)
		  (not (end-of-text-p))
		  (not (blink-paren-just-inserted))
		  (member (current-char) *parenthesis-list*))
	     (jump-and-return-balanced-paren)
	     (sync-point-and-cursor))
	   (cond
	    ((not (kill-commands-p cmd))
	     (kill-ring-suspend t))
	    ((not (allow-yank-pop-commands-p cmd))
	     (setf *allow-yank-pop* nil)))))))))

;;;
;;; end-inputޥϸθԤ
;;;
(defun postlude-input ()
  (when (line-edit-put-newline) (writeln))
  (setf *cursor-position* 0)
  (search-from nil)
  (setf *mark* nil)
  (reset-undo-info)
  (reset-scroll))

;;;
;;; ޥɼ¹ԤܽƤʣ礷ƹԤ
;;;
(defun complex-do-command (cmd)
  (when (pre-do-command cmd)
    (do-command cmd)
    (post-do-command cmd)))

;;;
;;; Ϥ줿ؿǤСδؿ¹Ԥ
;;; ʸǤ*otherwise-function*ǻꤵ줿ؿ
;;; ¹Ԥ롣
;;;
(defun do-command (cmd)
  (setf *call-by-keyboard* t)
  (cond
   ((functionp cmd) (funcall cmd))
   ((stringp cmd)
    (self-insert-string cmd))
   ((characterp cmd)
    (cond
     ((not (ascii-char-p cmd))	;  ((not (graphic-char-p cmd))
      (self-insert-string (make-description cmd)))
     (t (funcall *otherwise-function* cmd))))
   (t (line-edit-break
       (format nil "do-command: function or character expected (~s)" cmd) t)))
  (setf *call-by-keyboard* nil))

;;;
;;; ޥɼ¹ԤνԤ
;;; ̤ nilǤʤ do-commandpost-do-command¹Ԥ롣
;;;
(defun pre-do-command (cmd)
  ;ǥХåѤΥȥ졼ꤵƤоϿ롣
  (when (command-trace)
    (push-command-trace-list
     (list (command-trace) (repeat-count) cmd (packed-text)
	   (point) (mark) (undo-info) (redo-info)))
    (set-command-trace (1+ (command-trace))))
  ;ְν򤹤롣
  (when (prefix-commands-p cmd)
    (when (not (macro-mode-p))
      (funcall cmd) ;ְꤹ롣
      (push-keyboard-macro (repeat-count)))
    (return-from pre-do-command nil))
  ;undoǤredoǤʤϡundoredoν롣
  (when (not
    (or
     (advertised-undo-command-p cmd)
     (advertised-redo-command-p cmd)))
    (set-undo-info (append (redo-info) (reverse-do (redo-info)) (undo-info)))
    (set-redo-info nil))
  (set-undo-control nil)
  (set-command-undo-info nil)
  (return-from pre-do-command t))

;;;
;;; undo/redoθԤ
;;;
(defun post-do-command (cmd)
  (when (command-undo-info)
    (cond
     ((equal (undo-control) 'undo)
      (push-redo-info (command-undo-info)))
     ((equal (undo-control) 'redo)
      (push-undo-info (command-undo-info)))
     (t (push-undo-info (command-undo-info)))))
  (if (not (numberp cmd)) (set-repeat-count-for nil)))

;;;
;;; ܡɤϤ¹Ԥ٤ؿ֤
;;; ƤʤϤ줿ϡפ
;;; ʤäʸ֤
;;;
(defun get-command ()
  (let (sym)
    (setf sym (true-get-command (getsym) (global-keymap)))
    (cond
     ((null sym)
      (clear-symbol-buffer)
      *last-char*)
     ((stringp sym)
      (setf sym (search-key sym (global-keymap)))
      (cond
       ((null sym)
	(audio-bell)
	(clear-symbol-buffer)
	(get-command))
       (t (car (second sym)))))
     (t sym))))

(defun true-get-command (key keymap)
  (let (p)
    (setf p (search-key key keymap))
    (cond
     ((null p) nil)
     ((null (cdr (second p)))
      (car (second p)))
     (t (true-get-command (getsym) (cdr p))))))

;;;
;;; ؿ getsymȴؿ clear-symbol-buffer˥ѿ
;;; *symbol-buffer*롣
;;;
(let ((*symbol-buffer* nil))
  (defun getsym ()	;ܡɤʸɤ߹ߡܥ֤
    (cond
     ((null *symbol-buffer*)
      (setf *symbol-buffer* (true-getsym))
      (pop *symbol-buffer*))
     (t (pop *symbol-buffer*))))

  (defun clear-symbol-buffer ()	;;; symbol buffer򥯥ꥢ롣
    (setf *symbol-buffer* nil)) )

(defun true-getsym ()
  (setf *last-char* (getchar))
  (when (and (inspect-keycode) (not (graphic-char-p *last-char*)))
    (format t "~a " (make-description *last-char*)))
  (case *last-char* 
    (#\^@		(list *control-string* *at-sign*))
    (#\^A		(list *control-string* *letter-a*))
    (#\^B		(list *control-string* *letter-b*))
    (#\^C		(list *control-string* *letter-c*))
    (#\^D		(list *control-string* *letter-d*))
    (#\^E		(list *control-string* *letter-e*))
    (#\^F		(list *control-string* *letter-f*))
    (#\^G		(list *control-string* *letter-g*))
;   (#\^H		(list *control-string* *letter-h*))
    (#\^I		(list *control-string* *letter-i*))
;   (#\^J		(list *control-string* *letter-j*))
    (#\^K		(list *control-string* *letter-k*))
    (#\^L		(list *control-string* *letter-l*))
;   (#\^M		(list *control-string* *letter-m*))
    (#\^N		(list *control-string* *letter-n*))
    (#\^O		(list *control-string* *letter-o*))
    (#\^P		(list *control-string* *letter-p*))
    (#\^Q		(list *control-string* *letter-q*))
    (#\^R		(list *control-string* *letter-r*))
    (#\^S		(list *control-string* *letter-s*))
    (#\^T		(list *control-string* *letter-t*))
    (#\^U		(list *control-string* *letter-u*))
    (#\^V		(list *control-string* *letter-v*))
    (#\^W		(list *control-string* *letter-w*))
    (#\^X		(list *control-string* *letter-x*))
    (#\^Y		(list *control-string* *letter-y*))
    (#\^Z		(list *control-string* *letter-z*))
    (#\^\\		(list *control-string* *backslash*))
    (#\^] 		(list *control-string* *right-square-bracket*))
    (#\^^ 		(list *control-string* *circumflex-accent*))
    (#\^_ 		(list *control-string* *underscore*))
    (#\^[		(list *meta-string*))
    (#\Rubout 		(list *rubout*))
    (#\Space		(list *space*))
    (#\Backspace	(list *backspace*))
    (#\Linefeed		(list *linefeed*))
    (#\Return		(list *return*))
    (otherwise		(list (gryph-to-description *last-char*)))))

;;;
;;; keymapǤcarsymǤǤ֤
;;;
(defun search-key (sym keymap)
  (cond
   ((atom (car keymap))
    (if (symbol= sym (car keymap)) keymap nil))
   (t (dolist (s keymap) (if (symbol= sym (car s)) (return-from search-key s))))))

(defun symbol= (p q) (and (stringp q) (string= p q)))

;;;
;;; ޥɤФ륭ȼ¹Ԥ٤ؿȤ
;;; ޥɼ¹Ԥɬפɽ¸롣
;;;
;;;	(global-set-key "\\C-f" 'forward-char) Ȥ
;;;		 C-f Ȥ
;;; 	(funcall 'forward-char)
;;;
;;; ɾ롣
;;; 1 nilꤹ
;;;
;;; 	(global-set-key nil 'do-nothing)
;;;
;;; (global-set-key ...)ǻꤵʤäФ
;;; ؿλȤʤ롣ǥեȤ self-insertǤ롣
;;;
(defun global-set-key (str func)
  (cond
   ((null str)
    (setf *otherwise-function* func))
   ((stringp str)
    (set-global-keymap (key-seq-to-normalized-symbol-seq str) func))
   (t (line-edit-break "global-set-key: string or nil expected" t))))

;;;
;;; global-set-keyΰ̷
;;;
(defun define-keymap (keymap str func)
  (set-keymap keymap (key-seq-to-nomalized-symbol-seq str) func))

;;;
;;; global-set-keyνԤ
;;;
(defun clear-global-set-key ()
  (setf *defined-key-list* nil)
  (setf *global-keymap* nil))

;;;
;;; ޥɡɽ֤
;;;
(defun global-keymap ()
  (cond
   ((atom (car *global-keymap*))
    (list *global-keymap*))
   (t *global-keymap*)))

;;;
;;; *global-keymap* global-set-keyɽѤ롣
;;;
(defun set-global-keymap (lst func)
  (set-keymap '*global-keymap* lst func))

;;;
;;; global-set-key¹Ԥ뤳ȤѤƤɽ
;;; ˥Ȥ롣¹ԤɬܤǤϤʤ
;;;
(defun finish-global-set-key ()
  (setf *global-keymap* (sort-keymap *global-keymap*))
  (if (null *global-keymap*) nil t))

;;;
;;; 󥹤ؿ
;;;
;;; 	(define-key "[delete]" "\\M-\\[3\~")
;;;
(defun define-key (sym str)
  (push (cons sym str) *defined-key-list*)
  (global-set-key str sym))

;;;
;;; define-key 줿ܥȥεϿ֤
;;;
(defun defined-key-list () *defined-key-list*)

;;;
;;; 'keymap'ɽѤ롣
;;;
(defun set-keymap (keymap lst func)
  (set keymap
	(condense (eval keymap) (make-trans-table (append lst (list func))))))

;;;
;;; ޥɡܥΥꥹȤȼ¹Ԥ٤ؿ̾ʤꥹȤ
;;; ޥñΤɽ֤
;;;
(defun make-trans-table (lst)
  (cond
   ((null lst) nil)
   ((atom lst) (list lst))
   ((null (cdr lst)) lst)
   (t (cons (car lst) (list (make-trans-table (cdr lst)))))))

;;;
;;; ĤɽĹΤʤñɽ˹֤
;;;
(defun condense (p q)
  (cond
   ((atom p) q)
   ((atom q) p)
   ((atom (car p))
    (more-simple-condense p q))
   ((atom (car q))
    (more-simple-condense q p))
   ((and
     (null (cdr p))
     (null (cdr q)))
    (condense (car p) (car q)))
   ((null (cdr p))
    (condense (car p) q))
   ((null (cdr q))
    (condense p (car q)))
   (t (cons (condense (car p) (car q)) (condense (cdr p) (cdr q))))))

;;;
;;; ĤɽǤȥǤνԤ
;;;
;;;	p ::=	(a (b (c)) (d (e)))
;;;	q ::=	((a (c (d))) (e (f (g (h)))) (i (j)))
;;;	r ::=	((v (c (d))) (w (e (h))))
;;;
;;;	(more-simple-condense p q)
;;;		=> ((a (b (c)) (d (e)) (c (d))) (e (f (g (h)))) (i (j)))
;;;
;;;	(more-simple-condense p r)
;;;		=> ((a	(b (c))	(d (e))) (v (c (d))) (w (e (h))))
;;;
(defun more-simple-condense (p q)
  (cond
   ((atom p) q)
   ((atom q) p)
   ((atom (car q))
    (most-simple-condense p q))
   ((null (cdr q))
    (more-simple-condense p (car q)))
   ((string= (car p) (caar q))
    (cons (most-simple-condense p (car q)) (cdr q)))
   (t (paste (car q) (more-simple-condense p (cdr q))))))

;;;
;;; ĤɽǤ˥ȥǤνԤ
;;;
;;; 	p ::=	(a (b (c)))
;;; 	q ::=	(a (f (c (d))))
;;;
;;;	(most-simple-condense p q)
;;;		=> (a (b (c)) (f (c (d))))
;;;
(defun most-simple-condense (p q)
  (cond
   ((atom p) q)
   ((atom q) p)
   ((string= (car p) (car q))
    (cons (car p) (condense (cdr p) (cdr q))))
   (t (list p q))))

;;;
;;; ޥʸ󤫤ޥʸΥꥹȤ롣
;;;
;;; (key-seq-to-normalized-symbol-seq "\\C-\\M-f \\M-p   r")
;;; 	=> ("meta" "control" "letter-f" "meta" "small-p" "small-r")
;;;
(defun key-seq-to-normalized-symbol-seq (str)
  (make-control-code (normalize-symbol-seq (key-seq-to-symbol-seq str))))

;;;
;;; ޥʸ򥳥ޥ̾ʸΥꥹȤѴ롣
;;;
;;; (key-seq-to-symbol-seq "\\C-\\M-f \\M-p  [delete]  r")
;;;	=> ("control" "meta" "small-f" "meta" "small-p" "[delete]" "small-r")
;;;
;;; '['  ']'ޤǤϰĤʸȤư
;;; äơ㤨 \\M-[delete] ("meta" "[delete]")Ѵ롣
;;; M-[  "\\M-\\[", M-]  "\\M-\\["ȵ
;;;
(defun key-seq-to-symbol-seq (str)
  (let (lst result ch word)
    (when (not (stringp str))
      (return-from key-seq-to-symbol-seq nil))
    (setf lst (unpack str))
    (setf result nil)
    (loop
      (if (null lst) (return))
      (setf lst (remove-preceding-white-spaces lst))
      (cond
       ((char= (first lst) *left-square-bracket-ch*)
	(setf word nil)
	(setf ch (pop lst))
	(push ch word)
	(loop
	  (if (null lst) (return))
	  (setf ch (pop lst))
	  (push ch word)
	  (if (char= ch *right-square-bracket-ch*) (return)))
	(push (concatenate 'string (reverse word)) result))
       ((and
	 (char= (first lst) *backslash-ch*) ; \C- or \M- ?
	 (>= (length lst) 3)
	 (or
	  (equal (subseq lst 1 3) *meta-prefix*)
	  (equal (subseq lst 1 3) *control-prefix*)))
	(pop lst)	; remove backslash.
	(cond
	 ((char= (first lst) (first *control-prefix*))
	  (setf lst (cdr lst))
	  (push *control-string* result))
	 ((char= (first lst) (first *meta-prefix*))
	  (setf lst (cdr lst))
	  (push *meta-string* result)) ))
       ((char= (first lst) *backslash-ch*)	; stand alone backslash.
	(pop lst)
	(when (not (null lst))
	  (push (gryph-to-description (first lst)) result)))
       (t (push (gryph-to-description (first lst)) result)))
      (pop lst))
    (reverse result)))

;;;
;;; Control MetaŽҤ "\\M-\\C-x"η줹롣
;;;
;;; ("control" "meta" "small-f" "meta" "small-p" "small-r")
;;; 	=> ("meta" "control" "small-f" "meta" "small-p" "small-r")
;;;
(defun normalize-symbol-seq (lst)
  (cond
   ((null lst) nil)
   ((atom lst) lst)
   ((null (cdr lst)) lst)
   ((and
     (string= (first lst) *control-string*)
     (string= (second lst) *meta-string*))
    (append (list *meta-string* *control-string*) (cddr lst)))
   (t (cons (car lst) (normalize-symbol-seq (cdr lst))))))

;;;
;;; ȥ롦ɤȤ߹碌ѻ줹롣
;;;
;;; ("meta" "control" "capital-f" "meta" "small-p" "small-r")
;;; 	=> ("meta" "control" "letter-f" "meta" "small-p" "small-r")
;;;
(defun make-control-code (lst)
  (cond
   ((null lst) nil)
   ((atom lst) lst)
   ((null (cdr lst)) lst)
   ((string= (car lst) *control-string*)
    (append (unified-control-symbol (cadr lst)) (make-control-code (cddr lst))))
   (t (cons (car lst) (make-control-code (cdr lst))))))

;;;
;;; ȥ롦ɤȤ߹碌ѻ羮ʸΤʤʸ줷
;;; ȥ롦ɡܥ֤
;;;
(defun unified-control-symbol (sym)
  (list *control-string* (unified-symbol sym)))

;;;
;;; ʸʸ줹롣
;;;
(defun unified-symbol (sym)
  (cond
    ((or (string= sym *capital-A*) (string= sym *small-a*))	*letter-a*)
    ((or (string= sym *capital-B*) (string= sym *small-b*))	*letter-b*)
    ((or (string= sym *capital-C*) (string= sym *small-c*))	*letter-c*)
    ((or (string= sym *capital-D*) (string= sym *small-d*))	*letter-d*)
    ((or (string= sym *capital-E*) (string= sym *small-e*))	*letter-e*)
    ((or (string= sym *capital-F*) (string= sym *small-f*))	*letter-f*)
    ((or (string= sym *capital-G*) (string= sym *small-g*))	*letter-g*)
    ((or (string= sym *capital-H*) (string= sym *small-h*))	*letter-h*)
    ((or (string= sym *capital-I*) (string= sym *small-i*))	*letter-i*)
    ((or (string= sym *capital-J*) (string= sym *small-j*))	*letter-j*)
    ((or (string= sym *capital-K*) (string= sym *small-k*))	*letter-k*)
    ((or (string= sym *capital-L*) (string= sym *small-l*))	*letter-l*)
    ((or (string= sym *capital-M*) (string= sym *small-m*))	*letter-m*)
    ((or (string= sym *capital-N*) (string= sym *small-n*))	*letter-n*)
    ((or (string= sym *capital-O*) (string= sym *small-o*))	*letter-o*)
    ((or (string= sym *capital-P*) (string= sym *small-p*))	*letter-p*)
    ((or (string= sym *capital-Q*) (string= sym *small-q*))	*letter-q*)
    ((or (string= sym *capital-R*) (string= sym *small-r*))	*letter-r*)
    ((or (string= sym *capital-S*) (string= sym *small-s*))	*letter-s*)
    ((or (string= sym *capital-T*) (string= sym *small-t*))	*letter-t*)
    ((or (string= sym *capital-U*) (string= sym *small-u*))	*letter-u*)
    ((or (string= sym *capital-V*) (string= sym *small-v*))	*letter-v*)
    ((or (string= sym *capital-W*) (string= sym *small-w*))	*letter-w*)
    ((or (string= sym *capital-X*) (string= sym *small-x*))	*letter-x*)
    ((or (string= sym *capital-Y*) (string= sym *small-y*))	*letter-y*)
    ((or (string= sym *capital-Z*) (string= sym *small-z*))	*letter-z*)
    (t	sym)))

;;;
;;; #\Space #\~ ޤǤο޷ʸФ description֤
;;; ޷ʸϤʸɤɽ "[0xa0]" Ȥ
;;; 216ʿɽ줿ʸ֤ʾʸˡ
;;;
(defun gryph-to-description (ch)
  (declare (character ch))
  (case ch
    (#\a	*small-a*)
    (#\b	*small-b*)
    (#\c	*small-c*)
    (#\d	*small-d*)
    (#\e	*small-e*)
    (#\f	*small-f*)
    (#\g	*small-g*)
    (#\h	*small-h*)
    (#\i	*small-i*)
    (#\j	*small-j*)
    (#\k	*small-k*)
    (#\l	*small-l*)
    (#\m	*small-m*)
    (#\n	*small-n*)
    (#\o	*small-o*)
    (#\p	*small-p*)
    (#\q	*small-q*)
    (#\r	*small-r*)
    (#\s	*small-s*)
    (#\t	*small-t*)
    (#\u	*small-u*)
    (#\v	*small-v*)
    (#\w	*small-w*)
    (#\x	*small-x*)
    (#\y	*small-y*)
    (#\z	*small-z*)
    (#\A	*capital-A*)
    (#\B	*capital-B*)
    (#\C	*capital-C*)
    (#\D	*capital-D*)
    (#\E	*capital-E*)
    (#\F	*capital-F*)
    (#\G	*capital-G*)
    (#\H	*capital-H*)
    (#\I	*capital-I*)
    (#\J	*capital-J*)
    (#\K	*capital-K*)
    (#\L	*capital-L*)
    (#\M	*capital-M*)
    (#\N	*capital-N*)
    (#\O	*capital-O*)
    (#\P	*capital-P*)
    (#\Q	*capital-Q*)
    (#\R	*capital-R*)
    (#\S	*capital-S*)
    (#\T	*capital-T*)
    (#\U	*capital-U*)
    (#\V	*capital-V*)
    (#\W	*capital-W*)
    (#\X	*capital-X*)
    (#\Y	*capital-Y*)
    (#\Z	*capital-Z*)
    (#\0	*digit-0*)
    (#\1	*digit-1*)
    (#\2	*digit-2*)
    (#\3	*digit-3*)
    (#\4	*digit-4*)
    (#\5	*digit-5*)
    (#\6	*digit-6*)
    (#\7	*digit-7*)
    (#\8	*digit-8*)
    (#\9	*digit-9*)
    (#\+	*plus-sign*)
    (#\<	*less-than-sign*)
    (#\=	*equals-sign*)
    (#\>	*greater-than-sign*)
    (#\$	*dollar-sign*)
    (#\`	*backquote*)
    (#\^	*circumflex-accent*)
    (#\~	*tilde*)
    (#\#	*sharp-sign*)
    (#\%	*percent-sign*)
    (#\&	*ampersand*)
    (#\*	*asterisk*)
    (#\@	*at-sign*)
    (#\[	*left-square-bracket*)
    (#\\	*backslash*)
    (#\]	*right-square-bracket*)
    (#\{	*left-brace*)
    (#\|	*vertical-bar*)
    (#\}	*right-brace*)
    (#\!	*exclamation-mark*)
    (#\"	*double-quote*)
    (#\'	*single-quote*)
    (#\(	*left-parenthesis*)
    (#\)	*right-parenthesis*)
    (#\,	*comma*)
    (#\_	*underscore*)
    (#\-	*minus-sign*)
    (#\.	*period*)
    (#\/	*slash*)
    (#\:	*colon*)
    (#\;	*semicolon*)
    (#\?	*question-mark*)
    (#\Space	*space*)
    (otherwise	(make-description ch) )))

(defun make-description (ch)
  (declare (character ch))
  (string-downcase (format nil "[0x~2,'0x]" (char-code ch))))


(defun remove-preceding-white-spaces (lst)
  (loop
    (when (null lst) (return-from remove-preceding-white-spaces nil))
    (when (not (white-space-p (car lst)))
	(return-from remove-preceding-white-spaces lst))
    (pop lst)))

(defun control-prefix-p (lst)
  (and
   (char= (first lst) (first *control-prefix*))
   (char= (second lst) (second *control-prefix*))))

(defun meta-prefix-p (lst)
  (and
   (char= (first lst) (first *meta-prefix*))
   (char= (second lst) (second *meta-prefix*))))

(defun paste (p q)
  (cond
   ((atom (car q))
    (list p q))
   (t (cons p q))))

;;;
;;; ɽ򼭽˥Ȥ롣
;;;
(defun sort-keymap (keymap)
  (sort-inner-level (sort-first-level keymap)))

;;;
;;; ɽǥ٥򥽡Ȥ롣
;;;
(defun sort-first-level (keymap)
  (stable-sort keymap #'string-lessp :key #'car))

;;;
;;; ɽǤ٥򥽡Ȥ롣
;;;
(defun sort-inner-level (keymap)
  (cond
   ((null keymap) nil)
   ((atom keymap) keymap)
   ((null (cdr keymap)) keymap)
   ((null (cddar keymap))
    (cons (car keymap) (sort-inner-level (cdr keymap))))
   (t (cons
       (cons (caar keymap) (sort-inner-level (sort-first-level (cdar keymap))))
       (sort-inner-level (cdr keymap))))))

;;;
;;; Hook function for C-p and C-n
;;;
;;; ߤιԤƬȥݥȤδ֤ʸ/˸äõ
;;; ФΥҥȥԤɽ롣 ݥȤΰ֤ΥҥȥԤ
;;; ݥȰ֤Ʊ
;;;

#+disable-hook
(eval-when (eval load compile)
  (defun previous-line (&optional n) nil)
  (defun next-line (&optional n) nil)
  (defun beginning-of-history (&optional count) nil)
  (defun end-of-history (&optional count) nil)
  (defun goto-history (n) nil)
  (defun set-global-mark () nil)
  (defun goto-global-mark () nil))

#-disable-hook
(eval-when (eval load compile)

  ;;; C-p
  ;;; ոʸŤʤˤ˸롣
  ;;; Ǹ˸˰פ֤ϿŤо
  ;;; ȤƸ롣
  ;;;
  ;;; (history-pkg::history-search-backward str search-frm) ϸ
  ;;; ȡ˰פʸ
  ;;;
  ;;; ((ҥȥֹ . ֹ) Ƭ) ֤
  ;;;
  ;;; ߤλͤǤϡԤƬΰפΤߤоݤȤƤ뤿ḡ
  ;;; Ƭ֤Ͼ˥Ǥ롣
  ;;;
  (defun previous-line (&optional (n 1))
    (declare (fixnum n))
    (setf n (select-repeat-count n))
    (cond
     ((zerop n) nil)
     ((> n 0)
      (dotimes (i n) (true-previous-line)))
     ((< n 0)
      (next-line (- n)))))

  (defun true-previous-line ()
    (let (pos found lst str search-frm s n)
      (setf pos *point*)
      (setf search-frm (search-from))
      (when (oldest-history-p search-frm) (return-from true-previous-line nil))
      (setf lst (subseq *text* 0 pos))
      (multiple-value-setq (s n) (expand-history lst))
      (setf str (pack lst))
      (if (null str) (setf str ""))
      (cond
       ((not (null s))
	(search-from n)
	(beginning-of-line 1)
	(kill-ring-suspend t)
	(kill-line 1)
	(pop-kill-ring)
	(init-line-edit s)
	(move-point-to 0)
	(need-refresh-line t))
       (t (setf found (history-pkg:history-search-backward str search-frm))
	  (when found
	    (search-from (caar found))
	    (beginning-of-line 1)
	    (kill-ring-suspend t)
	    (kill-line 1)
	    (pop-kill-ring)
	    (init-line-edit (history-pkg:hist-buf (cdar found)))
	    (move-point-to pos)
	    (need-refresh-line t))))))


  ;;; un-binded function
  ;;; ǻꤵ줿ֹ˰ư롣
  ;;;
  (defun goto-history (num)
    (let (s n pos)
      (if (or (not (integerp num)) (not (history-pkg:hist-range-p num)))
	  (return-from goto-history nil))
      (setf pos *point*)
      (multiple-value-setq (s n) (expand-history (make-hist-string num)))
      (search-from num)
      (beginning-of-line 1)
      (kill-ring-suspend t)
      (kill-line 1)
      (pop-kill-ring)
      (init-line-edit s)
      (move-point-to pos)
      (need-refresh-line t)))

  (defun make-hist-string (n)
    (let ((lst nil) num i)
      (setf num n)
      (loop
	(if (zerop num) (return-from make-hist-string (push #\! lst)))
	(multiple-value-setq (num i) (floor num 10))
	(push (digit-to-char i) lst))))

  
  ;;; C-n
  ;;; ʿʤˤ˸롣
  ;;; Ǹ˸˰פ֤Ͽ꿷о
  ;;; ȤƸ롣
  ;;;
  ;;; (history-pkg::history-search-forward str search-frm)ϸ
  ;;; ȡ˰פʸ
  ;;;
  ;;; ((ҥȥֹ . ֹ) Ƭ) ֤
  ;;;
  ;;; ߤλͤǤϡԤƬΰפΤߤоݤȤƤ뤿ḡ
  ;;; Ƭ֤Ͼ˥Ǥ롣
  ;;;
  (defun next-line (&optional (n 1))
    (declare (fixnum n))
    (setf n (select-repeat-count n))
    (cond
     ((zerop n) nil)
     ((> n 0)
      (dotimes (i n) (true-next-line)))
     ((< n 0)
      (previous-line (- n)))))

  (defun true-next-line ()
    (let (pos found lst str search-frm s n)
      (setf pos *point*)
      (setf search-frm (search-from))
      (when (and search-frm (= search-frm (car (history-pkg:current-history))))
	(kill-text)
	(return-from true-next-line nil))
      (setf lst (subseq *text* 0 pos))
      (setf str (pack lst))
      (if (null str) (setf str ""))
      (multiple-value-setq (s n) (expand-history lst))
      (cond
       ((not (null s))
	(search-from n)
	(beginning-of-line 1)
	(kill-ring-suspend t)
	(kill-line 1)
	(pop-kill-ring)
	(init-line-edit s)
	(move-point-to 0)
	(need-refresh-line t))
       (t (setf found (history-pkg:history-search-forward str search-frm))
	  (when found
	    (search-from (caar found))
	    (beginning-of-line 1)
	    (kill-ring-suspend t)
	    (kill-line 1)
	    (pop-kill-ring)
	    (init-line-edit (history-pkg:hist-buf (cdar found)))
	    (move-point-to pos)
	    (need-refresh-line t))))))

  ;;;
  ;;; Ǹ˸ֹϿ롣
  ;;; ꤷʤäϡϿƤֹ֤
  ;;; nilϺǿֹ̣롣
  ;;;
  (defun search-from (&optional (num nil sw))
    (cond
     ((null sw) *search-from*)
     (t (setf *search-from* num))))

  ;;;
  ;;; 󥰹¤Хåե˵ϿƤǤŤ򤫡
  ;;;
  (defun oldest-history-p (hist-num)
    (and (numberp hist-num)
	 (= hist-num (car (history-pkg:last-live-history)))))

  ;;;
  ;;; ҥȥֹɽʸꥹȤǤ
  ;;; бʸֹ֤Ǥʤ nil֤
  ;;;
  ;;; <ҥȥֹ> ::= '!' ['-'] <number>
  ;;;			 '!' '!'
  ;;;
  ;;; ex. (expand-history (unpack "!-3"))
  ;;; 	  (expand-history (make-hist-string 1234)))
  ;;;
  (defun expand-history (lst)
    (let (num n s)
      (loop				;skip white space
	(cond
	 ((null lst) (return-from expand-history nil))
	 ((char= (car lst) #\!)
	  (pop lst)
	  (return))
	 ((member (car lst) *white-space-char*) (pop lst))
	 (t (return-from expand-history nil))))
      (loop
	(cond
	 ((null lst) (return-from expand-history nil))
	 ((member (car lst) *white-space-char*) (pop lst))
	 (t (return))))
      (cond
       ((null lst)
	(return-from expand-history nil))
       ((char= (car lst) #\!)
	(setf n (car (history-pkg:current-history)))
	(setf s (history-pkg:hist-buf (history-pkg::get-hist n))))
       ((char= (car lst) #\-)
	(setf num (meta-string-to-number (pack lst) :default 4))
	(setf n (+ (car (history-pkg:current-history)) num 1))
	(setf s (history-pkg:hist-buf (history-pkg::get-hist n))))
       ((digit-char-p (car lst))
	(setf n (meta-string-to-number (pack lst) :default 4))
	(setf s (history-pkg:hist-buf (history-pkg::get-hist n))))
       (t (return-from expand-history nil)))
      (return-from expand-history (values s n))))

  ;;; M-<
  ;;; 󥰹¤ХåեƬʺǸŤ򤬵ϿƤ
  ;;; ˰ư롣
  ;;;
  (defun beginning-of-history ()
    (let (last-live)
      (setf last-live (history-pkg:last-live-history))
      (search-from (car last-live))
      (beginning-of-line 1)
      (kill-ring-suspend t)
      (kill-line 1)
      (pop-kill-ring)
      (init-line-edit (history-pkg:hist-buf (cdr last-live)))
      (if *undo-info* (line-edit-break "beginning-of-history"))
      (need-refresh-line t)
      (beginning-of-line 1)))

  ;;; M->
  ;;; 󥰹¤ХåեʺǿϿƤ֡
  ;;; ˰ư롣
  ;;;
  (defun end-of-history () (kill-text))

  ;;; [C-u [num]] M-m x (Mark global-mark)
  ;;; ְǻꤵ줿ֹޡ 'x'˵Ͽ롣ְ
  ;;; ꤵƤʤиߤֹϿ롣
  ;;; ְ 0 ʤ顢Х롦ޡ 'x'ΤΤ롣
  ;;; 衢ꤵ줿󥰡ХåեäƤޤ
  ;;; ơƤ¸Ƥ
  ;;;
  ;;; ꥹȤϰʲ̤ꡧ
  ;;;	(ޡ ֹ ݥȰ )
  ;;;
  (defun set-global-mark ()
    (let (mark hist-num str)
      (setf mark (getchar))
      (cond
       ((repeat-count)
	(setf hist-num (repeat-count))
	(setf str (expand-history (make-hist-string hist-num))))
       (t (setf hist-num (1+ (car (history-pkg:current-history))))
	  (setf str (packed-text))))
      (pure-set-global-mark mark hist-num (point) str)))

  (defun pure-set-global-mark (mark hist-num point str)
    (cond
     ((null mark) nil)
     ((null (assoc mark *global-mark*))
      (push (list mark hist-num point str) *global-mark*))
     (t (setf (cdr (assoc mark *global-mark*)) (list hist-num point str))))
    (setf *global-mark*
	  (remove-if #'(lambda (x) (zerop (second x))) *global-mark*))
    (sort-global-mark))

  ;;;
  ;;; Х롦ޡ 'x'Ƥ֤
  ;;;
  (defun get-global-mark (x)
    (assoc x *global-mark*))

  ;;;
  ;;; Х롦ޡ 'x'ΥݥȰֹ֤֤¿͡ˡ
  ;;; ꤵ줿Х롦ޡ¸ߤƤʤ nil֤
  ;;;
  (defun eval-global-mark (x)
    (let (lst)
      (setf lst (get-global-mark x))
      (cond
       ((null lst) (values nil nil))
       (t (values (third lst) (second lst))))))

  ;;;
  ;;; Х롦ޡƤɽ롣
  ;;;
  (defun view-global-mark ()
    (if (null *global-mark*) (return-from view-global-mark nil))
    (dolist (i *global-mark* t) (prin1 i) (terpri)))

  ;;;
  ;;; Х롦ޡ򥽡Ȥ롣
  ;;;
  (defun sort-global-mark ()
    (setf *global-mark*
	  (sort *global-mark* #'(lambda (x y) (char< (car x) (car y))))))

  ;;;
  ;;; ٤ƤΥХ롦ޡõ롣
  (defun clear-global-mark ()
    (setf *global-mark* nil))

  ;;;
  ;;; Х롦ޡƤե¸롣
  ;;;
  (defun save-global-mark (&optional fname)
    (if (null fname) (setf fname (global-mark-file)))
    (with-open-file (stream fname :direction :output)
      (dolist (i *global-mark* t)
	(prin1 i stream)
	(terpri stream))))

  ;;;
  ;;; Х롦ޡ¸եꤹ롣
  ;;;
  (defun global-mark-file (&optional fname)
    (cond
     ((null fname) *global-mark-file*)
     (t (setf *global-mark-file* fname))))

  ;;;
  ;;; Х롦ޡɤ߹ࡣ
  ;;;
  (defun load-global-mark (&optional fname)
    (let (form)
      (if (null fname) (setf fname (global-mark-file)))
      (when (probe-file fname)
	(clear-global-mark)
	(with-open-file (stream fname :direction :input)
	  (loop
	    (setf form (read stream nil *eos*))
	    (when (eq form *eos*) (return-from load-global-mark t))
	    (pure-set-global-mark
	     (first form) (second form) (third form) (fourth form)))))))

  ;;; M-g x (Go to global-mark)
  ;;; ꤵ줿Х롦ޡΰ֤˰ư롣
  ;;; Х롦ޡǻꤵ줿򤬥󥰡Хåե˻ĤäƤ
  ;;; ˰ưǤʤС餫¸ƤƤɽ
  ;;; 롣
  ;;;
  (defun goto-global-mark ()
    (let (lst mark hist-num point str)
      (setf mark (getchar))
      (setf lst (get-global-mark mark))
      (setf hist-num (second lst))
      (setf point (third lst))
      (setf str (fourth lst))
      (cond
       ((history-pkg:hist-range-p hist-num)
	(goto-history hist-num))
       (t
	(beginning-of-line 1)
	(kill-ring-suspend t)
	(kill-line 1)
	(pop-kill-ring)
	(init-line-edit str)))
      (if (null point) (move-point-to 0)
	(move-point-to point)))) )
